//
//  ivAppObjAlgoAddShadowConstraints.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoAddShadowConstraints.h"

//the function to remap shadow point
inline float updateShdToPtLinear
( cv::Point2f &shdToPt, float height, float WRP_btmT, const cv::Point2f &objToPt )
{
    height = std::max( 0.0f, height );
    
    float weight = 0.0f;
    if( WRP_btmT>0.0f )
        weight = 1.0f - std::min( height/WRP_btmT, 1.0f );
    
    shdToPt = (1.0f-weight) * shdToPt + weight * objToPt;
    return weight;
}

void ivAppObjAlgoAddShadowConstraints::adjustPointConstraints(ivAppObject *obj)
{
    const ivAppObjParams &objParams = obj->objParams;
    ivAppObjectData &objData = obj->objData;
    std::vector<ivAppObjectPart> &objParts = objData.parts;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    
    //for each cluster
    const int clusterNum = (int)objParts.size();
    for( int i=0; i<clusterNum; i++ )
    {
        ivAppObjectPart &curPart = objParts[i];
        
        ObjShdWarpData &oswData = objFrameData[i];
        const WarpData &object = oswData.object;
        ShadowWarpData &shadow = oswData.shadow;
        if( !object.isReadyForWarping || !shadow.isReadyForWarping )
            continue;
        
        //for each visible inpatch point from camera & light in the cluster
        int oriPtNum = (int)curPart.oriPtNum;
        for( int j=0; j<oriPtNum; j++ )
        {
            int ptID = curPart.ptIDs[j];
            
            int pid_obj = object.sfm2fp[ptID];
            int pid_shd = shadow.sfm2fp[ptID];
            if( pid_obj<0 || pid_shd<0 )
                continue;
            
            const FrontPoint &objfp = object.frontPoints[ pid_obj ];
            FrontPoint &shdfp = shadow.frontPoints[ pid_shd ];
            
            //blending the bottom point near the bottom
            float height = objfp.pt3d(2);
            if( height >= objParams.WRP_btmT )
                continue;
            
            float weight = updateShdToPtLinear( shdfp.toPt, height, objParams.WRP_btmT, objfp.toPt );
            weight *= objfp.pWeight;
            shadow.addBottomPoint( shdfp.fromPt, shdfp.toPt, weight );
        }
    }
}

void ivAppObjAlgoAddShadowConstraints::adjustEdgeConstraints(ivAppObject *obj)
{
    const ivAppObjParams &objParams = obj->objParams;
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();

    //for each cluster
    const int clusterNum = (int)objData.parts.size();
    for( int i=0; i<clusterNum; i++ )
    {
        ObjShdWarpData &oswData = objFrameData[i];
        const WarpData &object = oswData.object;
        ShadowWarpData &shadow = oswData.shadow;
        if( !object.isReadyForWarping || !shadow.isReadyForWarping )
            continue;
        
        //for each visible inpatch edge point from camera & light in the cluster
        const int tvptNum = (int)obj->objTvEdges.vtvPoint.size();
        for( int j=0; j<tvptNum; j++ )
        {
            //if the point is visible & inpatch from camera and light
            int ftvpid_obj = object.tvp2ftvp[j];
            int ftvpid_shd = shadow.tvp2ftvp[j];
            if( ftvpid_obj<0 || ftvpid_shd<0 )
                continue;
            
            const FrontPoint &objftvp = object.frontTvPoints[ ftvpid_obj ];
            FrontPoint &shdftvp = shadow.frontTvPoints[ ftvpid_shd ];
            
            //blending the bottom point near the bottom
            float height = objftvp.pt3d(2);
            if( height >= objParams.WRP_btmT )
                continue;
            
            float weight = updateShdToPtLinear( shdftvp.toPt, height, objParams.WRP_btmT, objftvp.toPt );
            weight *= objftvp.pWeight;
            shadow.addBottomPoint( shdftvp.fromPt, shdftvp.toPt, weight );
        }
        
        //update shadow edge parameters
        const std::vector<FrontPoint> sfrontTvPoints = shadow.frontTvPoints;
        std::vector<FrontEdge> sfrontEdges = shadow.frontEdges;
        std::vector<FrontEdge>::iterator sfeit = sfrontEdges.begin();
        for( ; sfeit!=sfrontEdges.end(); sfeit++ )
            sfeit->calcEdgeParams( sfrontTvPoints );
    }
}



bool ivAppObjAlgoAddShadowConstraints::execute(ivAppObject *obj)
{
    //adjust shadow point correspondence
    this->adjustPointConstraints(obj);

    //adjust shadow edge correspondence
    this->adjustEdgeConstraints(obj);
    
    return true;
}
