//
//  ivAppObjAlgoCalcMotionField.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoCalcMotionField.h"

//calc dense object motion field
inline void calcMotionField(ivAppObject *obj, bool isPatch)
{
    ivAppObjectData &objData = obj->objData;
    const bool enoughFrames = objData.frames.size() >= 2;
    
    const ivAppObjParams &objParams = obj->objParams;
    const ivAppTvEdges &objTvEdges = obj->objTvEdges;
    const int tvpNum = (int)objTvEdges.vtvPoint.size();
    
    //the frame data of the current frame
    std::list<ivAppObjectFrameData>::iterator it_cur = --(objData.frames.end());
    
    //initialize motion maps and their visualization
    cv::Mat &motionField = isPatch ? it_cur->objectMotionField : it_cur->shadowMotionField;
    motionField = cv::Mat( OSZ_HEIGHT, OSZ_WIDTH, CV_32FC2, cv::Scalar::all(0.0f) );
    
    if( !enoughFrames )
        return;

    //the frame data of the previous frame
    std::list<ivAppObjectFrameData>::iterator it_pre = it_cur; it_pre--;
    
    //sparse to dense
    lsqSparseToDenseMap s2dx( OSZ_WIDTH, OSZ_HEIGHT, objParams.S2D_gridSz );
    lsqSparseToDenseMap s2dy( OSZ_WIDTH, OSZ_HEIGHT, objParams.S2D_gridSz );
    
    //collect motion data
    const int clusterNum = (int)objData.parts.size();
    for(int i=0;i<clusterNum;i++)
    {
        const ivAppObjectPart &curPart = objData.parts[i];
        
        const ObjShdWarpData &oswDataPre = (*it_pre)[i];
        const ObjShdWarpData &oswDataCur = (*it_cur)[i];
        
        const WarpData &wDataPre = isPatch ? oswDataPre.object : oswDataPre.shadow;
        const WarpData &wDataCur = isPatch ? oswDataCur.object : oswDataCur.shadow;
        
        if( !wDataPre.isReadyForWarping || !wDataCur.isReadyForWarping )
            continue;
        
        //for each point in the cluster
        for( int j=0; j<curPart.oriPtNum; j++ )
        {
            int ptID = curPart.ptIDs[j];
            
            int fptIDPre = wDataPre.sfm2fp[ptID];
            int fptIDCur = wDataCur.sfm2fp[ptID];
            
            if( fptIDPre<0 || fptIDCur<0 )
                continue;
            //the point is visible in both frames
            
            const cv::Point2f &ptPre = wDataPre.frontPoints[fptIDPre].toPt;
            const cv::Point2f &ptCur = wDataCur.frontPoints[fptIDCur].toPt;
            cv::Point2f v = ptPre - ptCur;
            
            s2dx.addValue( ptCur, v.x );
            s2dy.addValue( ptCur, v.y );
        }    //calc dense object motion field
        
        //for each time-varying point
        for( int j=0; j<tvpNum; j++ )
        {
            int ftvptIDPre = wDataPre.tvp2ftvp[j];
            int ftvptIDCur = wDataCur.tvp2ftvp[j];
            
            if( ftvptIDPre<0 || ftvptIDCur<0 )
                continue;
            //the tv point is visible in both frames
            
            const cv::Point2f &ptPre = wDataPre.frontTvPoints[ftvptIDPre].toPt;
            const cv::Point2f &ptCur = wDataCur.frontTvPoints[ftvptIDCur].toPt;
            cv::Point2f v = ptPre - ptCur;
            
            s2dx.addValue( ptCur, v.x );
            s2dy.addValue( ptCur, v.y );
        }
    }
    
    //least squre optimization
    s2dx.goLSQ( objParams.S2D_dataW, objParams.S2D_smthW );
    s2dy.goLSQ( objParams.S2D_dataW, objParams.S2D_smthW );
    
    //get dense motion field map
    cv::Mat dx = s2dx.getDenseMap();
    cv::Mat dy = s2dy.getDenseMap();
    
    for(int i=0;i<OSZ_HEIGHT;i++) for(int j=0;j<OSZ_WIDTH;j++)
    {
        cv::Vec2f &v = motionField.at<cv::Vec2f>(i,j);
        v[0] = dx.at<float>(i,j);
        v[1] = dy.at<float>(i,j);
    }
}

bool ivAppObjAlgoCalcObjectMotionField::execute(ivAppObject *obj)
{
    calcMotionField( obj, true );
    
    return true;
}

bool ivAppObjAlgoCalcShadowMotionField::execute(ivAppObject *obj)
{
    calcMotionField( obj, false );
    
    return true;
}