//
//  ivAppObjAlgoCoGEPWarping.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoCoGEPWarping.h"

inline void coWarp(ivAppObject *obj, bool isPatch, bool useShared)
{
    ivAppObjectFrameData &objFrameData = obj->objData.frames.back();
    const ivAppObjParams &objParams = obj->objParams;
    const int sfmPtNum = (int)obj->objSfM.vPointData.size();
    const int tvptNum = (int)obj->objTvEdges.vtvPoint.size();
    const cv::Size osz = cv::Size(OSZ_WIDTH, OSZ_HEIGHT);
    
    CoGEPWarper &coWarper = isPatch ? objFrameData.objectCoWarper : objFrameData.shadowCoWarper;
    
    const std::vector<int> &usedClusterIDs =
        isPatch ? objFrameData.objectUsedClusterIDs : objFrameData.shadowUsedClusterIDs;
 
    if( useShared )
    {
        //align shared points
        for( int i=0; i<sfmPtNum; i++ )
        {
            //the cluster pair where the clusters are ready for warping
            for( int j=0; j<(int)usedClusterIDs.size(); j++ )
            for( int k=j+1; k<(int)usedClusterIDs.size(); k++ )
            {
                int cidJ = usedClusterIDs[j];
                const WarpData &wDataJ = isPatch ? objFrameData[cidJ].object : objFrameData[cidJ].shadow;
                int fpid_j = wDataJ.sfm2fp[i];
                
                int cidK = usedClusterIDs[k];
                const WarpData &wDataK = isPatch ? objFrameData[cidK].object : objFrameData[cidK].shadow;
                int fpid_k = wDataK.sfm2fp[i];
                
                //the point is not in one of the clusters
                if( fpid_j<0 || fpid_k<0 )
                    continue;
                
                const cv::Point2f &fromPtJ = wDataJ.frontPoints[fpid_j].fromPt;
                const cv::Point2f &fromPtK = wDataK.frontPoints[fpid_k].fromPt;

                float pWeight = wDataJ.frontPoints[fpid_j].pWeight;
                coWarper.addCoPointConstraint( j, fromPtJ, k, fromPtK, pWeight );
            }
        }

        //align shared edge end-points
        for( int i=0; i<tvptNum; i++ )
        {
            //the cluster pair where the clusters are ready for warping
            for( int j=0; j<(int)usedClusterIDs.size(); j++ )
            for( int k=j+1; k<(int)usedClusterIDs.size(); k++ )
            {
                int cidJ = usedClusterIDs[j];
                const WarpData &wDataJ = isPatch ? objFrameData[cidJ].object : objFrameData[cidJ].shadow;
                int ftvpid_j = wDataJ.tvp2ftvp[i];
                
                int cidK = usedClusterIDs[k];
                const WarpData &wDataK = isPatch ? objFrameData[cidK].object : objFrameData[cidK].shadow;
                int ftvpid_k = wDataK.tvp2ftvp[i];
                
                //the end point is not in one of the clusters
                if( ftvpid_j<0 || ftvpid_k<0 )
                    continue;
                
                const cv::Point2f &fromPtJ = wDataJ.frontTvPoints[ftvpid_j].fromPt;
                const cv::Point2f &fromPtK = wDataK.frontTvPoints[ftvpid_k].fromPt;
                
                float pWeight = wDataJ.frontTvPoints[ftvpid_j].pWeight;
                coWarper.addCoPointConstraint( j, fromPtJ, k, fromPtK, pWeight );
            }
        }
    }
    
    //go collaborative warping (return the errors of warps)
    Eigen::VectorXf warpErrors = coWarper.optimizeForWarps
    (objParams.WRP_ftrW, objParams.WRP_simW, objParams.WRP_edgeW, objParams.WRP_coW,
     objParams.WRP_btmW );
    
    //remap the features of object patch ( for shadow alignement )
    for( int i=0; i<(int)usedClusterIDs.size(); i++ )
    {
        GEPWarper &warper = coWarper.at( i );
        
        int cidI = usedClusterIDs[i];
        WarpData &wData = isPatch ? objFrameData[cidI].object : objFrameData[cidI].shadow;
        
        //remap the points to warped space
        std::vector<FrontPoint> &frontPoints = wData.frontPoints;
        for( std::vector<FrontPoint>::iterator fpit = frontPoints.begin(); fpit!=frontPoints.end(); fpit++ )
            warper.remapPoint( fpit->fromPt, fpit->toPt );

        //remap the tv edge end points to warped space
        std::vector<FrontPoint> &frontTvPoints = wData.frontTvPoints;
        for(std::vector<FrontPoint>::iterator ftvpit = frontTvPoints.begin(); ftvpit!=frontTvPoints.end(); ftvpit++ )
            warper.remapPoint( ftvpit->fromPt, ftvpit->toPt );

        //update edge params
        std::vector<FrontEdge> &frontEdges = wData.frontEdges;
        for(std::vector<FrontEdge>::iterator feit = frontEdges.begin(); feit!=frontEdges.end(); feit++ )
            feit->calcEdgeParams( frontTvPoints );
    }
    
    //warp results
    for( int i=0; i<(int)usedClusterIDs.size(); i++ )
    {
        GEPWarper &warper = coWarper.at( i );
        
        int cidI = usedClusterIDs[i];
        WarpData &wData = isPatch ? objFrameData[cidI].object : objFrameData[cidI].shadow;
        
        //new weight map for collage
        wData.prepareWeightMap( osz );
        
        //warp results
        wData.warpFrame = warper.getDeformedImg( osz, objParams.WRP_rgrid );

        //warp resulting feature maps
        wData.warpPointMap = warper.getPointMap( osz );
        wData.warpEdgeMap = warper.getEdgeMap( osz, objParams.CLG_edgeRange );
        wData.warpEdgeCornerMap = warper.getEdgeCornerMap( osz );
        
        //warp errors
        wData.warpError = warpErrors[i];
        wData.warpDistortionMap = warper.getDistortionMap( osz ) / objParams.WRP_gridSize;
    }
}

bool ivAppObjAlgoCoGEPWarpingObjectPatch::execute(ivAppObject *obj)
{
    coWarp( obj, true, true );
    return true;
}

bool ivAppObjAlgoCoGEPWarpingShadowPatch::execute(ivAppObject *obj)
{
    coWarp( obj, false, true );
    return true;
}
















