//
//  ivAppObjAlgoCollage.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoCollage.h"

inline void callCollager(ivAppObject *obj, bool isPatch, bool useEdge, bool considerTemporal)
{
    const ivAppObjParams &objParams = obj->objParams;
    ivAppObjectData &objData = obj->objData;
    
    float CLG_feature   = objParams.CLG_feature;
    float CLG_dstrn     = objParams.CLG_dstrn * isPatch;
    float CLG_sm_clr    = objParams.CLG_sm_clr;
    float CLG_sm_grad   = objParams.CLG_sm_grad;
    float CLG_sm_edge   = objParams.CLG_sm_edge * useEdge;
    float CLG_tm_alpha  = objParams.CLG_tm_alpha;
    float CLG_tm_w      = objParams.CLG_tm_w * considerTemporal;
    float CLG_edgeRange = objParams.CLG_edgeRange;
    
    const int preFrameNum = (int)objData.frames.size() - 1;
    const int clusterNum = (int)objData.parts.size();
    
    //prepare temporal Weight Map
    //===========================================================================================
    std::vector< cv::Mat > tmprlClusterWeightMaps(clusterNum);
    for(int i=0;i<clusterNum;i++)
        tmprlClusterWeightMaps[i] = cv::Mat( OSZ_HEIGHT, OSZ_WIDTH, CV_32FC1, cv::Scalar(0.0f) );

    //remapping index
    cv::Mat remapIndexMap = cv::Mat( OSZ_HEIGHT, OSZ_WIDTH, CV_32FC2, cv::Scalar::all(0.0f) );
    for(int i=0;i<OSZ_HEIGHT;i++) for(int j=0;j<OSZ_WIDTH;j++)
    {
        cv::Vec2f &av = remapIndexMap.at<cv::Vec2f>(i,j);
        av << j, i;
    }

    //prepare previous collagers and motion fields
    std::list<ivAppObjectFrameData>::reverse_iterator oft_rit0 = ++( objData.frames.rbegin() ); // i-1
    std::list<ivAppObjectFrameData>::reverse_iterator oft_rit1 = objData.frames.rend();         // i-CLG_tm_fnum

    for( std::list<ivAppObjectFrameData>::reverse_iterator rit = oft_rit0; rit!=oft_rit1; rit++ )
    {
        std::list<ivAppObjectFrameData>::reverse_iterator rit2 = rit; rit2--;
        cv::Mat &motionField = isPatch ? rit2->objectMotionField : rit2->shadowMotionField;
        
        //calculate the remapping index map
        for(int pi=0;pi<OSZ_HEIGHT;pi++) for(int pj=0;pj<OSZ_WIDTH;pj++)
        {
            cv::Vec2f &rim = remapIndexMap.at<cv::Vec2f>(pi,pj);
            
            //boundary check
            if( rim[0]<0 || rim[0]>=OSZ_WIDTH || rim[1]<0 || rim[1]>=OSZ_HEIGHT )
                continue;
            
            const cv::Vec2f &v = motionField.at<cv::Vec2f>( rim[1], rim[0] );
            rim += v;
        }
        
        
        patchCollager &preCollager = isPatch ? rit->objectCollager : rit->shadowCollager;
        
        // prepare weight maps
        for(int i=0;i<clusterNum;i++)
        {
            cv::Mat srcidwmap;
            preCollager.getSrcIDWeightMap( srcidwmap, i );
            
            for(int pi=0;pi<OSZ_HEIGHT;pi++) for(int pj=0;pj<OSZ_WIDTH;pj++)
            {
                const cv::Vec2f &rim = remapIndexMap.at<cv::Vec2f>(pi,pj);
                
                //boundary check
                if( rim[0]<0 || rim[0]>=OSZ_WIDTH || rim[1]<0 || rim[1]>=OSZ_HEIGHT )
                    continue;
                
                float &srcw = srcidwmap.at<float>( rim[1], rim[0] );
                float &tcw = tmprlClusterWeightMaps[i].at<float>( pi, pj );
                
                tcw += srcw / preFrameNum;
            }
            
        }
    }
    //===========================================================================================
    
    
    //collage the patch
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    patchCollager &collager = isPatch ? objFrameData.objectCollager : objFrameData.shadowCollager;
    collager.clear();
    collager.initSize( OSZ_WIDTH, OSZ_HEIGHT );

    for(int i=0;i<clusterNum;i++)
    {
        Eigen::Vector3f color;
        Utils_Ocv::instance().getColorBGR(color, i, clusterNum-1, 0);
        
        //collage object patch
        ObjShdWarpData &oswData = objFrameData[i];
        WarpData &wData = isPatch ? oswData.object : oswData.shadow;
        if( !wData.isReadyForWarping )
            continue;

        //do collage
        collager.add(i, color,
                     wData.warpFrame, wData.warpWeightMap, wData.warpDistortionMap,
                     wData.warpEdgeMap, wData.warpEdgeCornerMap, tmprlClusterWeightMaps[i],
                     CLG_feature, CLG_dstrn, CLG_sm_clr, CLG_sm_grad, CLG_sm_edge, CLG_edgeRange, CLG_tm_w );
    }
    
    //calculate source weight map for each cluster
    collager.calcSrcIDWeightMaps( CLG_tm_alpha );
}

bool ivAppObjAlgoObjectColorGradEdgeTemporalCollage::execute(ivAppObject *obj)
{
    callCollager( obj, true, true, true );
    return true;
}

bool ivAppObjAlgoShadowColorGradEdgeTemporalCollage::execute(ivAppObject *obj)
{
    callCollager( obj, false, true, true );
    return true;
}