//
//  ivAppObjAlgoMakeObject.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoMakeObject.h"

inline cv::Point2f subPoint(const cv::Point2f &p, int x, int y)
{
    return p-cv::Point2f(x,y);
}

//calc dense object depth map / diffuse intensity map
void ivAppObjAlgoMakeObject::interpolateAttributeMaps
(const ivAppObject *obj, const cv::Rect &objBB, cv::Mat &depthMap, cv::Mat &diffuseMap)
{
    const ivAppObjectData &objData = obj->objData;
    const ivAppObjParams &objParams = obj->objParams;
    const ivAppObjectFrameData &objFrameData = objData.frames.back();
    const int clusterNum = (int)objData.parts.size();
    
    //sparse to dense
    //====================================================================================
    lsqSparseToDenseMap s2d_depth( objBB.width, objBB.height, objParams.S2D_gridSz );
    lsqSparseToDenseMap s2d_intensity( objBB.width, objBB.height, objParams.S2D_gridSz );
    
    //collect point data with depth value (value, value0, value1)
    for(int i=0;i<clusterNum;i++)
    {
        const WarpData &wData = objFrameData[i].object;
        
        if( !wData.isReadyForWarping )
            continue;
        
        //for each inpatch front point
        for( int j=0; j<wData.oriPtNum; j++ )
        {
            cv::Point2f p = subPoint( wData.frontPoints[j].toPt, objBB.x, objBB.y );
            s2d_depth.addValue( p, wData.frontPoints[j].value );
            s2d_intensity.addValue( p, wData.frontPoints[j].intensity );
        }
        
        //for each inpatch edge
        for(std::vector<FrontPoint>::const_iterator fpit = wData.frontTvPoints.begin();
            fpit!=wData.frontTvPoints.end(); fpit++)
        {
            cv::Point2f p = subPoint( fpit->toPt, objBB.x, objBB.y );
            s2d_depth.addValue( p, fpit->value );
        }
    }
    
    //least squre optimization
    s2d_depth.goLSQ( objParams.S2D_dataW, objParams.S2D_smthW );
    s2d_intensity.goLSQ( objParams.S2D_dataW, objParams.S2D_smthW );
    
    //get dense map
    depthMap = s2d_depth.getDenseMap();
    diffuseMap = s2d_intensity.getDenseMap();
    //====================================================================================
}

//adjust object intensity
void ivAppObjAlgoMakeObject::adjustObjectIntensity(ivAppObject *obj, cv::Mat &objDiffuse)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    ivAppObjParams &objParams = obj->objParams;
    
    const ivAppScene &scene = ivAppScene::instance();
    cv::Mat objPatchBGRA = objFrameData.objPatch;
    
    //apply diffuse lighting
    //====================================================================================
    cv::Mat diffuseAdjMap = 1.0f + ( objDiffuse - 0.5f ) * objParams.LT_kd * scene.scnLight.lightIntensity;
    for(int i=0;i<objPatchBGRA.rows;i++)
    for(int j=0;j<objPatchBGRA.cols;j++)
    {
        cv::Vec4b &c = objPatchBGRA.at<cv::Vec4b>(i,j);
        const float &d = diffuseAdjMap.at<float>(i,j);
        for(int k=0;k<3;k++) c[k] = cv::saturate_cast<uchar>( c[k] * d );
    }
    //====================================================================================
}


bool ivAppObjAlgoMakeObject::execute(ivAppObject *obj)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    ivAppObjParams &objParams = obj->objParams;
    patchCollager &collager = objFrameData.objectCollager;
    
    if( !collager.isReady() )
        return true;
    
    //collage result
    if( !objParams.CLG_usePoisson ) //original collage
        objFrameData.objPatch = collager.getImgCollage( objParams.CLG_alpha );
    else //poisson blended version of collage
        objFrameData.objPatch = collager.getBlendedCollage();
    
    //initial depth map (infinity)
    cv::Scalar infDepth = cv::Scalar( std::numeric_limits<float>::max() );
    objFrameData.objDepth = cv::Mat( OSZ_HEIGHT, OSZ_WIDTH, CV_32FC1, infDepth );
    
    //calc bounding box
    std::vector<cv::Mat> objPatchChannels;
    cv::split( objFrameData.objPatch, objPatchChannels );
    Utils_Ocv &uocv = Utils_Ocv::instance();
    cv::Rect objBB;
    if( !uocv.getFGBB( objBB, objPatchChannels[3] ) )//invisible
        return true;
    
    //make object depth map / intensity map
    //====================================================================================
    //object roi
    cv::Mat objMskROI = objPatchChannels[3]( objBB );
    
    //attribute interpolation
    cv::Mat depthMap, diffuseMap;
    interpolateAttributeMaps( obj, objBB, depthMap, diffuseMap );
    
    //depth map
    cv::Mat roiDepth( objFrameData.objDepth, objBB );
    depthMap.copyTo( roiDepth, objMskROI>0 );
    
    //intensity map
    cv::Mat objDiffuse = cv::Mat( OSZ_HEIGHT, OSZ_WIDTH, CV_32FC1, cv::Scalar(0.0f) );
    cv::Mat roiIntensity( objDiffuse, objBB );
    diffuseMap.copyTo( roiIntensity, objMskROI>0 );
    //====================================================================================
    
    //adjust object intensity
    adjustObjectIntensity( obj, objDiffuse );
    
    return true;
}