//
//  ivAppObjAlgoMakeShadow.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoMakeShadow.h"

inline cv::Point2f subPoint(const cv::Point2f &p, int x, int y)
{
    return p-cv::Point2f(x,y);
}

//calc toObj Distance and adjust shadow intensity
void ivAppObjAlgoMakeShadow::applyShadowIntensity
(cv::Mat &shadowMask, cv::Rect &shdBB, ivAppObject *obj)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    ivAppObjParams &objParams = obj->objParams;
    
    //shadow patch
    cv::Mat shdROI = shadowMask(shdBB);
    
    //object center
    Eigen::Vector4f objCenter;
    objCenter << 0.0f, 0.0f, 0.0f, 1.0f;
    objCenter = objFrameData.gTransform * objCenter;
    
    //object scale
    float invObjScale = 1.0f / obj->objAnimation.objPose._scl.maxCoeff();
    
    //light intensity
    float lightIntensity = ivAppScene::instance().scnLight.lightIntensity;
    
    //identity matrix
    const Eigen::Matrix4f id4f = Eigen::Matrix4f::Identity();
    
    //initial weight map : distance to object center in shdBB
    for(int i=0;i<shdBB.height;i++)for(int j=0;j<shdBB.width;j++)
    {
        uchar &c = shdROI.at<uchar>(i,j);
        
        Eigen::Vector2f pt2d;
        pt2d << j+shdBB.x, i+shdBB.y;
        
        Eigen::Vector4f pt3d;
        Plane::ground.getIntersection( pt3d, pt2d, id4f, objFrameData.prjM_scn );
        
        //the attenuation is related to object size
        float dist2Cntr = (pt3d-objCenter).norm() * invObjScale;
        
        //remap to 1(close) ~ 0(far) by exp function
        float attenuationWeight = objParams.SHDW_gain * std::exp( -dist2Cntr * objParams.SHDW_alpha );
        
        //apply toObject distance weight and apply lighting intensity
        float newC = c * lightIntensity * attenuationWeight;
        newC = std::max( newC, 0.0f );
        newC = std::min( newC, 255.0f );
        c = newC;
    }
}

//calc dense shadow blurring weight and apply it to shadow patch
void ivAppObjAlgoMakeShadow::applyShadowBlurring
(cv::Mat &shadowMask, cv::Rect &shdBB, ivAppObject *obj)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    ivAppObjParams &objParams = obj->objParams;
    
    cv::Mat shdROI = shadowMask( shdBB );   //shadow mask in the bounding box
    
    //add border to the shadow patch
    cv::Mat shdBROI;
    const int &SHDW_bKSz = objParams.SHDW_bKSz;
    int hSHDW_bKSz = SHDW_bKSz / 2;
    cv::copyMakeBorder(shdROI, shdBROI, hSHDW_bKSz, hSHDW_bKSz, hSHDW_bKSz, hSHDW_bKSz,
                       cv::BORDER_CONSTANT, cv::Scalar::all(0) );
    cv::Rect shdBBB( shdBB.x-hSHDW_bKSz, shdBB.y-hSHDW_bKSz, shdBB.width+2*hSHDW_bKSz, shdBB.height+2*hSHDW_bKSz );
    
    //sparse to dense
    lsqSparseToDenseMap s2d( shdBBB.width, shdBBB.height, objParams.S2D_gridSz );
    
    //collect point data with shadow blurring weight value (value, value0, value1)
    const int clusterNum = (int)objData.parts.size();
    for(int i=0;i<clusterNum;i++)
    {
        const ObjShdWarpData &oswData = objFrameData[i];
        const WarpData &wData = oswData.shadow;
        
        //for each inpatch front point
        for( int j=0; j<wData.oriPtNum; j++ )
        {
            cv::Point2f p = subPoint( wData.frontPoints[j].toPt, shdBBB.x, shdBBB.y );
            s2d.addValue( p, wData.frontPoints[j].value );
        }
        
        //for each inpatch edge
        for(std::vector<FrontPoint>::const_iterator fpit = wData.frontTvPoints.begin();
            fpit!=wData.frontTvPoints.end(); fpit++)
        {
            cv::Point2f p = subPoint( fpit->toPt, shdBBB.x, shdBBB.y );
            s2d.addValue( p, fpit->value );
        }
    }
    
    //least square optimization
    s2d.goLSQ( objParams.S2D_dataW, objParams.S2D_smthW );
    
    //get dense shadow blurring weights
    cv::Mat shadowWeightMapBROI = s2d.getDenseMap();
    
    //shadow blur kernel weight map
    cv::Mat shadowBlurWeight;
    cv::threshold( shadowWeightMapBROI, shadowBlurWeight, 1.0f, 0.0, cv::THRESH_TRUNC );
    shadowBlurWeight *= objParams.SHDW_bKSz;
    
    //local blur shadow mask
    Utils_Ocv &ocv = Utils_Ocv::instance();
    cv::Mat blurredShdBROI = ocv.weightedGaussianBlur( shdBROI, shadowBlurWeight, objParams.SHDW_bsigma );
    
    //copy to shadow mask
    cv::Rect shdBBB_roi = shdBBB & cv::Rect( 0, 0, shadowMask.cols, shadowMask.rows );
    cv::Rect shdBBB_roi2 = shdBBB_roi;
    shdBBB_roi2.x -= shdBBB.x;
    shdBBB_roi2.y -= shdBBB.y;
    blurredShdBROI( shdBBB_roi2 ).copyTo( shadowMask( shdBBB_roi ) );
}


bool ivAppObjAlgoMakeShadow::execute(ivAppObject *obj)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    patchCollager &collager = objFrameData.shadowCollager;
    
    if( !collager.isReady() )
        return false;
    
    //the shadow mask
    cv::Mat shadowMask = collager.getImgCollage(0.0f);
    
    //foreground shadow region
    Utils_Ocv &uocv = Utils_Ocv::instance();
    cv::Rect shdBB;
    if( uocv.getFGBB( shdBB, shadowMask ) )//if the shadow is visible
    {
        //adjust shadow intensity
        this->applyShadowIntensity( shadowMask, shdBB, obj );
        
        //blur the shadow
        this->applyShadowBlurring( shadowMask, shdBB, obj );
    }

    //generate shadow patch
    objFrameData.shadowPatch = cv::Mat( shadowMask.size(), CV_8UC4, cv::Scalar::all(0) );
    for(int i=0;i<OSZ_HEIGHT;i++)for(int j=0;j<OSZ_WIDTH;j++)
    {
        uchar &a = shadowMask.at<uchar>(i,j);
        cv::Vec4b &c = objFrameData.shadowPatch.at<cv::Vec4b>(i,j);
        c << 0,0,0,a;
    }

    return true;
}

