//
//  ivAppObjAlgoPrepareEdges.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoPrepareEdges.h"

inline bool prepareTvFrontPoint
( FrontPoint &fp, const WarpData &wData, const ivAppObject *obj, bool isPatch )
{
    const ivAppObjectVideo &objVideo = obj->objVideo;
    const ivAppSfM &objSfM = obj->objSfM;
    const tvPoint3DSet &tvptset = obj->objTvEdges.vtvPoint;

    const ivAppObjectFrameData &objFrameData = obj->objData.frames.back();
    const Eigen::Matrix4f &prjM_scn = objFrameData.prjM_scn;
    const int &nframeID = wData.nfData.nframeID;
    const Eigen::Matrix4f prjM_obj = objSfM.vFrameData[nframeID].KM;
    
    const Eigen::Vector2f &p_obj2d = tvptset[ fp.ptID ].tvv2D[ nframeID ];
    const Eigen::Vector4f &p_scn = fp.pt3d;
    
    //2D projection of point in object patch
    cv::Point2f &objMsr = fp.fromPt;
    objMsr.x = p_obj2d(0);
    objMsr.y = p_obj2d(1);
    
    //2D projection of point in novel view
    cv::Point2f &scnMsr = fp.toPt;
    Eigen::Vector4f msr_scn = prjM_scn * p_scn;
    scnMsr.x = msr_scn(0) / msr_scn(2);
    scnMsr.y = msr_scn(1) / msr_scn(2);
    
    //point depth
    if( isPatch )
        fp.value = msr_scn(2);
    
    //the 2d point in image space should be in the patch/shadow boundary
    if( objVideo.remapPoint( objMsr, nframeID ) )
        return true;
    
    return false;
}

void ivAppObjAlgoPrepareEdges::prepareValidEdges
(WarpData &wData, const ivAppObject *obj, int partID, bool isPatch)
{
    if( !wData.isReadyForWarping )
        return;
    
    const ivAppScene &scene = ivAppScene::instance();
    const ivAppObjectData &objData = obj->objData;
    const ivAppObjectFrameData &objFrameData = objData.frames.back();
    const std::vector<ivAppObjectPart> &objParts = objData.parts;
    const ivAppTvEdges &objTvEdges = obj->objTvEdges;
    const tvPoint3DSet &tvptset = objTvEdges.vtvPoint;
    const int tvpNum = (int)tvptset.size();
    const std::vector<int> &edgeIDs = objParts[partID].edgeIDs;
    const int edgeNum = (int) edgeIDs.size();
    const int nframeID = wData.nfData.nframeID;
    
    std::vector<int> &tvp2ftvp = wData.tvp2ftvp;
    tvp2ftvp = std::vector<int>( tvpNum, -1 );
    
    std::vector<FrontEdge> &frontEdges = wData.frontEdges;
    std::vector<FrontPoint> &frontTvPoints = wData.frontTvPoints;

    //for each edge in the cluster
    for(int i=0;i<edgeNum;i++)
    {
        //the edge is visible in the selected frame
        const tvEdge &curEdge = objTvEdges.vtvEdge[ edgeIDs[i] ];
        
        //check if the edge is in the patch
        if( !curEdge.isVisible[nframeID] )
            continue;
        
        FrontPoint fp[2];
        bool isEdgeValid = true;
        for(int k=0;k<2;k++)
        {
            // if the tv point is used before, continue
            if( tvp2ftvp[ curEdge.vids(k) ] >= 0 )
                continue;
            
            fp[k].ptID = curEdge.vids(k);
            fp[k].pt3d = objFrameData.gTransform * tvptset[ fp[k].ptID ].v3D;
            fp[k].pWeight = 1.0f;
            
            //front shadow edge
            if( !isPatch )
            {
                Eigen::Vector4f v_pt3d = scene.scnLight.pos - fp[k].pt3d;
                Eigen::Vector4f new_pt3d;
                Plane::ground.getIntersection( new_pt3d, scene.scnLight.pos, v_pt3d );
                fp[k].value = ( new_pt3d - fp[k].pt3d ).norm(); //shadow blurring weight
                fp[k].pt3d = new_pt3d;
            }
            
            isEdgeValid = isEdgeValid && prepareTvFrontPoint( fp[k], wData, obj, isPatch );
        }
 
        if( !isEdgeValid )
            continue;

        //add the edge tv points
        for(int k=0;k<2;k++)
        {
            // if the tv point is used before, continue
            if( tvp2ftvp[ curEdge.vids(k) ] >= 0 )
                continue;
            
            frontTvPoints.push_back( fp[k] );
            tvp2ftvp[ curEdge.vids(k) ] = (int)frontTvPoints.size()-1;
        }
        
        //valid front patch edge
        FrontEdge fe;
        fe.edgeID = edgeIDs[i];
        for(int k=0;k<2;k++)
            fe.ftvptIDs(k) = tvp2ftvp[ curEdge.vids(k) ];
        fe.calcEdgeParams( frontTvPoints );
        
        frontEdges.push_back(fe);
    }
}

bool ivAppObjAlgoPrepareEdges::execute(ivAppObject *obj)
{
    //prepare edges
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    const int clusterNum = (int)objData.parts.size();
    for(int i=0;i<clusterNum;i++)
    {
        this->prepareValidEdges( objFrameData[i].object, obj, i,  true );
        this->prepareValidEdges( objFrameData[i].shadow, obj, i, false );
    }
    
    return true;
}
