//
//  ivAppObjAlgoPreparePoints.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoPreparePoints.h"

inline bool prepareFrontPoint
( FrontPoint &fp, const WarpData &wData, const ivAppObject *obj, bool isPatch )
{
    const ivAppObjectVideo &objVideo = obj->objVideo;
    const ivAppSfM &objSfM = obj->objSfM;
    
    const ivAppObjectFrameData &objFrameData = obj->objData.frames.back();
    const Eigen::Matrix4f &prjM_scn = objFrameData.prjM_scn;
    const int &nframeID = wData.nfData.nframeID;
    const Eigen::Matrix4f &prjM_obj = objSfM.vFrameData[nframeID].KM;
    
    const Eigen::Vector4f &p_obj = obj->objSfM.vPointData[ fp.ptID ].point;
    const Eigen::Vector4f &p_scn = fp.pt3d;
    
    //2D projection of point in novel view
    cv::Point2f &scnMsr = fp.toPt;
    Eigen::Vector4f msr_scn = prjM_scn * p_scn;
    scnMsr.x = msr_scn(0) / msr_scn(2);
    scnMsr.y = msr_scn(1) / msr_scn(2);

    //point depth
    if( isPatch )
        fp.value = msr_scn(2);
    
    //boundary check for the point in novel view
    const int sw = 60; //safe width
    cv::Rect svp( -sw, -sw, OSZ_WIDTH+2*sw, OSZ_HEIGHT+2*sw); //safe viewport
    if( !svp.contains(scnMsr) )
        return false;
    
    //2D projection of point in object patch
    cv::Point2f &objMsr = fp.fromPt;
    Eigen::Vector4f msr_obj = prjM_obj * p_obj;
    objMsr.x = msr_obj(0) / msr_obj(2);
    objMsr.y = msr_obj(1) / msr_obj(2);

    //boundary check for the point in object patch
    if( objVideo.remapPoint( objMsr, nframeID ) )
        return true;

    return false;
}

inline float calcTheta( Eigen::Vector4f v1, Eigen::Vector4f v2 )
{
    v1.normalize();
    v2.normalize();
    float cosTheta = v1.dot(v2);
    
    cosTheta = std::max( -1.0f, cosTheta );
    cosTheta = std::min( 1.0f, cosTheta );
    
    return std::acos( cosTheta ) * 180.0f / M_PI;
}

inline float calc_pWeight( float theta )
{
    const float thetaT = 60.0f;
    
    if( theta <= thetaT )
        return 1.0f;
    else if( theta <= 90.0f )
        return 1.0f - (theta-thetaT)/(90.0f-thetaT);
    else
        return 0.0f;
}

bool ivAppObjAlgoPreparePoints::prepareValidPoints
(WarpData &wData, const ivAppObject *obj, int partID, bool isPatch)
{
    const ivAppScene &scene = ivAppScene::instance();
    const ivAppSfM &objSfM = obj->objSfM;
    const ivAppObjectData &objData = obj->objData;
    const ivAppObjectFrameData &objFrameData = objData.frames.back();
    const std::vector<ivAppObjectPart> &objParts = objData.parts;
    
    //cull back points
    const std::vector<int> &ptIDs = objParts[partID].ptIDs;
    const int oriPtNum = objParts[partID].oriPtNum;
    std::vector<int> &sfm2fp = wData.sfm2fp;
    std::vector<FrontPoint> &frontPoints = wData.frontPoints;

    
    int sfmPtNum = (int)objSfM.vPointData.size();
    sfm2fp = std::vector<int>(sfmPtNum, -1);
    
    wData.oriPtNum = 0;
    
    //for each point in the cluster
    int ptNum = (int)ptIDs.size();
    for(int i=0;i<ptNum;i++)
    {
        int ptid = ptIDs[i];
        const Eigen::Vector4f &p = objSfM.vPointData[ ptid ].point;
        const Eigen::Vector4f &n = objSfM.vPointData[ ptid ].normal;
        Eigen::Vector4f gp = objFrameData.gTransform * p;
        
        //front patch point
        if( isPatch )
        {
            float theta = calcTheta( n, objFrameData.cameraPos_obj-p );
            if( theta <= 90.0f )
            {
                FrontPoint fp;
                fp.ptID = ptid;
                fp.pt3d = gp;
                fp.pWeight = calc_pWeight( theta );
                
                if( prepareFrontPoint( fp, wData, obj, isPatch ) )
                {
                    //calc diffuse intensity
                    Eigen::Vector4f L = objFrameData.lightPos_obj - p;
                    L.normalize();
                    fp.intensity = n.dot(L);
                    
                    frontPoints.push_back(fp);
                    sfm2fp[fp.ptID] = (int)frontPoints.size()-1;
                    if( i >= oriPtNum )
                        continue;
                    
                    wData.oriPtNum++;
                }
            }
        }
        
        else //front shadow point
        {
            float theta = calcTheta( n, objFrameData.lightPos_obj-p );
            if( theta <= 90.0f )
            {
                Eigen::Vector4f ground_p;
                Eigen::Vector4f v = scene.scnLight.pos - gp;
                Plane::ground.getIntersection(ground_p, scene.scnLight.pos, v);
                
                FrontPoint fp;
                fp.ptID = ptid;    
                fp.pt3d = ground_p;
                fp.value = (ground_p-gp).norm(); //blurring weight
                fp.pWeight = calc_pWeight( theta );
                
                if( prepareFrontPoint( fp, wData, obj, isPatch ) )
                {
                    frontPoints.push_back(fp);
                    sfm2fp[fp.ptID] = (int)frontPoints.size()-1;
                    if( i < oriPtNum )
                        wData.oriPtNum++;
                }
            }
        }
    }
    
    //check if points are sufficient for warping
    wData.isReadyForWarping = wData.oriPtNum >= 3;
    return wData.isReadyForWarping;
}

bool ivAppObjAlgoPreparePoints::execute(ivAppObject *obj)
{
    //cull back points
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    const int clusterNum = (int)objData.parts.size();
    for(int i=0;i<clusterNum;i++)
    {
        ObjShdWarpData &oswData = objFrameData[i];
        
        if( this->prepareValidPoints( oswData.object, obj, i, true ) )
            objFrameData.objectUsedClusterIDs.push_back(i);
        
        if( this->prepareValidPoints( oswData.shadow, obj, i, false ) )
            objFrameData.shadowUsedClusterIDs.push_back(i);
    }
    
    return true;
}


