//
//  ivAppObjAlgoSetWarps.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoSetWarps.h"


#include "ivDirParams.h"

inline void setWarps(ivAppObject *obj, bool isPatch, bool useEdge)
{
    ivAppObjectVideo &objVideo = obj->objVideo;
    ivAppObjectFrameData &objFrameData = obj->objData.frames.back();
    
    const ivAppObjParams &objParams = obj->objParams;
    const cv::Size osz = cv::Size(OSZ_WIDTH, OSZ_HEIGHT);
    std::vector< float* > pwarpWeights;
    
    CoGEPWarper &coWarper = isPatch ? objFrameData.objectCoWarper : objFrameData.shadowCoWarper;
    
    const std::vector<int> &usedClusterIDs =
        isPatch ? objFrameData.objectUsedClusterIDs : objFrameData.shadowUsedClusterIDs;
    
    //setup warping data for each visible cluster
    for( std::vector<int>::const_iterator it = usedClusterIDs.begin();
         it!=usedClusterIDs.end(); it++ )
    {
        WarpData &wData = isPatch ? objFrameData[*it].object : objFrameData[*it].shadow;
        
        // prepare warp *it
        //=========================================================================================
        //get the frame for ith cluster
        cv::Mat curFrame;
        const int &nframeID  = wData.nfData.nframeID;
        if( isPatch )
            curFrame = objVideo.objVideo.frames[nframeID];
        else
            curFrame = objVideo.objMask.frames[nframeID];
        
        //ith gradient edge preserving warper
        GEPWarper warper( curFrame, objParams.WRP_gridSize );
        
        //step1: positional constraints
        for(int j=0;j<wData.oriPtNum;j++)
            warper.addPointConstraint
            ( wData.frontPoints[j].fromPt, wData.frontPoints[j].toPt, wData.frontPoints[j].pWeight );
        
        //harder constraints for shadow ( bottom alignment constraints )
        if( !isPatch )
        {
            ShadowWarpData &swData = objFrameData[*it].shadow;
            int hcNum = (int)swData.fromPtsBottom.size();
            for(int j=0;j<hcNum;j++)
                warper.addHarderPointConstraint
                ( swData.fromPtsBottom[j], swData.toPtsBottom[j], swData.weightsBottom[j] );
        }
        
        //step2: edge contraints
        if( useEdge )
        {
            const std::vector<FrontPoint> &frontTvPoints = wData.frontTvPoints;
            const std::vector<FrontEdge> &frontEdges = wData.frontEdges;
            for(std::vector<FrontEdge>::const_iterator eid = frontEdges.begin(); eid!=frontEdges.end(); eid++ )
            {
                const cv::Point2f &fromPt0 = frontTvPoints[ eid->ftvptIDs(0) ].fromPt;
                const cv::Point2f &fromPt1 = frontTvPoints[ eid->ftvptIDs(1) ].fromPt;
                warper.addEdgeConstraint( fromPt0, fromPt1, objParams.WRP_eintrvl, eid->alpha, eid->b );
            }
        }
        //=========================================================================================

        // estimate warp weight
        //=========================================================================================
        wData.prepareWeightMap( osz );
        wData.warpWeight = 1.0f + cv::sum( wData.warpWeightMap )[0];
        pwarpWeights.push_back( &wData.warpWeight );
        //=========================================================================================
        
        coWarper.addGepWarp( warper, wData.warpWeight );
    }
    
    //normalize the weights
    float warpWeightSum = 0.0f;
    for( std::vector<float*>::iterator it = pwarpWeights.begin(); it!=pwarpWeights.end(); it++ )
        if( **it > warpWeightSum )
            warpWeightSum = **it;
    
    for(int i=0;i<(int)pwarpWeights.size();i++)
    {
        *(pwarpWeights[i]) /= warpWeightSum;
        coWarper.weightAt(i) /= warpWeightSum;
    }
}

//===========================================================================
// Warping with edges
//===========================================================================

bool ivAppObjAlgoSetObjectWarps::execute(ivAppObject *obj)
{
    setWarps( obj, true, true );
    return true;
}

bool ivAppObjAlgoSetShadowWarps::execute(ivAppObject *obj)
{
    setWarps( obj, false, true );
    return true;
}

