//
//  ivAppObjAlgoClustering.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoClustering.h"

//plane estimation
inline bool estimatePlane(ivAppObjectPart &objPart, const ivAppObject *obj)
{
    const std::vector<pointData> &vPointData = obj->objSfM.vPointData;
    std::vector<int> &ptIDs = objPart.ptIDs;
    Plane &tanPlane = objPart.tanPlane;
    
    // setup the pointcloud for plane fitting
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    cloud->width = (int)ptIDs.size();
    cloud->height = 1;
    cloud->points.resize (cloud->width * cloud->height);
    for (size_t j = 0; j < cloud->points.size(); ++j)
    {
        int ptid = ptIDs[j];
        const Eigen::Vector4f &curPoint = vPointData[ptid].point;
        cloud->points[j].x = curPoint(0);
        cloud->points[j].y = curPoint(1);
        cloud->points[j].z = curPoint(2);
    }
    
    // set up the ground
    if( !tanPlane.estimatePlane( cloud ) )
    {
        std::cout << "[X] unable to fit a plane for the cluster." << std::endl;
        return false;
    }
    Eigen::Vector3f planeNormal;
    tanPlane.getNormal(planeNormal);
    
    //correct orientation
    int votingPositive = 0;
    for( std::vector<int>::iterator it = ptIDs.begin(); it!=ptIDs.end(); it++ )
    {
        Eigen::Vector3f curNormal = vPointData[ *it ].normal.block(0,0,3,1);
        float dotValue = planeNormal.dot( curNormal );
        if( dotValue >= 0.0f )
            votingPositive++;
    }
    if( votingPositive < (int)ptIDs.size()/2 )
        tanPlane.inverseNormal();
    
    return true;
}

//find shared points
inline void expandCluster(ivAppObject *obj)
{
    const ivAppObjParams &objParams = obj->objParams;
    const float CLSTR_sharedW = objParams.CLSTR_sharedW;
    const std::vector<pointData> &vPointData = obj->objSfM.vPointData;
    
    ivAppObjectData &objData = obj->objData;
    std::vector<ivAppObjectPart> &objParts = objData.parts;
    const int clusterNum = (int)objData.parts.size();
    for(int i=0;i<clusterNum;i++) for(int j=0;j<clusterNum;j++)
    {
        if( i==j )
            continue;
        
        std::vector<int> &ptIDs_i = objParts[i].ptIDs;
        std::vector<int> &ptIDs_j = objParts[j].ptIDs;
        
        //for each point of j, if the dist to a point of i is smaller than CLSTR_sharedW, the point is shared.
        for( int jj=0; jj<objParts[j].oriPtNum; jj++ )
        for( int ii=0; ii<objParts[i].oriPtNum; ii++ )
        {
            int pii = ptIDs_i[ii];
            int pjj = ptIDs_j[jj];
            
            const Eigen::Vector4f &ptii = vPointData[pii].point;
            const Eigen::Vector4f &ptjj = vPointData[pjj].point;
            
            if( (ptii-ptjj).norm() < CLSTR_sharedW )
            {
                ptIDs_i.push_back( pjj );
                break;
            }
        }
    }
}


//k means clustering
bool ivAppObjAlgoClustering_KMeans::execute(ivAppObject *obj)
{
    //input
    //======================================================
    const ivAppObjParams &objParams = obj->objParams;
    
    //cluster parameters
    const int CLSTR_cnum    = objParams.CLSTR_cnum;
    const float CLSTR_nweight = objParams.CLSTR_nweight;
    
    // dimension number
    const int dimNum  = 6;
    
    const std::vector<pointData> &vPointData = obj->objSfM.vPointData;
    
    //number of data
    const int dataNum = (int)vPointData.size();
    //======================================================
    
    //k mean clustering
    //======================================================
    //data tobe clustered
    std::vector<double> data( dataNum * dimNum );
    std::vector<double> weights( dataNum , 1.0f );
    
    //set data properly
    for(int i=0;i<dataNum;i++)
    {
        const Eigen::Vector4f &curPt = vPointData[i].point;
        const Eigen::Vector4f &curN  = vPointData[i].normal;
        
        for(int j=0;j<3;j++)
        {
            data[i*dimNum+j] = static_cast<double>( curPt(j) );
            data[i*dimNum+3+j] = CLSTR_nweight * static_cast<double>( curN(j) );
        }
    }
    
    //clustered data
    std::vector<int> ptsClusterID(dataNum);
    std::vector<int> cluster_population( CLSTR_cnum );
    std::vector<double> cluster_energy( CLSTR_cnum );
    
    //parameters
    int seed = 12345678;
    double *cluster_center;
    int it_num;
    const int it_max = 20;
    
    cluster_center = cluster_initialize_2 ( dimNum, dataNum, CLSTR_cnum, &data[0], &seed );
    kmeans_w_03 ( dimNum, dataNum, CLSTR_cnum, it_max, it_num,
                 &data[0], &weights[0], &ptsClusterID[0], cluster_center,
                 &cluster_population[0], &cluster_energy[0] );
    //======================================================
    
    //output
    //=============================================================
    //copy clustering results
    ivAppObjectData &objData = obj->objData;

    //setup object parts
    std::vector<ivAppObjectPart> &objParts = objData.parts;
    objParts.clear();
    objParts.resize( CLSTR_cnum );
    
    for( int i=0; i<dataNum; i++ )
    {
        int clusterID = ptsClusterID[i];
        objParts[clusterID].ptIDs.push_back(i);
    }
    
    for(int i=0;i<CLSTR_cnum;i++)
        objParts[i].oriPtNum = (int)objParts[i].ptIDs.size();

    //estimate tangent plane
    for( int i=0; i<CLSTR_cnum; i++ )
        estimatePlane( objParts[i], obj );
    
    expandCluster( obj );
    //=============================================================
    
    //init the frame data (at least 1 frame)
    objData.initCurrentFrameData(true);
    
    delete [] cluster_center;

    return true;
}

//mean shift clustering
bool ivAppObjAlgoClustering_MeanShift::execute(ivAppObject *obj)
{
    //input
    //======================================================
    //normal estimation
    const ivAppObjParams &objParams = obj->objParams;
    
    //cluster parameters
    const int k_neigh = objParams.NRM_nnum;
    const float CLSTR_width = objParams.CLSTR_width;
    const float CLSTR_nweight = objParams.CLSTR_nweight;
    
    // dimension number
    const int dimNum  = 6;
    
    const std::vector<pointData> &vPointData = obj->objSfM.vPointData;
    
    //number of data
    const int dataNum = (int)vPointData.size();
    //======================================================

    //mean shift clustering
    //======================================================
    //input points - data tobe clustered
    Eigen::Matrix<float,Eigen::Dynamic,Eigen::Dynamic> inputPoints;
    inputPoints.resize( dataNum, dimNum );
    for(int i=0;i<dataNum;i++)
    {
        const Eigen::Vector4f &curPt = vPointData[i].point;
        const Eigen::Vector4f &curN  = vPointData[i].normal;
        
        for(int j=0;j<3;j++)
        {
            inputPoints(i,j) = curPt(j);
            inputPoints(i,j+3) = CLSTR_nweight * curN(j);
        }
    }
    
    Eigen::VectorXi ptClusterIDs;
    MeanShiftCluster::cluster( ptClusterIDs, CLSTR_width, k_neigh, inputPoints );
    int clusterNum = 1 + ptClusterIDs.maxCoeff();
    //======================================================

    //output
    //=============================================================
    ///copy clustering results
    ivAppObjectData &objData = obj->objData;

    //setup object parts
    std::vector<ivAppObjectPart> &objParts = objData.parts;
    objParts.clear();
    objParts.resize( clusterNum );
    
    for( int i=0; i<dataNum; i++ )
    {
        int clusterID = ptClusterIDs(i);
        if( clusterID < 0 )//outlier elimination
            continue;
        objParts[clusterID].ptIDs.push_back(i);
        objParts[clusterID].oriPtNum++;
    }
    
    for(int i=0;i<clusterNum;i++)
        objParts[i].oriPtNum = (int)objParts[i].ptIDs.size();
    
    //estimate tangent plane
    for( int i=0; i<clusterNum; i++ )
        estimatePlane( objParts[i], obj );
    
    expandCluster( obj );
    //=============================================================

    //init the frame data (at least 1 frame)
    objData.initCurrentFrameData(true);

    return true;
}