//
//  ivAppObjAlgoNormalEstimation.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoNormalEstimation.h"

void ivAppObjAlgoNormalEstimation::corretPointNormal(ivAppObject *obj)
{
    ivAppSfM &objSfM = obj->objSfM;
    const std::vector<frameData> &vFrameData = objSfM.vFrameData;
    std::vector<pointData> &vPointData = objSfM.vPointData;
    
    ivAppTvEdges &objTvEdges = obj->objTvEdges;
    tvTriPrimSet &vtvPrim = objTvEdges.vtvPrim;
    tvPoint3DSet &vtvPoint = objTvEdges.vtvPoint;
    
    //for each primitive
    for(std::vector<tvTriPrim>::iterator pit = vtvPrim.begin(); pit!=vtvPrim.end(); pit++ )
    {
        const std::vector<int> &polygon = pit->polygon;
        
        //edges are ignored
        int npts = (int)polygon.size();
        if( npts<=2 )
            continue;
        
        //find the points lying in the primitive
        int ptNum = (int)vPointData.size();
        std::vector<bool> inPrimPoints(vPointData.size(), true);
        for(int f=0;f<objSfM.vFrameData.size();f++)
        {
            cv::Mat canvas = cv::Mat::zeros(objSfM.height, objSfM.width, CV_8UC1);
            
            //if the polygon is visible
            if( pit->isFront(f) )
            {
                std::vector<cv::Point> pts;
                for(int i=0;i<npts;i++)
                {
                    Eigen::Vector2f &lblpt2d = vtvPoint[polygon[i]].tvv2D[f];
                    cv::Point p( lblpt2d(0), lblpt2d(1) );
                    pts.push_back( p );
                }
                const cv::Point* elementPoints[1] = { &pts[0] };
                cv::fillPoly( canvas, elementPoints, &npts, 1, cv::Scalar::all(255) );
            }
            
            int vptNum = (int)vFrameData[f].visiblePIDs.size();
            for(int i=0;i<vptNum;i++)
            {
                const Eigen::Vector2f &p = vFrameData[f].visibleMsr2ds[i];
                uchar &c = canvas.at<uchar>(p(1),p(0));
                if( c==0 )
                {
                    int ptID = vFrameData[f].visiblePIDs[i];
                    inPrimPoints[ptID] = false;
                }
            }
        }
        
        //calc average primitive normal (the triangle order is corrected when labeling)
        Eigen::Vector3f primNormal;
        const std::vector<Eigen::Vector3i> &triangles = pit->triangles;
        for(std::vector<Eigen::Vector3i>::const_iterator tit = triangles.begin(); tit!=triangles.end(); tit++ )
        {
            Eigen::Vector4f v1_4 = vtvPoint[(*tit)(1)].v3D - vtvPoint[(*tit)(0)].v3D;
            Eigen::Vector4f v2_4 = vtvPoint[(*tit)(2)].v3D - vtvPoint[(*tit)(0)].v3D;
            Eigen::Vector3f v1_3 = v1_4.block(0,0,3,1);
            Eigen::Vector3f v2_3 = v2_4.block(0,0,3,1);
            
            primNormal += v1_3.cross( v2_3 );
        }
        primNormal.normalize();
        
        //correct point normal
        for(int i=0;i<ptNum;i++)
        {
            if( !inPrimPoints[i] )
                continue;
            
            vPointData[i].normal.block(0,0,3,1) = primNormal;
            vPointData[i].normal(3) = 0.0f;
        }
    }
}

bool ivAppObjAlgoNormalEstimation::execute(ivAppObject *obj)
{
    //normal estimation
    ivAppSfM &objSfM = obj->objSfM;
    const ivAppObjParams &objParams = obj->objParams;
    objSfM.estimateNormal( objParams.NRM_nnum );
    
    //correct normal by primitives
    if( objParams.NRM_label )
        this->corretPointNormal(obj);
    
    return true;
}


