//
//  ivAppObjAlgoClusterNearestFrames.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoClusterNearestFrames.h"

void ivAppObjAlgoClusterNearestFrame::calcOriDiff
( std::vector<float> &oriDiffs, const std::vector<int> &cluster, const Eigen::Vector4f &c_obj_target, const ivAppSfM &objSfM )
{
    int ptNum = (int)cluster.size();
    
    int fNum = (int)objSfM.vFrameData.size();
    oriDiffs = std::vector<float>(fNum);//in degrees
    for( int k=0; k<fNum; k++ )
    {
        float error = 0.0f;
        
        const Eigen::Vector4f &c_obj = objSfM.vFrameData[k].camPosition;
        for( int j=0; j<ptNum; j++ )
        {
            const Eigen::Vector4f &p = objSfM.vPointData[ cluster[j] ].point;
            
            Eigen::Vector4f v_obj_target = c_obj_target - p;
            v_obj_target.normalize();
            
            Eigen::Vector4f v_obj = c_obj - p;
            v_obj.normalize();
            
            float dotValue = v_obj_target.dot( v_obj );
            dotValue = std::max( dotValue, -1.0f );
            dotValue = std::min( dotValue, 1.0f );
            error += std::acos( dotValue ) * 180.0f / M_PI;
        }
        
        oriDiffs[k] = error/ptNum;
    }
}

float ivAppObjAlgoClusterNearestFrame::FindNearest
( int &nframeID, const std::vector<int> &cluster, const Eigen::Vector4f &c_obj_target, const ivAppSfM &objSfM )
{
    //calc the orientation difference between c_obj_target and source view
    std::vector<float> oriDiffs;//in degrees
    ivAppObjAlgoClusterNearestFrame::calcOriDiff( oriDiffs, cluster, c_obj_target, objSfM );
    
    //find the min
    std::vector<float>::iterator minit = std::min_element( oriDiffs.begin(), oriDiffs.end() );
    nframeID = (int)std::distance(oriDiffs.begin(), minit);
    
    return *minit;
}

bool ivAppObjAlgoClusterNearestFrame::execute(ivAppObject *obj)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    
    int clusterNum = (int)objData.parts.size();
    for(int i=0;i<clusterNum;i++)
    {
        ObjShdWarpData &oswData = objFrameData[i];
        NearestFrameData &pnfd = oswData.object.nfData;
        NearestFrameData &snfd = oswData.shadow.nfData;
        
        ivAppObjectPart &objPart = objData.parts[i];
        pnfd.oDiff = this->FindNearest( pnfd.nframeID, objPart.ptIDs, objFrameData.cameraPos_obj, obj->objSfM );
        snfd.oDiff = this->FindNearest( snfd.nframeID, objPart.ptIDs,  objFrameData.lightPos_obj, obj->objSfM );
        
        pnfd.isSafe = pnfd.oDiff < obj->objParams.OSG_oriDiff;
        snfd.isSafe = snfd.oDiff < obj->objParams.OSG_oriDiff;
    }
    
    return true;
}

bool ivAppObjAlgoClusterNearestFrames::execute(ivAppObject *obj)
{
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    ivAppGuiParams *guiParams = [ivAppGuiParams instance];
    
    //cache the original animation pose
    int oriScnPos = guiParams.scnPos;
    int oriObjPos = guiParams.objPos;
    int oriAniPos = guiParams.aniPos;
    
    const int aniPosNum = guiParams.aniMaxPos+1;
    const int clusterNum = (int)objData.parts.size();
    
    //initialize space for selected frames
    objData.objNFrames.resize( clusterNum, aniPosNum );
    objData.shdNFrames.resize( clusterNum, aniPosNum );
    
    //setup procedure
    ivAppObjAlgoSetup oaSetup;
    
    //for each motion pose
    ivRenderOrder &ivRO = ivRenderOrder::instance();
    for(int i=0;i<aniPosNum;i++)
    {
        //goto pose i
        guiParams.aniPos = i;
        ivRO.setAnimationPose();
        
        //setup
        oaSetup.execute(obj);
        
        //find cluster nearest frame
        ivAppObjAlgoClusterNearestFrame::execute(obj);
        
        //copy curFE to nframeData
        for(int j=0;j<clusterNum;j++)
        {
            objData.objNFrames(j,i) = objFrameData[j].object.nfData;
            objData.shdNFrames(j,i) = objFrameData[j].shadow.nfData;
        }
    }
    
    //recover original pose
    guiParams.scnPos = oriScnPos;
    guiParams.objPos = oriObjPos;
    guiParams.aniPos = oriAniPos;
    
    return true;
}
