//
//  ivAppObjAlgoFrontGraphCutFrames.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjAlgoFrontGraphCutFrames.h"

#define FS_CG_Precision 1000.0f

inline bool isFrontCluster
(const Eigen::Vector4f &c_obj_target, const ivAppObjectPart &objPart, const ivAppSfM &objSfM)
{
    const std::vector<int> &ptIDs = objPart.ptIDs;
    for( int i=0; i<objPart.oriPtNum; i++ )
    {
        int ptid = ptIDs[i];
        const Eigen::Vector4f &p = objSfM.vPointData[ ptid ].point;
        const Eigen::Vector4f &n = objSfM.vPointData[ ptid ].normal;
        
        Eigen::Vector4f v_obj_target = c_obj_target - p;
        v_obj_target.normalize();
        
        if( v_obj_target.dot(n) > 0.0f )
            return true;
    }
    
    return false;
}

void ivAppObjAlgoFrontGraphCutFrame::selectFramesForObjShd
(bool isPatch, ivAppObject *obj)
{
    const ivAppObjParams &objParams = obj->objParams;
    
    //number of nodes in the graph to be cut
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    const Eigen::Vector4f &targetPos_obj = isPatch ? objFrameData.cameraPos_obj : objFrameData.lightPos_obj;
    
    std::vector<ivAppObjectPart> &objParts = objData.parts;
    int clusterNum = (int)objParts.size();
    int clusterNodeNum = clusterNum + 1;
    
    //number of labels
    const ivAppSfM &objSfM = obj->objSfM;
    int fNum = (int)objSfM.vFrameData.size();
    
    //data penalty
    std::vector<int> dataPenalty( clusterNodeNum*fNum, 0 );
    std::vector<float> oDiff( clusterNodeNum*fNum, 0.0f );
    std::vector<bool> isFronts( clusterNum );
    for(int k=0;k<clusterNodeNum;k++)
    {
        std::vector<float> oriDiffs;
        if( k == clusterNodeNum-1 ) //representative that enforces spatial coherence
            oriDiffs = std::vector<float>( fNum, 0.0f );
        else //cluster
        {
            isFronts[k] = isFrontCluster( targetPos_obj, objParts[k], objSfM );
            ivAppObjAlgoClusterNearestFrames::calcOriDiff( oriDiffs, objParts[k].ptIDs, targetPos_obj, objSfM );
        }
        
        for(int j=0;j<fNum;j++)
        {
            dataPenalty[k*fNum+j] = (int)( FS_CG_Precision * oriDiffs[j] );
            oDiff[k*fNum+j] = oriDiffs[j];
        }
    }
    
    //smoothness penalty
    const ivAppObjectSfMSimGraph &objSimGraph = obj->objSimGraph;
    std::vector<int> smoothPenalty( fNum * fNum, 0 );
    for(int i=0;i<fNum;i++)for(int j=0;j<fNum;j++)
        smoothPenalty[i*fNum+j] = (int)( FS_CG_Precision * objSimGraph(i,j) * objParams.FS_simW );
    
    //frame selection by graph cut optimization(labeling)
    try{
        GCoptimizationGeneralGraph *gc = new GCoptimizationGeneralGraph( clusterNodeNum, fNum );
        gc->setDataCost( &(dataPenalty[0]) );
        gc->setSmoothCost( &(smoothPenalty[0]) );
        
        //spatial edges of nodes
        for(int k=0;k<clusterNodeNum-1;k++)
        {
            if( !isFronts[k] )
                continue;
            
            gc->setNeighbors( k, clusterNodeNum-1 );
        }
        
        //go optimization
        gc->swap( -1 );
        
        //copy results (ignore the representative)
        for ( int i=0; i<clusterNodeNum-1; i++ )
        {
            int nframeID = gc->whatLabel(i);
            
            NearestFrameData &nfData = isPatch ? objFrameData[i].object.nfData : objFrameData[i].shadow.nfData;
            nfData.nframeID = nframeID;
            nfData.oDiff = oDiff[i*fNum+nframeID];
            nfData.isSafe = nfData.oDiff < objParams.OSG_oriDiff;
        }
        
        delete gc;
    }
    catch (GCException e)
    {
        e.Report();
    }
}

bool ivAppObjAlgoFrontGraphCutFrame::execute(ivAppObject *obj)
{
    this->selectFramesForObjShd( true,  obj );
    this->selectFramesForObjShd( false, obj );
    
    return true;
}





void ivAppObjAlgoFrontGraphCutFrames::selectFramesForObjShd
(bool isPatch, ivAppObject *obj)
{
    const ivAppObjParams &objParams = obj->objParams;
    
    ivAppObjectData &objData = obj->objData;
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    std::vector<ivAppObjectPart> &objParts = objData.parts;
    int clusterNum = (int)objParts.size();
    int clusterNodeNum = clusterNum + 1;
    
    //number of labels
    const ivAppSfM &objSfM = obj->objSfM;
    int fNum = (int)objSfM.vFrameData.size();
    
    const ivAppGuiParams *guiParams = [ivAppGuiParams instance];
    int aniPosNum = guiParams.aniMaxPos+1;
    
    //initialize space for selected frames
    objData.objNFrames.resize( clusterNum, aniPosNum );
    objData.shdNFrames.resize( clusterNum, aniPosNum );
    
    //the node number in the graph to be cut
    int nodeNum = aniPosNum * clusterNodeNum;
    
    ivAppObjAlgoSetup oaSetup;
    ivRenderOrder &ivRO = ivRenderOrder::instance();
    
    //data penalty
    std::vector<int> dataPenalty( nodeNum*fNum, 0 );
    std::vector<float> oDiff( nodeNum*fNum, 0.0f );
    Eigen::Matrix<bool, Eigen::Dynamic, Eigen::Dynamic> isFronts( aniPosNum, clusterNum ); //cluster visibility
    for(int i=0;i<aniPosNum;i++)
    {
        guiParams.aniPos = i;
        ivRO.setAnimationPose();
        
        //setup the motion pose parameters
        oaSetup.execute(obj);

        const Eigen::Vector4f &targetPos_obj = isPatch ? objFrameData.cameraPos_obj : objFrameData.lightPos_obj;
        for(int k=0;k<clusterNodeNum;k++)
        {
            std::vector<float> oriDiffs;
            if( k==clusterNodeNum-1 ) //representative that enforce spatial coherence
                oriDiffs = std::vector<float>( fNum, 0.0f );
            else
            {
                isFronts( i, k ) = isFrontCluster( targetPos_obj, objParts[k], objSfM );
                ivAppObjAlgoClusterNearestFrame::calcOriDiff( oriDiffs, objParts[k].ptIDs, targetPos_obj, objSfM );
            }
            
            //cluster k in object motion pose i
            int kNode_i = i*clusterNodeNum + k;
            for(int j=0;j<fNum;j++)
            {
                dataPenalty[kNode_i*fNum+j] = (int)( FS_CG_Precision * oriDiffs[j] );
                oDiff[kNode_i*fNum+j] = oriDiffs[j];
            }
        }
    }
    
    //smoothness penalty
    const ivAppObjectSfMSimGraph &objSimGraph = obj->objSimGraph;
    std::vector<int> smoothPenalty( fNum * fNum, 0 );
    for(int i=0;i<fNum;i++)for(int j=0;j<fNum;j++)
        smoothPenalty[i*fNum+j] = (int)( FS_CG_Precision * objSimGraph(i,j) * objParams.FS_simW );
    
    try{
        GCoptimizationGeneralGraph *gc = new GCoptimizationGeneralGraph( nodeNum, fNum );
        gc->setDataCost( &(dataPenalty[0]) );
        gc->setSmoothCost( &(smoothPenalty[0]) );
        
        //temporal edges of nodes
        for(int i=0;i<aniPosNum-1;i++)
            for(int k=0;k<clusterNodeNum-1;k++)
            {
                //cluster k in object motion pose i, j
                int kNode_i = i*clusterNodeNum + k;
                int kNode_j = (i+1)*clusterNodeNum + k;
                if( !isFronts( i, k ) )
                    continue;
                
                gc->setNeighbors( kNode_i, kNode_j );
            }
        
        //spatial edges of nodes
        for(int i=0;i<aniPosNum;i++)
            for(int k=0;k<clusterNodeNum-1;k++)
            {
                //cluster k, c in object motion pose i
                int kNode_i = i*clusterNodeNum + k;
                int cNode_i = i*clusterNodeNum + clusterNodeNum-1;
                if( !isFronts( i, k ) )
                    continue;

                gc->setNeighbors( kNode_i, cNode_i );
            }
        
        //go optimization
        gc->swap( -1 );
        
        //copy results
        for ( int i=0; i<nodeNum; i++ )
        {
            int clusterID = i % clusterNodeNum;
            int motionID = i / clusterNodeNum;
            int nframeID = gc->whatLabel(i);
            
            //ignore the representative
            if( clusterID == clusterNodeNum-1 )
                continue;
            
            NearestFrameData &nfData = isPatch ? objData.objNFrames(clusterID, motionID):
            objData.shdNFrames(clusterID, motionID);
            
            nfData.nframeID = nframeID;
            nfData.oDiff = oDiff[i*fNum+nframeID];
            nfData.isSafe = nfData.oDiff < objParams.OSG_oriDiff;
        }
        
        delete gc;
    }
    catch (GCException e){
        e.Report();
    }
}

bool ivAppObjAlgoFrontGraphCutFrames::execute(ivAppObject *obj)
{
    const ivAppGuiParams *guiParams = [ivAppGuiParams instance];
    
    //no animation
    int aniPosNum = guiParams.aniMaxPos+1;
    if( aniPosNum <= 0 )
        return false;
    
    //cache the original animation pose
    int oriScnPos = guiParams.scnPos;
    int oriObjPos = guiParams.objPos;
    int oriAniPos = guiParams.aniPos;
    
    this->selectFramesForObjShd( true, obj );
    this->selectFramesForObjShd( false, obj );
    
    //recover original pose
    guiParams.scnPos = oriScnPos;
    guiParams.aniPos = oriObjPos;
    guiParams.aniPos = oriAniPos;
    
    return true;
}