//
//  ivAppObjAnimation.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "Utils_Ogl.h"
#include "ivAppObjAnimation.h"

//estimate control points
//input: keyPoses
//output: controlPoints
void ivAppObjAnimation::estimateBezierControls()
{
    this->controlPoints.clear();
    
    int keyNum = (int)keyPoses.size();
    if( keyNum < 2 )
    {
        for(int i=0;i<keyNum;i++)
            controlPoints.push_back(keyPoses[i]);
        return;
    }
    
    // the first pose
    {
        ivAppObjPose &p1 = keyPoses[0];
        ivAppObjPose &p2 = keyPoses[1];
        ivAppObjPose tangent = (p2 - p1).normalized();
        ivAppObjPose q1 = p1 + tangent * controlScl;
        controlPoints.push_back(p1);
        controlPoints.push_back(q1);
    }
    
    //pose in between
    for(int i=1;i<keyNum-1;i++)
    {
        ivAppObjPose &p0 = keyPoses[i - 1];
        ivAppObjPose &p1 = keyPoses[i];
        ivAppObjPose &p2 = keyPoses[i + 1];
        ivAppObjPose tangent = (p2 - p0).normalized();
        float mag0 = (p1 - p0).magnitude();
        float mag1 = (p2 - p1).magnitude();
        ivAppObjPose q0 = p1 - tangent * controlScl * mag0;
        ivAppObjPose q1 = p1 + tangent * controlScl * mag1;
        
        controlPoints.push_back(q0);
        controlPoints.push_back(p1);
        controlPoints.push_back(q1);
    }
    
    //the last pose
    {
        ivAppObjPose &p0 = keyPoses[keyNum - 2];
        ivAppObjPose &p1 = keyPoses[keyNum - 1];
        ivAppObjPose tangent = (p1 - p0).normalized();
        ivAppObjPose q0 = p1 - tangent * controlScl;
        controlPoints.push_back(q0);
        controlPoints.push_back(p1);
    }
}

//draw a bezier curve
//input: starting index of controlPoints
//output: push_back a path in vM
void ivAppObjAnimation::bezierCurveCubic(int key)
{
    int intervalNum3 = intervalNum * 3;
    ivAppObjPose tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;
    for(int t=0;t<intervalNum3;t++)
    {
        float alpha = 1.0f - (float)t/intervalNum3;
        
        tmp1 = controlPoints[key+0].interpolate( alpha, controlPoints[key+1] );
        tmp2 = controlPoints[key+1].interpolate( alpha, controlPoints[key+2] );
        tmp3 = controlPoints[key+2].interpolate( alpha, controlPoints[key+3] );
        
        tmp4 = tmp1.interpolate( alpha, tmp2 );
        tmp5 = tmp2.interpolate( alpha, tmp3 );
        
        tmp6 = tmp4.interpolate( alpha, tmp5 );
        
        this->vM.push_back( tmp6 );
    }
}

//reset
void ivAppObjAnimation::reset()
{
    this->keyPoses.clear();
    this->vM.clear();
    this->controlPoints.clear();
    objPose.reset();
}

//push & pop key poses
void ivAppObjAnimation::push()
{
    this->keyPoses.push_back(objPose);
}
void ivAppObjAnimation::pop()
{
    if( !this->keyPoses.empty() )
        this->keyPoses.pop_back();
}

//set current object motion
void ivAppObjAnimation::gotoMotion(int i)
{
    if( i<0 )
        return;
    
    int motionNum = this->getMotionNum();
    if( i>=motionNum )
        return;

    if( motionNum==0 )
        return;
    
    objPose = vM[i];
}
Eigen::Matrix4f ivAppObjAnimation::getMotion(int i)
{
    int motionNum = this->getMotionNum();
    if( motionNum == 0 )
        return objPose.gTransform();
    else
    {
        i = std::min( motionNum-1, i );
        i = std::max( 0, i );
        
        return vM[i].gTransform();
    }
}
int ivAppObjAnimation::getMotionNum()
{
    return (int)vM.size();
}

//set source object motion
void ivAppObjAnimation::setObjMotion(ivAppSfM &objSfM)
{
    int fNum = (int)objSfM.vFrameData.size();
    this->vM.resize(fNum);
    
    Eigen::Matrix4f usrRt = objPose.rtTransform();
    Eigen::Matrix4f refRt = objSfM.vFrameData[0].M.inverse();
    for(int i=0;i<fNum;i++)
    {
        Eigen::Matrix4f newRt = usrRt * refRt * objSfM.vFrameData[i].M;
        
        Eigen::Matrix3f newR = newRt.block(0,0,3,3);
        Eigen::Quaternionf q(newR);
        this->vM[i]._rot = q;
        this->vM[i]._trs = newRt.block(0,3,3,1);
        this->vM[i]._scl = objPose._scl;
    }
}

//set bezier path params
void ivAppObjAnimation::setAniParams(int intervalNum, float controlScl)
{
    this->intervalNum = intervalNum;
    this->controlScl = controlScl;
}

//1. estimateBezierControls
//2. bezierPathCubic
//input: keyPoses
//output: controlPoints
void ivAppObjAnimation::estimateObjAnimation()
{
    //estimate control points
    this->estimateBezierControls();
    this->vM.clear();
    
    int cpNum = (int)controlPoints.size();
    if( cpNum == 0 )
        return ;
    
    //calc bezier curves
    for(int i=0;i<cpNum;i+=3)
        if(i+3<cpNum)
            bezierCurveCubic( i );
    
    //make up the last point
    int keyNum = (int)keyPoses.size();
    this->vM.push_back( keyPoses[keyNum-1] );
}

//render the animation path
void ivAppObjAnimation::renderObjAnimationPath()
{
    int keyPosNum = (int)keyPoses.size();
    int posNum = (int)vM.size();
    
    //key poses
    glPushAttrib( GL_LINE_BIT );
    glLineWidth(4);
    for(int i=0;i<keyPosNum;i++)
    {
        glPushMatrix();
        glMultMatrixf(keyPoses[i].gTransform().data());
        glScalef( 0.5f, 0.5f, 0.5f );
        Utils_Ogl::drawAxis();
        glPopMatrix();
    }
    glPopAttrib();
    
    //get transform matrix
    glPushAttrib( GL_POINT_BIT );
    glPointSize( 1 );
    for(int i=0;i<posNum;i++)
    {
        glPushMatrix();
        glMultMatrixf(vM[i].gTransform().data());
        glScalef( 0.2f, 0.2f, 0.2f );
        Utils_Ogl::drawAxis();
        glPopMatrix();
    }
    glPopAttrib();
    
    //render path
    glColor3ub(255,255,255);
    glBegin(GL_LINES);
    for(int i=0;i<posNum-1;i++)
    {
        int j = i+1;
        glVertex4fv(vM[i].gTransform().block(0,3,4,1).data());
        glVertex4fv(vM[j].gTransform().block(0,3,4,1).data());
    }
    glEnd();
}

//read the params
std::istream& operator>>(std::istream& is, ivAppObjAnimation& objani)
{
    std::string trash;
    is >> trash;
    
    int dataNum;
    std::string type;
    is >> trash >> type >> trash >> dataNum;
    
    //key poses added by user
    if( type=="KEY" )
    {
        is >> trash >> objani.intervalNum;
        is >> trash >> objani.controlScl;
        
        objani.keyPoses.resize(dataNum);
        for(int i=0;i<dataNum;i++)
            is >> objani.keyPoses[i];
        
        objani.setAniParams(objani.intervalNum, objani.controlScl);
        objani.estimateObjAnimation();
    }
    //object motion
    else if( type=="OBJ" )
    {
        objani.vM.resize(dataNum);
        for(int i=0;i<dataNum;i++)
            is >> objani.vM[i];
    }
    
    //current object pose
    is >> trash;
    is >> objani.objPose;

    return is;
}

//export the params
std::ostream& operator<<(std::ostream& os, const ivAppObjAnimation& objani)
{
    os << "OBJ_ANIMATION" << std::endl;
    
    //key poses added by user
    if( !objani.keyPoses.empty() )
    {
        os << "\tTYPE: KEY" << std::endl;
        int kpNum = (int)objani.keyPoses.size();
        os << "\tKEYNUM: " << kpNum << std::endl;
        
        os << "\tintervalNum: " << objani.intervalNum << std::endl;
        os << " \tcontrolScl: " << objani.controlScl  << std::endl;
        
        for(int i=0;i<kpNum;i++)
            os << objani.keyPoses[i];
    }
    //object motion
    else
    {
        os << "\tTYPE: OBJ" << std::endl;
        int vmNum = (int)objani.vM.size();
        os << "\tVMNUM: " << vmNum << std::endl;
        for(int i=0;i<vmNum;i++)
            os << objani.vM[i];
    }

    //current object pose
    os << "\tCURRENTPOS: " << objani.objPose;
    
    return os;
}







