//
//  ivAppObject.mm
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObject.h"

//read the object data
void ivAppObject::init( std::string tcName )
{
    this->tcName = tcName;

    //object parameters and animation
    this->reset();
    
    //read video data
    this->objVideo.init(tcName);
    
    //read sfm data
    std::string objsfmPath = ivDirParams::getObjSfMPath(tcName);
    objSfM.readObjSfM( objsfmPath );

    //read tvedges
    std::string tvmPath = ivDirParams::getTvMPath(tcName);
    this->objTvEdges.readTVModel( tvmPath, &objSfM );
    
    this->isReady = true;
}

//reset status
void ivAppObject::reset()
{
    objParams.reset();
    objAnimation.reset();
    mGuidePose.reset();
    cameraDensity = 10.0f;
}

//render proxy
void ivAppObject::renderProxy
(bool isActive, bool OBJ_GUIDE, bool OBJ_NCAMERA,
 float objID, bool OBJ_CLSTR, bool OBJ_PROXY, bool SHD_PROXY )
{
    if( !isReady )
        return;
    
    // render current frame data (point, normal, edge)
    //=========================================================================
    objData.renderCurrentFrameData( objSfM, objID, OBJ_CLSTR, OBJ_PROXY, SHD_PROXY );
    //=========================================================================
    
    // render manipulation guidance for current object
    //=========================================================================
    int clusterNum = (int)objData.parts.size();
    ivAppObjectFrameData &objFrameData = objData.frames.back();
    
    glPushMatrix();
    glMultMatrixf( objFrameData.gTransform.data() );
    
    if( OBJ_PROXY )
    {
        //render the axis
        glPushAttrib(GL_LINE_BIT);
        glLineWidth(6);
        glPushMatrix();
        if( !isActive )
            glScalef(0.5f,0.5f,0.5f);
        Utils_Ogl::drawAxis();
        glPopMatrix();
        glPopAttrib();
    }
    
    glPushMatrix();
    Eigen::Matrix4f mGuideM = mGuidePose.gTransform();
    glMultMatrixf( mGuideM.data() );
    
    //render the manipulation guidance
    //===========================================================
    if( OBJ_GUIDE && isActive )
    {
        glColor4ub(128,128,128, 255);
        int fNum = (int)objSfM.vFrameData.size();
        for(int i=0;i<fNum;i++)
            if( i%(int)cameraDensity==0 )
                objSfM.renderCameraPyramid( 0.03f, i, true );
        
        glPushAttrib(GL_LINE_BIT);
        glColor4ub(0,0,0, 255);
        glLineWidth(4);
        glBegin(GL_LINE_STRIP);
        for(int i=0;i<fNum;i++)
            glVertex3fv( objSfM.vFrameData[i].unitCam.data() );
        glEnd();
        glPopAttrib();
    }
    //===========================================================
    
    //the selected frames
    for(int i=0;i<clusterNum&&OBJ_GUIDE&&isActive;i++)
    {
        int aniPoseNum = (int)objData.objNFrames.cols();
        
        //selected frames for object motion
        for(int j=0;j<aniPoseNum;j++)
        {
            glColor4ub( 255, 255, 255, 255 );
            if( !objData.ifObjectMotionFramesValid[j] )
                glColor4ub( 192, 0, 0, 255 );

            int nframeID = objData.objNFrames(i,j).nframeID;
            objSfM.renderCameraSphere(0.025f, nframeID, true);
        }
        
        //draw current frame for object using box
        if(OBJ_NCAMERA)
        {
            glColor4ub( 255, 255, 255, 255 );
            if( !objFrameData.ifObjectPoseFramesValid )
                glColor4ub( 192, 0, 0, 255 );
            
            int nframeID = objFrameData[i].object.nfData.nframeID;
            objSfM.renderCameraPyramid( 0.06f, nframeID, true );
            objSfM.renderCameraSphere(0.025f, nframeID, true);
        }
        
        //selected frames for shadow motion
        for(int j=0;j<aniPoseNum;j++)
        {
            glColor4ub( 255, 255, 0, 255 );
            if( !objData.ifShadowMotionFramesValid[j] )
                glColor4ub( 192, 0, 0, 255 );
            
            int nframeID = objData.shdNFrames(i,j).nframeID;
            objSfM.renderCameraSphere(0.02f, nframeID, true);
        }
        
        //draw current frame for shadow using box
        if(OBJ_NCAMERA)
        {
            glColor4ub( 255, 255, 0, 255 );
            if( !objFrameData.ifShadowPoseFramesValid )
                glColor4ub( 192, 0, 0, 255 );
            
            int nframeID = objFrameData[i].shadow.nfData.nframeID;
            objSfM.renderCameraPyramid( 0.05f, nframeID, true );
            objSfM.renderCameraSphere(0.02f, nframeID, true);
        }
    }
    
    glPopMatrix();
    
    glPopMatrix();
    //=========================================================================
}

//io
std::istream& operator>>(std::istream& is, ivAppObject& obj)
{
    is >> obj.tcName;
    obj.init(obj.tcName);
    
    is >> obj.objParams;
    
    is >> obj.objAnimation;
    
    return is;
}
std::ostream& operator<<(std::ostream& os, const ivAppObject& obj)
{
    os << obj.tcName << std::endl;
    os << obj.objParams;
    os << obj.objAnimation;
    
    return os;
}












