//
//  ivAppObjectData.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjectData.h"

void FrontEdge::calcEdgeParams(const std::vector<FrontPoint> &frontTvPoints)
{
    const FrontPoint &fp0 = frontTvPoints[ ftvptIDs(0) ];
    const FrontPoint &fp1 = frontTvPoints[ ftvptIDs(1) ];
    
    cv::Point2f edge_v = fp1.toPt - fp0.toPt;
    alpha = std::atan2( edge_v.y, edge_v.x );
    b = std::sin(alpha) * fp1.toPt.x - std::cos(alpha) * fp1.toPt.y;
}

void WarpData::resetData()
{
    //release the object points and edges
    this->sfm2fp.clear();
    this->frontPoints.clear();
    
    this->tvp2ftvp.clear();
    this->frontEdges.clear();
    this->frontTvPoints.clear();
}

void WarpData::renderPoint(bool coloredByAttribute) const
{
    Utils_Ocv &uocv = Utils_Ocv::instance();
    
    //render front points
    glBegin(GL_POINTS);
    
    for( std::vector<FrontPoint>::const_iterator it = frontPoints.begin();
        it!=frontPoints.end(); it++ )
    {
        if( coloredByAttribute )
        {
            cv::Vec3b color;
            uocv.getColorRGB( color, it->value * 255.0f, 255, 0 );
            glColor3ub( color[0], color[1], color[2] );
        }
        
        glVertex4fv( it->pt3d.data() );
    }
    
    glEnd();
}

void WarpData::renderPointNormal(const ivAppSfM &objSfM) const
{
    glBegin(GL_LINES);
    for( std::vector<FrontPoint>::const_iterator it = frontPoints.begin();
        it!=frontPoints.end(); it++ )
    {
        const pointData &curPD = objSfM.vPointData[ it->ptID ];
        Eigen::Vector4f npt = curPD.point + 0.05f * curPD.normal;
        
        glVertex4fv( curPD.point.data() );
        glVertex4fv( npt.data() );
    }
    glEnd();
}

void WarpData::renderEdge() const
{
    glPushAttrib(GL_LINE_BIT);
    glLineWidth(4.0f);
    
    //render front edges
    glBegin(GL_LINES);
    for( std::vector<FrontEdge>::const_iterator feit = frontEdges.begin();
        feit != frontEdges.end(); feit++ )
    {
        for(int i=0;i<2;i++)
        {
            const FrontPoint &fp = frontTvPoints[ feit->ftvptIDs(i) ];
            glVertex4fv( fp.pt3d.data() );
        }
    }
    glEnd();
    
    glPopAttrib();
}

//prepare warpWeightMap
void WarpData::prepareWeightMap(cv::Size osz)
{
    this->warpWeightMap = cv::Mat( osz, CV_32FC1, cv::Scalar(0.0f) );
    
    Utils_Ocv &uocv = Utils_Ocv::instance();
    
    //step1: point importance
    for(std::vector<FrontPoint>::iterator pit = this->frontPoints.begin();
        pit != this->frontPoints.end(); pit++ )
    {
        uocv.drawImpWeight( this->warpWeightMap, pit->toPt, pit->pWeight * PTKERNELSZ );
    }
    
    //step2: edge importance
    cv::Mat edgeMap = cv::Mat( osz, CV_8UC1, cv::Scalar(0) );
    for(std::vector<FrontEdge>::const_iterator eid = frontEdges.begin(); eid!=frontEdges.end(); eid++ )
    {
        const cv::Point2f &toPt0 = this->frontTvPoints[ eid->ftvptIDs(0) ].toPt;
        const cv::Point2f &toPt1 = this->frontTvPoints[ eid->ftvptIDs(1) ].toPt;
        cv::line( edgeMap, toPt0, toPt1, cv::Scalar(255), 1, CV_AA );
    }
    for(int i=0;i<osz.height;i++)for(int j=0;j<osz.width;j++)
    {
        uchar ei = edgeMap.at<uchar>(i,j);
        if( !ei )
            continue;
        
        float ew = (float)ei/255.0f;
        uocv.drawImpWeight( this->warpWeightMap, cv::Point(j,i), ew * PTKERNELSZ );
    }
}

void ShadowWarpData::resetData()
{
    WarpData::resetData();
    this->fromPtsBottom.clear();
    this->toPtsBottom.clear();
    this->weightsBottom.clear();
}

void ShadowWarpData::addBottomPoint( const cv::Point2f &fromPt, const cv::Point2f &toPt, float weight )
{
    this->fromPtsBottom.push_back( fromPt );
    this->toPtsBottom.push_back( toPt );
    this->weightsBottom.push_back( weight );
}

//prepare warpWeightMap
void ShadowWarpData::prepareWeightMap(cv::Size osz)
{
    WarpData::prepareWeightMap(osz);
    
    Utils_Ocv &uocv = Utils_Ocv::instance();
    
    //harder point importance ( bottom alignment constraints )
    int hcNum = (int)this->fromPtsBottom.size();
    for(int j=0;j<hcNum;j++)
        uocv.drawImpWeight( this->warpWeightMap, this->toPtsBottom[j], this->weightsBottom[j] * PTKERNELSZ );
}


void ObjShdWarpData::reset()
{
    this->object.resetData();
    this->shadow.resetData();
}

void ivAppObjectFrameData::resetData()
{
    this->objectUsedClusterIDs.clear();
    this->shadowUsedClusterIDs.clear();
    
    this->objectCoWarper.clear();
    this->shadowCoWarper.clear();
    
    std::vector<ObjShdWarpData>::iterator it = this->begin();
    for(;it!=this->end();it++)
        it->reset();
}

//initialize the current(back) frame data
void ivAppObjectData::initCurrentFrameData(bool singleFD)
{
    if( singleFD )
    {
        this->frames.clear();
        this->frames.resize(1);
    }
    
    ivAppObjectFrameData &objFrameData = this->frames.back();
    const int clusterNum = (int)this->parts.size();
    objFrameData.resize( clusterNum );

    assert( (int)this->frames.size() > 0 );
}

//render object proxy
void ivAppObjectData::renderCurrentFrameData
(const ivAppSfM &objSfM, float objID, bool OBJ_CLSTR, bool OBJ_PROXY, bool SHD_PROXY) const
{
    const int clusterNum = (int)this->parts.size();
    const ivAppObjectFrameData &objFrameData = this->frames.back();
    
    // render order of clusters
    std::vector<int> renderOrder = std::vector<int>(clusterNum);
    for(int i=0;i<clusterNum;i++)
        renderOrder[i]=i;
    std::random_shuffle( renderOrder.begin(), renderOrder.end() );
    
    // render the front points & edges & point normals
    cv::Vec3b color;
    for( std::vector<int>::iterator cit=renderOrder.begin(); cit!=renderOrder.end(); cit++)
    {
        int i = *cit;

        //use cluster color
        if( OBJ_CLSTR )
            Utils_Ocv::instance().getColorRGB( color, i, clusterNum-1, 0 );
        else
            Utils_Ocv::instance().getColorRGB( color, objID*255, 255, 0 );
            
        //render object proxy
        if( OBJ_PROXY )
        {
            const WarpData &object = objFrameData[i].object;
            
            //render object point/edge
            cv::Vec3b color_obj = color * object.nfData.isSafe;
            glColor4ub(color_obj[0],color_obj[1],color_obj[2],128);
            object.renderPoint();
            
            //object point normal
//            glPushMatrix();
//            glMultMatrixf( objFrameData.gTransform.data() );
//            object.renderPointNormal(objSfM);
//            glPopMatrix();
        }
        
        //render shadow proxy        
        if( SHD_PROXY )
        {
            //render shadow point/edge
            const WarpData &shadow = objFrameData[i].shadow;
            cv::Vec3b color_shd = color * shadow.nfData.isSafe;
            glColor4ub(color_shd[0]/2,color_shd[1]/2,color_shd[2]/2,128);
            shadow.renderPoint();
            shadow.renderEdge();
        }
        
        if( OBJ_PROXY )
        {
            const WarpData &object = objFrameData[i].object;
            cv::Vec3b color_obj = color * object.nfData.isSafe;
            glColor4ub(color_obj[0],color_obj[1],color_obj[2],255);
            object.renderEdge();
        }
        
        //render shadow proxy
        if( SHD_PROXY )
        {
            //render shadow point/edge
            const WarpData &shadow = objFrameData[i].shadow;
            cv::Vec3b color_shd = color * shadow.nfData.isSafe;
            glColor4ub(color_shd[0]/2,color_shd[1]/2,color_shd[2]/2,128);
            shadow.renderEdge();
        }
        
    }
}


