//
//  ivAppObjectData.h
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#ifndef __ivManipulation__ivAppObjectData__
#define __ivManipulation__ivAppObjectData__

#import "Utils_Ocv.h"

#import "Plane.h"
#import "ivAppSfM.h"
#import "CoGEPWarper.h"
#import "patchCollager.h"


#define PTKERNELSZ 15

//nearest frame data
class NearestFrameData
{
public:
    int nframeID;
    float oDiff;
    bool isSafe;
};

class FrontPoint
{
public:
    int ptID;               //object point id
    Eigen::Vector4f pt3d;   //3D point in scene space
    
    cv::Point2f fromPt; // 2D point in image space
    cv::Point2f toPt;   // 2D point in novel view
    float value;        // object depth or shadow blurring weight
    float pWeight;      // point weight for warping
    float intensity;    // object diffuse intensity
};

//the visible edge
class FrontEdge
{
public:
    int edgeID;
    Eigen::Vector2i ftvptIDs;
    
    float alpha, b;     // line parameters
    void calcEdgeParams(const std::vector<FrontPoint> &frontTvPoints);
};

//warping data
class WarpData
{
public:
    void resetData();

    //rendering
    void renderPoint(bool coloredByAttribute = false) const ;
    void renderPointNormal(const ivAppSfM &objSfM) const;
    void renderEdge() const ;
    
    //ivAppObjAlgoMain setp 2 : ivAppObjAlgoFindNearestFrame
    NearestFrameData nfData;
    
    //ivAppObjAlgoMain step 3 : ivAppObjAlgoPreparePoints
    std::vector<int> sfm2fp; // sfm points to front points
    std::vector<FrontPoint> frontPoints; //front points (sz=|ptIDs|)
    int oriPtNum;//0~oriPtT in frontPoints is cluster points, others are shared points
    bool isReadyForWarping; //if true, the warp will be executed
    
    //ivAppObjAlgoMain step 4 : ivAppObjAlgoPrepareEdges
    std::vector<int> tvp2ftvp; // tv pts to front tv pts
    std::vector<FrontEdge> frontEdges;
    std::vector<FrontPoint> frontTvPoints;
    
    //ivAppObjAlgoMain step 5, 8 : ivAppObjAlgoSetWarps, ivAppObjAlgoCoGEPWarping
    
    //the weight of the warp
    float warpWeight;
    cv::Mat warpWeightMap;
    void prepareWeightMap(cv::Size osz);
    
    //warp result
    cv::Mat warpFrame;
    
    //point feature map & edge features & edge corners
    cv::Mat warpPointMap;
    cv::Mat warpEdgeMap;
    cv::Mat warpEdgeCornerMap;
    
    //the error of the warp result
    float warpError;
    
    //texture distortion error map
    cv::Mat warpDistortionMap;
};

class ShadowWarpData : public WarpData
{
public:
    void resetData();

    //ivAppObjAlgoMain step 8: ivAppObjAlgoAddShadowConstraints
    std::vector<cv::Point2f> fromPtsBottom; //2D points in image space
    std::vector<cv::Point2f> toPtsBottom;   //2D points in novel view
    std::vector<float> weightsBottom;       //weight of each point
    void addBottomPoint( const cv::Point2f &fromPt, const cv::Point2f &toPt, float weight );
    
    //prepare warpWeightMap
    void prepareWeightMap(cv::Size osz);
};

//planar part of the object
class ObjShdWarpData
{
public:
    //ivAppObjAlgoMain: warping data for object & shadow
    WarpData object;
    ShadowWarpData shadow;
    
    //reset data
    void reset();
};

class ivAppObjectFrameData : public std::vector<ObjShdWarpData>
{
public:
    //ivAppObjAlgoMain step 1 : ivAppObjAlgoSetup - setup params
    Eigen::Matrix4f gTransform;
    Eigen::Matrix4f gTransformInv;
    Eigen::Vector4f cameraPos_obj;
    Eigen::Vector4f lightPos_obj;
    Eigen::Matrix4f prjM_scn;
    void resetData();
    
    //ivAppObjAlgoMain setp 2: show if the selected frames are valid
    bool ifObjectPoseFramesValid;
    bool ifShadowPoseFramesValid;
    
    //ivAppObjAlgoMain step 3 : ivAppObjAlgoPreparePoints
    std::vector<int> objectUsedClusterIDs; //the clusters used in object
    std::vector<int> shadowUsedClusterIDs; //the clusters used in shadow
    
    //ivAppObjAlgoMain step 5, 8 : ivAppObjAlgoSetWarps & ivAppObjAlgoCoGEPWarping
    CoGEPWarper objectCoWarper;
    CoGEPWarper shadowCoWarper;
    
    //ivAppObjAlgoMain step 5, 8: ivAppObjAlgoCalcMotionField
    //from current frame i to the previous one i-1
    cv::Mat objectMotionField;
    cv::Mat shadowMotionField;
    
    //ivAppObjAlgoMain step 6, 9 : ivAppObjAlgoCollage - do collage for object & shadow
    // new collager is put at the front
    patchCollager objectCollager;
    patchCollager shadowCollager;
    
    //ivAppObjAlgoMain step 10 : ivAppObjAlgoMakeObject
    cv::Mat objPatch;       //the object patch
    cv::Mat objDepth;       //the object depth
    
    //ivAppObjAlgoMain step 11 : ivAppObjAlgoMakeShadow
    cv::Mat shadowPatch;    //the shadow patch
};

class ivAppObjectPart
{
public:
    //ivAppObjAlgoPreProcess : ivAppObjAlgoClustering : the clustered point IDs
    std::vector<int> ptIDs;
    int oriPtNum; //0~oriPtNum in ptIDs are cluster points, the others are shared points
    Plane tanPlane;
    
    //ivAppObjAlgoPreProcess : ivAppObjAlgoEdgeClassification
    std::vector<int> edgeIDs;
};

class ivAppObjectData
{
public:
    //ivAppObjAlgoPreProcess : ivAppObjAlgoClustering, ivAppObjAlgoEdgeClassification
    //time in-varying object part data
    std::vector<ivAppObjectPart> parts;
    
    //ivAppObjAlgoMain step 2->12
    //time varying object synthesis data
    //(new frame data are put at the back)
    std::list<ivAppObjectFrameData> frames;
    
    //ivAppObjAlgoMain setp 2: show if the selected frames are valid
    std::vector<bool> ifObjectMotionFramesValid;
    std::vector<bool> ifShadowMotionFramesValid;

    //the selected frames for each pose of object animation
    // cluster\pose    0, 1, 2, ...
    //              0, i, k, l, ...
    //              1, j, m, n, ...
    Eigen::Matrix< NearestFrameData, Eigen::Dynamic, Eigen::Dynamic > objNFrames;
    Eigen::Matrix< NearestFrameData, Eigen::Dynamic, Eigen::Dynamic > shdNFrames;

    //initialize the current(back) frame data
    //if singleFD, the size of framedata is set to 1
    void initCurrentFrameData(bool singleFD);

    //render object proxy
    void renderCurrentFrameData(const ivAppSfM &objSfM,
                                float objID, bool OBJ_CLSTR,
                                bool OBJ_PROXY, bool SHD_PROXY) const;
};

#endif /* defined(__ivManipulation__ivAppObjectData__) */



