//
//  ivAppObjects.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjects.h"

ivAppObjects ivAppObjects::inst;

//preprocess object
void ivAppObjects::goPreProcess()
{
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        ivAppObjAlgoPreProcess::instance().execute( &(*it) );
}

//cull back and find nearest frame for each object
void ivAppObjects::goCBFNFs(bool useSelectedFrames)
{
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        ivAppObjAlgoCBFNF::instance().execute( &(*it), useSelectedFrames );
}

//go selecting frames
void ivAppObjects::goSelectFrames()
{
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        ivAppObjAlgoSelectFrames::instance().execute( &(*it) );
}

//go main for each object
void ivAppObjects::goMains(bool useSelectedFrames, bool isFirstPose)
{
    ivAppScene &scn = ivAppScene::instance();
    scn.estimateGroundDepth();
    
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        ivAppObjAlgoMain::instance().execute( &(*it), useSelectedFrames, isFirstPose );
}

//get the max animation length
int ivAppObjects::getAniMax()
{
    std::vector<int> numVec;
    
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        numVec.push_back( it->objAnimation.getMotionNum()-1 );
    
    if( numVec.empty() )
        return 0;
    else
        return *std::max_element( numVec.begin(), numVec.end() );
    
    return 0;
}

//set object motion
void ivAppObjects::setMotion(int i)
{
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        it->objAnimation.gotoMotion(i);
}

//set parameters for exporting
void ivAppObjects::setCLG_alpha(float alpha)
{
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
        it->objParams.CLG_alpha = alpha;
}

class pixelData
{
public:
    float depth;
    cv::Vec4b c;
};

bool depCompare ( const pixelData& l, const pixelData& r)
{ return l.depth > r.depth; }

//rendering object pathes
void ivAppObjects::renderObjects(bool OBJ_EDGE, BOOL DEPTHMAP)
{
    if( ivaObjList.empty() )
        return;
    
    std::vector<cv::Mat> objPatches;
    std::vector<cv::Mat> objDepths;
    
    std::list< ivAppObject >::iterator it = ivaObjList.begin();
    for( int i=0; it!=ivaObjList.end(); it++,i++)
    {
        ivAppObjectData &objData = it->objData;
        ivAppObjectFrameData &objFrameData = objData.frames.back();
        patchCollager &objectCollager = objFrameData.objectCollager;
        
        if( !objectCollager.isReady() )
            continue;
        
        //get object patch
        cv::Mat objPatch = objFrameData.objPatch.clone();
        cv::Mat &objDepth = objFrameData.objDepth;
       
        //cluster color
        float CLG_alpha = it->objParams.CLG_alpha;
        if( CLG_alpha > 0.0f )
            objPatch = objectCollager.getImgCollage( CLG_alpha );
        
        //visualize the edges & edge corners
        if( OBJ_EDGE )
        {
            cv::Mat edgeMap = objectCollager.getEdgeMapCollage();
            objPatch.setTo( cv::Scalar::all(255), edgeMap>128 );
            
            cv::Mat edgeCornerMap = objectCollager.getEdgeMapCornerCollage();
            objPatch.setTo( cv::Scalar(0,0,0,255), edgeCornerMap>128 );
        }
        
        objPatches.push_back( objPatch );
        objDepths.push_back( objDepth );
    }
    
    //no object patch can be rendered
    if( objPatches.empty() )
        return;
    
    const cv::Scalar infDepth( std::numeric_limits<float>::max() );
    cv::Mat colorBuffer = cv::Mat::zeros( OSZ_HEIGHT, OSZ_WIDTH, CV_8UC4 );
    cv::Mat depthBuffer = cv::Mat( OSZ_HEIGHT, OSZ_WIDTH, CV_32FC1, infDepth );
    
    int objNum = (int)objPatches.size();
    //for each pixel
    for(int i=0;i<OSZ_HEIGHT;i++)for(int j=0;j<OSZ_WIDTH;j++)
    {
        //for each object, get the pixel data
        std::vector<pixelData> pData(objNum);
        for(int k=0;k<objNum;k++)
        {
            pData[k].depth = objDepths[k].at<float>(i,j);
            pData[k].c = objPatches[k].at<cv::Vec4b>(i,j);
        }
        
        //sort according to depth (descending order)
        std::sort( pData.begin(), pData.end(), depCompare );
        
        //current pixel
        cv::Vec4b &curColor = colorBuffer.at<cv::Vec4b>(i,j);
        float &curDepth = depthBuffer.at<float>(i,j);
        float curAlpha = 1.0f;
        
        //determine the final color and depth
        for(int k=0;k<objNum;k++)
        {
            //depth test
            if( curDepth <= pData[k].depth )
                continue;

            curDepth = pData[k].depth;
            
            //blending
            float alpha = (float)pData[k].c[3]/255.0f;
            curAlpha *= 1.0f-alpha;
            curColor = curColor * (1.0f-alpha) + pData[k].c * alpha;
        }
        
        //adjust color to correct blending result
        curAlpha = 1.0f - curAlpha;
        curColor /= curAlpha;
        curColor[3] = 255 * curAlpha;
    }
    
    if( DEPTHMAP )
    {
        cv::Mat groundDepth = ivAppScene::instance().groundDepthMap;
        groundDepth.copyTo( depthBuffer, groundDepth<depthBuffer );
        cv::normalize(depthBuffer, depthBuffer, 0.0, 1.0, cv::NORM_MINMAX);
        depthBuffer.convertTo( colorBuffer, CV_8UC1, -255, 255 );
    }
    
    Utils_Ogl::renderMat( colorBuffer );
}

//rendering shadow pathes
void ivAppObjects::renderShadows(bool OBJ_EDGE)
{
    if( ivaObjList.empty() )
        return;

    //combine objects
    std::list< ivAppObject >::iterator it = ivaObjList.begin();
    for( int i=0; it!=ivaObjList.end(); it++,i++)
    {
        ivAppObjectData &objData = it->objData;
        ivAppObjectFrameData &objFrameData = objData.frames.back();
        patchCollager &shadowCollager = objFrameData.shadowCollager;
        
        //if the shadow is not ready
        if( !shadowCollager.isReady() )
            continue;

        cv::Mat shadow = objFrameData.shadowPatch.clone();
        
        //visualize the edges & edge corners
        if( OBJ_EDGE )
        {
            cv::Mat edgeMap = shadowCollager.getEdgeMapCollage();
            shadow.setTo( cv::Scalar::all(255), edgeMap>128 );
            
            cv::Mat edgeCornerMap = shadowCollager.getEdgeMapCornerCollage();
            shadow.setTo( cv::Scalar(0,0,0,255), edgeCornerMap>128 );
        }
        
        Utils_Ogl::renderMat(shadow);
    }
}

void ivAppObjects::renderProxies
(bool OBJ_GUIDE, bool OBJ_NCAMERA,
bool OBJ_CLSTR, bool OBJ_PROXY, bool SHD_PROXY)
{
    std::list< ivAppObject >::iterator it=ivaObjList.begin();
    for(;it!=ivaObjList.end();it++)
    {
        float objID = (float)std::distance( ivaObjList.begin(), it ) / (float)( ivaObjList.size() - 1 );
        bool isActive = it == pCurObject;
        it->renderProxy( isActive, OBJ_GUIDE, OBJ_NCAMERA,
                        objID, OBJ_CLSTR, OBJ_PROXY, SHD_PROXY );
    }
}

void ivAppObjects::renderObectMotions()
{
    if(this->empty())
        return;
    
    ivAppObject *obj = this->getCurrentObject();
    obj->objAnimation.renderObjAnimationPath();
}

//io
std::istream& operator>>(std::istream& is, ivAppObjects& objects)
{
    objects.clear();
    
    std::string trash;
    
    int objNum;
    is >> trash >> objNum;
    
    int objInd;
    is >> trash >> objInd;
    
    is >> trash;

    for(int i=0;i<objNum;i++)
    {
        ivAppObject obj;
        objects.addObject(obj);
        ivAppObject *pobj = objects.getCurrentObject();
        is >> *pobj;
    }
    objects.goPreProcess();
    objects.pointToObjAt(objInd);

    is >> trash;
    
    return is;
}

std::ostream& operator<<(std::ostream& os, const ivAppObjects& objects)
{

    os << "OBJ_NUMBER: " << objects.size() << std::endl;
    os << "OBJ_INDEX: " << objects.getCurPos() << std::endl;
    os << "===============OBJECT_LIST===============" << std::endl;
    std::list< ivAppObject >::const_iterator it=objects.ivaObjList.begin();
    for(;it!=objects.ivaObjList.end();it++)
        os << *it;
    os << "=============EOF_OBJECT_LIST==============" << std::endl;
    
    return os;
}