//
//  ivAppUsrLabel.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppUsrLabel.h"

//initialization
void ivAppUsrLabel::init( ivAppSfM *sfmData, ivAppVideoCanny *canny, std::string tcName )
{
    this->canny = canny;
    this->sfmData = sfmData;

    this->lblPoly.init(canny);
    
    std::string tvmPath = ivDirParams::getTvMPath(tcName);
    this->tvm.readTVModel(tvmPath, sfmData);
    
    this->modelPose.reset();
    
    this->pickedPointID = -1;
}

//add a point
void ivAppUsrLabel::addPoint(int framePos)
{
    //when picking point, dont add
    if( this->pickedPointID>=0 )
        return;
    
    int fNum = (int)this->sfmData->vFrameData.size();
    
    //check the user-added point exists
    int selectedPID = this->tvm.selectPoint(framePos, cursor);
    bool ifToAddCandidate = selectedPID < 0;
    
    //if not, add a point candidate
    if( ifToAddCandidate )
        selectedPID = this->tvm.pushPointCandidate(fNum, framePos, cursor);
    Eigen::Vector2f *selectedPt2D = this->tvm.getPoint2D(framePos, selectedPID);
    
    //try to add a point to the polygon
    bool accept = this->lblPoly.addPoint2D( tvm.vtvPoint, framePos, *selectedPt2D, selectedPID, lblParams::isEdgeSnapping );
    if( !accept && ifToAddCandidate )
        this->tvm.popPointCandidate(framePos,selectedPID);
    
    if( !this->lblPoly.isReady )
        return;

    this->tvm.addPrimitive( framePos, sfmData, this->lblPoly );
    
    this->lblPoly.reset();
}

//remove 2D points from polygon
void ivAppUsrLabel::rmPolygon2DPoint()
{
    int ptNum = (int)this->lblPoly.plgn2d.size();
    if( ptNum <= 0 )
        return;
    
    int pid = this->lblPoly.prim[ptNum-1];
    int framePos = this->lblPoly.curFrame;

    this->tvm.popPointCandidate(framePos, pid);
    this->lblPoly.rmPoint2D();
}

//pick a point
void ivAppUsrLabel::pickPoint(int framePos)
{
    //when adding point, dont pick
    if( this->lblPoly.plgn2d.size() > 0 )
        return;
    
    if( this->pickedPointID<0 )
        this->pickedPointID = this->tvm.selectPoint(framePos, cursor);
    else
    {
        //estimate 3D, digestion
        lblPolygon2D lblPolyAdjust(canny);
        this->tvm.adjustOneRing(this->pickedPointID, framePos, cursor, lblPolyAdjust, sfmData);
        this->pickedPointID = -1;
    }
}

//adjust picked point
void ivAppUsrLabel::adjustPickedPoint(int framePos)
{
    //no picked point, no adjusting
    if( this->pickedPointID<0 )
        return;

    //adjust 2d point , do not estimate 3D
    lblPolygon2D lblPolyAdjust(canny);
    this->tvm.adjustOneRing(this->pickedPointID, framePos, cursor, lblPolyAdjust);
}

//cancel adjusting
void ivAppUsrLabel::cancelAdjusting(int framePos)
{
    //no picked point, no adjusting
    if( this->pickedPointID<0 )
        return;
    
    this->tvm.cancelAdjusting(framePos, sfmData, this->pickedPointID );
    this->pickedPointID = -1;
}

//remove picked one ring
void ivAppUsrLabel::removePickedOneRing(int framePos)
{
    if( this->pickedPointID<0 )
        return;
    
    this->tvm.rmOneRingPrims( framePos, this->pickedPointID );
    this->pickedPointID = -1;
}

//remove picked primitive
void ivAppUsrLabel::removePickedPrim(int framePos)
{
    if( this->pickedPointID<0 )
        return;
    
    this->tvm.rmPrimitive( framePos, this->pickedPointID );
    this->cancelAdjusting(framePos);
}

//goto next selected prim
void ivAppUsrLabel::nextPrim()
{
    tvm.primPointer++;
}

//set primitive critical
void ivAppUsrLabel::setCritical(int framePos)
{
    if( this->pickedPointID<0 )
        return;
    
    int fNum = (int)sfmData->vFrameData.size();
    this->tvm.setPrimCritical( framePos, fNum, this->pickedPointID );
    this->cancelAdjusting(framePos);
}

//clear user labeling
void ivAppUsrLabel::clearCurrentUserLabeling(int framePos)
{
    //clear all the labeling
    int ptNum = (int)this->lblPoly.plgn2d.size();
    for(int i=0;i<ptNum;i++)
        this->rmPolygon2DPoint();
    
    //cancel adjusting
    this->cancelAdjusting(framePos);
}

//rendering 2D status
cv::Mat ivAppUsrLabel::render2D(int framePos)
{
    cv::Mat canvas = cv::Mat::zeros( this->sfmData->height, this->sfmData->width, CV_8UC4 );
    
    if(lblParams::isEdgeSnapping)
        cv::circle( canvas, cv::Point(cursor(0),cursor(1)), lblParams::POINT_SZ*2, Utils_Color::RED_A, -1 );
    
    this->tvm.renderOCV( canvas, framePos, this->pickedPointID );
    
    this->lblPoly.render2D_prev(canvas, tvm.vtvPoint, framePos, cursor );
    
    return canvas;
}

//rendering edge
cv::Mat ivAppUsrLabel::render2DEdge(int framePos, float edgeWidth, bool optimized)
{
    Eigen::Matrix4f gtransform = optimized ? modelPose.gTransform() : Eigen::Matrix4f::Identity();
    cv::Mat edgeMap = this->tvm.getEdgeMap( this->sfmData, framePos, gtransform, framePos, edgeWidth, true, optimized );
    
    //make it have 4 channels
    std::vector<cv::Mat> rgbaEdgeMaps(4, edgeMap);
    cv::Mat edgeMaprgba;
    cv::merge(rgbaEdgeMaps,edgeMaprgba);
    
    return edgeMaprgba;
}

//optimize the model points
void ivAppUsrLabel::doSBA()
{
    this->tvm.doSBA( this->sfmData );
}

//read the model
void ivAppUsrLabel::readModel(std::string objFileName)
{
    this->tvm.readTVModel(objFileName, this->sfmData);
}

//export the model
void ivAppUsrLabel::exportModel(std::string objFileName)
{
    int fNum = (int)this->sfmData->vFrameData.size();
    this->tvm.exportTVModel(objFileName, fNum);
}
