//
//  tvmRasterizer.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by sulaispig on 1/5/15.
//  Copyright (c) 2015 sulaispig. All rights reserved.
//

#include "tvmRasterizer.h"

//set data by interpolation
void VER::interpolate(const VER &V1, const VER &V2, int minV, int maxV, int v)
{
    //interpolation
    if( maxV == minV )
        *this = V1;
    else
    {
        pt = (v-minV)*V2.pt + (maxV-v)*V1.pt;
        pt /= maxV-minV;
    }
}


void tvmRasterizer::renderPixel( int x, int y, const VER &cur, bool renderColor )
{
    if( isOutOfBoundary(x,y) )
        return;
    
    float &d = this->depthBuffer.at<float>(y,x);
    if( d < cur.pt(2) )
        return;
    
    d = cur.pt(2);
    if( renderColor )
        this->colorBuffer.at<uchar>(y,x) = 255;
}
bool tvmRasterizer::isOutOfBoundary(int x, int y)
{
    bool isOutOfBoundaryX = x < 0 || x >= this->imgSz.width;
    bool isOutOfBoundaryY = y < 0 || y >= this->imgSz.height;
    return isOutOfBoundaryX || isOutOfBoundaryY;
}
void tvmRasterizer::renderHorizontalTrapezoid( const VER &ul, const VER &ur, const VER &bl, const VER &br )
{
    //floor
    int y_min = ul.pt(1);
    int y_max = bl.pt(1);
    
    for( int y=y_min; y<=y_max; y++ )
    {
        VER L,R;
        L.interpolate(ul, bl, y_min, y_max, y);
        R.interpolate(ur, br, y_min, y_max, y);
        
        //floor
        int x_min = L.pt(0);
        int x_max = R.pt(0);
        
        for(int x=x_min;x<=x_max;x++)
        {
            VER CUR;
            CUR.interpolate(L, R, x_min, x_max, x);
            this->renderPixel(x, y, CUR, false);
        }
    }
}
void tvmRasterizer::renderTriangleFILL( VER v1, VER v2, VER v3 )
{
    //sorting y
    if( v1.pt(1) > v2.pt(1) )
        std::swap(v1,v2);
    if( v1.pt(1) > v3.pt(1) )
        std::swap( v1, v3 );
    if( v2.pt(1) > v3.pt(1) )
        std::swap( v2, v3 );
    
    VER v4;
    v4.interpolate( v1, v3, v1.pt(1), v3.pt(1), v2.pt(1) );
    
    if( v2.pt(0) > v4.pt(0) )
        std::swap(v2,v4);
    
    //split the triangles into horizontal trapezoid
    this->renderHorizontalTrapezoid(v1, v1, v2, v4);
    this->renderHorizontalTrapezoid(v2, v4, v3, v3);
}
void tvmRasterizer::renderEdge( VER v1, VER v2 )
{
    int x1 = (int)v1.pt(0);
    int y1 = (int)v1.pt(1);
    int x2 = (int)v2.pt(0);
    int y2 = (int)v2.pt(1);
    
    if( y1 > y2 )
    {
        std::swap(x1,x2);
        std::swap(y1,y2);
    }
    
    //categorize the edge---------------------------------------------------------------------
    int edgeType;
    int delta_x = x2-x1;
    int delta_y = y2-y1;
    
    if( delta_x >= 0 )
    {
        if( delta_x >= delta_y ) edgeType = 1;
        else edgeType = 2;
    }
    else
    {
        if( -delta_x >= delta_y ) edgeType = 4;
        else edgeType = 3;
    }
    
    int xx1, yy1, xx2, yy2;
    //mirror-reflection
    if( edgeType == 1 )
    {
        xx1 = x1;
        yy1 = y1;
        xx2 = x2;
        yy2 = y2;
    }
    else if( edgeType == 2 )
    {
        xx1 = y1;
        yy1 = x1;
        xx2 = y2;
        yy2 = x2;
    }
    else if( edgeType == 3 )
    {
        xx1 = y1;
        yy1 = -x1;
        xx2 = y2;
        yy2 = -x2;
    }
    else if( edgeType == 4 )
    {
        xx1 = -x1;
        yy1 = y1;
        xx2 = -x2;
        yy2 = y2;
    }
    //-------------------------------------------------------------------
    
    delta_x = xx2-xx1;
    delta_y = yy2-yy1;
    
    int ei = 2 * ( delta_y - delta_x );
    int x = xx1;
    int y = yy1;
    int x_draw = x;
    int y_draw = y;
    
    for( int x = xx1; x<=xx2; x++ ){
        
        //mirror-reflect back
        if( edgeType == 1 ){
            x_draw = x;
            y_draw = y;
        }else if( edgeType == 2 ){
            x_draw = y;
            y_draw = x;
        }else if( edgeType == 3 ){
            x_draw = -y;
            y_draw = x;
        }else if( edgeType == 4 ){
            x_draw = -x;
            y_draw = y;
        }
        
        VER CUR;
        CUR.interpolate( v2, v1, xx2, xx1, x );
        this->renderPixel( x_draw, y_draw, CUR, true );
        
        //mid-point check
        if( ei >= 0 ){
            ei += 2 * ( delta_y - delta_x );
            y++;
        }else{
            ei += 2 * delta_y;
        }
    }
    
}




//init
void tvmRasterizer::init( ivAppSfM *sfmData )
{
    this->sfmData = sfmData;
    this->imgSz = cv::Size( sfmData->width, sfmData->height );
    this->colorBuffer = cv::Mat( this->imgSz, CV_8UC1, cv::Scalar(0) );
    this->depthBuffer = cv::Mat( this->imgSz, CV_32FC1, cv::Scalar(0.0f) );
}


//raster the tv model at framePos
cv::Mat tvmRasterizer::raster(ivAppTvModel &tvm, ivAppObjPose &pose, int framePos)
{
    //clear colorbuffer & depthbuffer
    this->colorBuffer = cv::Scalar::all(0);
    this->depthBuffer = cv::Scalar::all( std::numeric_limits<float>::max() );
    
    //project all points
    Eigen::Matrix4f prjM = this->sfmData->K * this->sfmData->vFrameData[framePos].M * pose.gTransform();
    int pNum = (int)tvm.vtvPoint.size();
    std::vector<VER> vs(pNum);
    for(int i=0;i<pNum;i++)
    {
        Eigen::Vector4f prjedPt = prjM * tvm.vtvPoint[i].tvv3D[framePos];
        vs[i].pt(0) = prjedPt(0)/prjedPt(2);
        vs[i].pt(1) = prjedPt(1)/prjedPt(2);
        vs[i].pt(2) = prjedPt(2);
    }
    
    //render the prims
    int primNum = (int)tvm.vtvPrim.size();
    for(int i=0;i<primNum;i++)
    {
        tvTriPrim &curPrim = tvm.vtvPrim[i];
        
        //back face check
        if( !curPrim.isFront(framePos) )
            continue;
        
        //render triangles
        int triNum = (int)curPrim.triangles.size();
        for(int j=0;j<triNum;j++)
        {
            const Eigen::Vector3i &tri = curPrim.triangles[j];
            this->renderTriangleFILL( vs[tri[0]], vs[tri[1]], vs[tri[2]] );
        }
    }
    
    for(int i=0;i<primNum;i++)
    {
        tvTriPrim &curPrim = tvm.vtvPrim[i];
        
        //back face check
        if( !curPrim.isFront(framePos) )
            continue;
        
        //render edges
        int plgnPtNum = (int)curPrim.polygon.size();
        for(int j=0;j<plgnPtNum;j++)
        {
            int k = (j+1)%plgnPtNum;
            int pid1 = curPrim.polygon[j];
            int pid2 = curPrim.polygon[k];
            this->renderEdge( vs[pid1], vs[pid2] );
        }
    }

    return this->colorBuffer.clone();
}



