//
//  ivAppData.m
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#import "ivAppData.h"

@implementation ivAppData

@synthesize testCases;
@synthesize selectedCase;
@synthesize framePos;
@synthesize frameMaxPos;
@synthesize isStroke;
@synthesize strokeSize;

- (id)init
{
    self = [super init];
    if (self)
    {
        //start initialization
        NSOpenPanel *openPanel = [NSOpenPanel openPanel];
        openPanel.title = @"Open The Working Folder.";
        [openPanel setCanChooseDirectories:YES];
        [openPanel setCanChooseFiles:NO];
        NSInteger status = [openPanel runModal];
        if( status != NSModalResponseOK )
            exit(EXIT_SUCCESS);
        
        std::string wPath = [[[[openPanel URL] path] stringByAppendingString:@"/"] UTF8String];
        chdir(wPath.c_str());
        
        //init the list of the input test cases
        testCases = [[NSMutableArray alloc] init];
        std::vector<std::string> dirs = Utils_DirList::parseDirList( ivDirParams::videoDir );
        std::vector<std::string>::iterator it = dirs.begin();
        for(;it!=dirs.end();it++)
            [testCases addObject:@(it->c_str())];
        
        //the name of the selected test case
        selectedCase = @(dirs[0].c_str());
        
        isReady = false;
    }
    
    return self;
}

//initialization
-(void) reset
{
    //the name of test case in std::string
    std::string tcName = [self.selectedCase UTF8String];
    std::string videoPath = ivDirParams::getVideoPath(tcName);
    std::string maskPath  = ivDirParams::getMaskPath(tcName);
    
    rStates = [NSMutableDictionary dictionaryWithObjectsAndKeys:
               @"1",  VIDEO_KEY,
               @"0",   MASK_KEY,
               @"1", STROKE_KEY,
               @"1",  SCENE_KEY,
               nil];
    
    //initialiation
    video.readVideo( videoPath, INT_MAX );
    masks.readVideo(  maskPath, INT_MAX );
    sfm_scn.readVoodooSfM( tcName );
    stroke.init( video.frameSz );
    objPose.reset();

    isReady = true;
    self.isStroke = 1;
    self.strokeSize = stroke.strkRadius;
    self.framePos = 0;
    self.frameMaxPos = video.frameNum-1;
}

//render results
-(void) render
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    if( !isReady )
        return;
    
//===============================================================
// OGL 2D rendering
//===============================================================
    std::vector< cv::Mat > imgs;
    
    //===============================================================
    // draw input video frame
    //===============================================================
    if( [[rStates objectForKey:VIDEO_KEY] boolValue] && video.isReady )
        imgs.push_back(video.frames[framePos]);
    
    //===============================================================
    // draw input object mask
    //===============================================================
    if( [[rStates objectForKey:MASK_KEY] boolValue] && masks.isReady )
        imgs.push_back(masks.frames[framePos]);
    
    //===============================================================
    // draw user stroke
    //===============================================================
    if( [[rStates objectForKey:STROKE_KEY] boolValue] )
        imgs.push_back( stroke.getCurStroke() );
    //===============================================================
    // rendering
    //===============================================================
    glDisable(GL_DEPTH_TEST);
    std::vector<cv::Mat>::iterator it = imgs.begin();
    for(;it!=imgs.end();it++)
    {
        Utils_Ogl::renderMat( *it );
        glClear( GL_DEPTH_BUFFER_BIT );
    }
    glEnable(GL_DEPTH_TEST);
    
//===============================================================
// OGL 3D rendering (Scene Camera)
//===============================================================
    
    //==============================================================
    // draw input SfM point cloud
    //==============================================================
    if( [[rStates objectForKey:SCENE_KEY] boolValue] )
    {
        //the tracking data
        frameData &fd_Scn = sfm_scn.vFrameData[ framePos ];
        Utils_Ogl::setMatrix( sfm_scn.width, sfm_scn.height, sfm_scn.K, fd_Scn.M );
        sfm_scn.render( framePos );
        Utils_Ogl::resetMatrix();
    }
    
    //==============================================================
    // draw object SfM
    //==============================================================
    if( [[rStates objectForKey:OBJECT_KEY] boolValue] )
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_CULL_FACE);
        
        //the tracking data
        frameData &fd_Scn = sfm_obj.vFrameData[ framePos ];
        Utils_Ogl::setMatrix( sfm_obj.width, sfm_obj.height, sfm_obj.K, fd_Scn.M );
        
        glPushAttrib( GL_CURRENT_BIT );
        glColor3ub(0,255,0);
        sfm_obj.render( framePos );
        glPopAttrib();
        
        glPushAttrib( GL_POLYGON_BIT | GL_LINE_BIT | GL_CURRENT_BIT );
        {
            glPushMatrix();
            Eigen::Matrix4f m = objPose.gTransform();
            glMultMatrixf(m.data());

            glLineWidth(2);
            Utils_Ogl::drawAxis();
            
            glPushMatrix();
            glTranslatef(0.0f, 0.0f, 0.5f);
            glScalef(1.0f, 1.0f, 0.5f);

            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            Utils_Ogl::drawCube();
            
            glPopMatrix();
            glPopMatrix();
        }
        glPopAttrib();
        
        Utils_Ogl::resetMatrix();
        
        glEnable(GL_CULL_FACE);
        glEnable(GL_DEPTH_TEST);
    }
}

@end






