//
//  easyPatchCollager.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "patchCollager.h"

//push the data
void collageData::push_back
(int patchID, const Eigen::Vector3f &colorID,
 cv::Mat &patch, cv::Mat &dataMap,
 cv::Mat &edgeMap, cv::Mat &edgeCornerMap)
{
    this->patchIDs.push_back( patchID );
    this->colorIDs.push_back( colorID );
    
    this->patchSet.push_back( patch );
    this->dataMapSet.push_back( dataMap );
    this->edgeMapSet.push_back( edgeMap );
    this->edgeCornerMapSet.push_back( edgeCornerMap );
}

//clear
void collageData::clear()
{
    this->patchIDs.clear();
    this->colorIDs.clear();
    
    this->patchSet.clear();
    this->dataMapSet.clear();
    this->edgeMapSet.clear();
    this->edgeCornerMapSet.clear();
}

//combine the imgSet
cv::Mat patchCollager::collageImgSet(const imgSet &iset, float alpha)
{
    cv::Mat imgCollage = iset[0].clone();
    int type = iset.imgCollage.type();
    
    float beta = 1.0f - alpha;
    
    for(int i=0;i<num_pixels;i++)
    {
        int rowID = i/width;
        int colID = i%width;
        
        const Eigen::Vector3f &color = clgData.colorIDs[ labels[i] ];
        
        if( type==CV_8UC1 )
        {
            const uchar &src = iset[labels[i]].at<uchar>(rowID,colID);
            uchar &trg = imgCollage.at<uchar>(rowID,colID);
            trg = src;
        }
        
        if( type==CV_8UC3 )
        {
            const cv::Vec3b &src = iset[labels[i]].at<cv::Vec3b>(rowID,colID);
            cv::Vec3b &trg = imgCollage.at<cv::Vec3b>(rowID,colID);
            for(int j=0;j<3;j++)
                trg[j] = color(j)*alpha + src[j]*beta;
        }
        
        if( type==CV_8UC4 )
        {
            const cv::Vec4b &src = iset[labels[i]].at<cv::Vec4b>(rowID,colID);
            cv::Vec4b &trg = imgCollage.at<cv::Vec4b>(rowID,colID);
            for(int j=0;j<3;j++)
                trg[j] = color(j)*alpha + src[j]*beta;
            trg[3] = src[3];
        }
        
        if( type==CV_32FC1 )
        {
            const float &src = iset[labels[i]].at<float>(rowID,colID);
            float &trg = imgCollage.at<float>(rowID,colID);
            trg = src;
        }
    }
    
    return imgCollage;
}

//combine all the sets
void patchCollager::collageAllSets()
{
    clgData.patchSet.imgCollage         = this->collageImgSet( clgData.patchSet         );
    clgData.dataMapSet.imgCollage       = this->collageImgSet( clgData.dataMapSet       );
    clgData.edgeMapSet.imgCollage       = this->collageImgSet( clgData.edgeMapSet       );
    clgData.edgeCornerMapSet.imgCollage = this->collageImgSet( clgData.edgeCornerMapSet );
}

//initialization
void patchCollager::init
(int patchID, const Eigen::Vector3f &colorID, cv::Mat &patch0, cv::Mat &dataMap0,
 cv::Mat &edgeMap0, cv::Mat &edgeCornerMap0)
{
    //size is not specified
    assert( this->width > 0 && this->height > 0 );

    //push_back collage data
    this->clgData.push_back( patchID, colorID, patch0, dataMap0, edgeMap0, edgeCornerMap0 );
    
    //initial labels
    labels = std::vector<int>(num_pixels,0);
    
    //collage imgSets
    this->collageAllSets();
}

//if the collage is ready
bool patchCollager::isReady()
{
    return this->clgData.patchSet.size()>0;
}

//clear all the data
void patchCollager::clear()
{
    this->width = this->height = this->num_pixels = 0;
    this->clgData.clear();
    this->labels.clear();
    this->srcIDWeightMaps.clear();
}

//initialize the collage size
void patchCollager::initSize( int width, int height )
{
    //image parameters
    this->width = width;
    this->height = height;
    this->num_pixels = width * height;
}

//add other images
void patchCollager::add
(int patchID, const Eigen::Vector3f &colorID,
 cv::Mat &patch, cv::Mat &weightMap, cv::Mat &distortionMap, 
 cv::Mat &edgeMap, cv::Mat &edgeCornerMap, cv::Mat &temporalWeightMap,
 float featureWeight, float distortionWeight,
 float cDiffWeight, float gDiffWeight, float edgeWeight,
 float edgeAttenuR, float tmWeight )
{
    assert( patch.type()==CV_8UC1 || patch.type()==CV_8UC3 || patch.type()==CV_8UC4 );
    assert( weightMap.type()==CV_32FC1 );
    assert( edgeMap.type()==CV_8UC1 );
    assert( edgeCornerMap.type()==CV_8UC1 );
    assert( patch.cols == this->width && patch.rows == this->height );
    
    // prepare data energy
    //==============================================================
    cv::Mat dataMap = featureWeight * ( 1.0f - weightMap ) + distortionWeight * distortionMap;
    
    //similar to previous collages (temporal consideration)
    dataMap += tmWeight * ( 1.0f - temporalWeightMap );
    //==============================================================
    
    //handle the first patch
    if( !this->isReady() )
    {
        this->init( patchID, colorID, patch, dataMap, edgeMap, edgeCornerMap );
        return;
    }
    
    //prepare smoothness energy
    //==============================================================
    Utils_Ocv &ocv = Utils_Ocv::instance();
    
    //color difference
    cv::Mat clrDiff_E = cDiffWeight * ocv.sumAbsDiff( clgData.patchSet.imgCollage, patch );
    
    //gradient difference
    cv::Mat clgGrad = ocv.getGradient( clgData.patchSet.imgCollage );
    cv::Mat imgGrad = ocv.getGradient( patch );
    cv::Mat gradDiff_E = gDiffWeight * ocv.sumAbsDiff( clgGrad, imgGrad );
    
    //edge penalty for smoothness edgeWeight
    cv::Mat unionEdgeCornerMap = clgData.edgeCornerMapSet.imgCollage | edgeCornerMap;
    cv::Mat cornerWeight;
    cv::distanceTransform( 255-unionEdgeCornerMap, cornerWeight, CV_DIST_L2, 3 );
    cv::threshold( cornerWeight, cornerWeight, edgeAttenuR, 0, cv::THRESH_TRUNC );
    cornerWeight *= 1.0f/edgeAttenuR;
    
    cv::Mat unionEdgeMap = clgData.edgeMapSet.imgCollage | edgeMap;
    cv::Mat edge_E;
    unionEdgeMap.convertTo( edge_E, CV_32FC1, edgeWeight/255.0f );
    
    edge_E.mul( cornerWeight );
    
    //smoothness energy
    cv::Mat SmoothnessEnergy = clrDiff_E + gradDiff_E + edge_E;
    
    //similar to previous collages (temporal consideration)
    SmoothnessEnergy += tmWeight * temporalWeightMap;
    //==============================================================
    
    // set data cost
    //==============================================================
    //prefer feature points
    cv::Mat data_0, data_1;
    data_0 = clgData.dataMapSet.imgCollage;
    data_1 = dataMap;
    
    //weighted by featureWeight
    std::vector<int> data(num_pixels*2,0);
    for ( int i = 0; i < num_pixels; i++ )
    {
        int rowID = i/width;
        int colID = i%width;
        
        data[i*2+0] = data_0.at<float>( rowID , colID );// * error0;
        data[i*2+1] = data_1.at<float>( rowID , colID );// * error1;
    }
    //==============================================================
    
    // set smooth costs
    //==============================================================
    int smooth[4] = {0,1,1,0};
    
    //spatially varying arrays V and H
    std::vector<int> V(num_pixels, 0);
    std::vector<int> H(num_pixels, 0);
    
    for(int i=0;i<num_pixels;i++)
    {
        int rowID_c = i/width;
        int colID_c = i%width;
        int rowID_d = rowID_c + 1;
        int colID_r = colID_c + 1;
        
        //current pixel
        float X_c =  SmoothnessEnergy.at<float>( rowID_c, colID_c );
        
        //bottom pixel
        // neighbor (rowID_c, colID_c) <-> (rowID_d, colID_c)
        if( rowID_d < height )
        {
            float X_d =  SmoothnessEnergy.at<float>( rowID_d, colID_c );
            V[i] = X_c + X_d;
        }
        
        //right pixel
        // neighbor (rowID_c, colID_c) <-> (rowID_c, colID_r)
        if( colID_r < width )
        {
            float X_r =  SmoothnessEnergy.at<float>( rowID_c, colID_r );
            H[i] = X_c + X_r;
        }
    }
    //==============================================================
    
    // start optimization
    //==============================================================
    try{
        GCoptimizationGridGraph gc(width,height,2);
        gc.setDataCost(&data[0]);
        gc.setSmoothCostVH(smooth,&V[0],&H[0]);
        gc.swap(-1);
        
        int newLabelID = (int)clgData.patchSet.size();
        for ( int  i = 0; i < num_pixels; i++ )
            if( gc.whatLabel(i)==1 )
                labels[i] = newLabelID;
    }
    catch (GCException e)
    {
        e.Report();
    }
    //==============================================================
    
    // save the results
    //==============================================================
    this->clgData.push_back( patchID, colorID, patch, dataMap, edgeMap, edgeCornerMap );
    this->collageAllSets();
    //==============================================================
}

//calculate source ID weight map
void patchCollager::calcSrcIDWeightMaps(float CLG_tm_alpha)
{
    //                        ------- 1
    //                       /
    //                  ...
    //               /
    // 0 ____________<-alpha->
    // other labels |        srcID
    int patchNum = (int)this->clgData.patchSet.size();
    
    //init
    std::vector<cv::Mat> srcIDMaps( patchNum );
    this->srcIDWeightMaps.resize( patchNum+1 );
    for(int k=0;k<patchNum;k++)
    {
        srcIDMaps[k] = cv::Mat( this->height, this->width, CV_8UC1, cv::Scalar(0) );
        this->srcIDWeightMaps[k] = cv::Mat( this->height, this->width, CV_32FC1, cv::Scalar(0.0f) );
    }
    //unused src ID
    this->srcIDWeightMaps[patchNum] = cv::Mat( this->height, this->width, CV_32FC1, cv::Scalar(0.0f) );
    
    if( !this->isReady() )
        return ;
    
    //find srcID maps
    cv::Mat &imgCollage = this->clgData.patchSet.imgCollage;
    for(int i=0;i<this->height;i++)for(int j=0;j<this->width;j++)
    {
        if( imgCollage.type() == CV_8UC1 )//shadow
        {
            uchar &c = imgCollage.at<uchar>(i,j);
            if( 0==c )
                continue;
        }
        
        if( imgCollage.type() == CV_8UC4 )//object
        {
            cv::Vec4b &c = imgCollage.at<cv::Vec4b>(i,j);
            if( 0==c[3] )
                continue;
        }
        
        int curLabel = this->labels[ i*width+j ];
        srcIDMaps[curLabel].at<uchar>(i,j) = 255;
    }
    
    //find cut maps & calc weight map
    Utils_Ocv &uocv = Utils_Ocv::instance();
    for(int k=0;k<patchNum;k++)
    {
        cv::Mat &srcIDMap = srcIDMaps[k];
        cv::Mat srcIDCutMap;
        cv::Mat &srcIDWeightMap = this->srcIDWeightMaps[k];
        
        srcIDCutMap = uocv.getGradient( srcIDMap ) > 0;
        cv::distanceTransform( 255-srcIDCutMap, srcIDWeightMap, CV_DIST_L2, 3 );
        cv::threshold( srcIDWeightMap, srcIDWeightMap, CLG_tm_alpha, 0, cv::THRESH_TRUNC );
        srcIDWeightMap /= CLG_tm_alpha;
        
        srcIDWeightMap.setTo( cv::Scalar(0.0f), srcIDMap==0 );
    }
}

//get ith src map
void patchCollager::getSrcIDWeightMap(cv::Mat &srcidwmap, int srcID)
{
    cv::vector<int> &patchIDs = this->clgData.patchIDs;
    cv::vector<int>::iterator it = std::find( patchIDs.begin(), patchIDs.end(), srcID );
    
    if( it!=patchIDs.end() )
    {
        long k = std::distance( patchIDs.begin(), it );
        srcidwmap = this->srcIDWeightMaps[k];
    }
    else //srcID can not be found, return a zero mat
        srcidwmap = this->srcIDWeightMaps.back();
}

//blend the collaged images
cv::Mat patchCollager::getBlendedCollage()
{
    imgSet &patchSet = this->clgData.patchSet;
    
    assert( this->isReady() );
    assert( patchSet[0].type()==CV_8UC4 );
    
    int imgNum = (int)patchSet.size();
    
    //prepare masks
    std::vector<cv::Mat> masks(imgNum);
    for(int i=0;i<imgNum;i++)
        masks[i] = cv::Mat::zeros(height, width, CV_8UC1);
    
    for(int i=0;i<num_pixels;i++)
    {
        int rowID = i/width;
        int colID = i%width;
        
        int labelID = this->labels[i];
        
        const cv::Vec4b &c= patchSet[labelID].at<cv::Vec4b>(rowID, colID);
        if( !c[3] )
            continue;
        
        masks[labelID].at<uchar>(rowID,colID) = 255;
    }

    //poisson blending
    Cloning obj;
    cv::Mat blend = patchSet[0].clone();
    for(int i=1;i<imgNum;i++)
        obj.normal_clone( blend, patchSet[i], masks[i], blend, 1);
    
    return blend;
}

//get the warped image which is colored in visible region
cv::Mat patchCollager::getColoredImg(int imgID, float alpha)
{
    cv::Mat coloredImg = clgData.patchSet[imgID].clone();
    int type = clgData.patchSet.imgCollage.type();
    if( type==CV_8UC1 )
        cv::cvtColor( coloredImg, coloredImg, CV_GRAY2BGR );
    
    float beta = 1.0f - alpha;
    
    for(int i=0;i<num_pixels;i++)
    {
        int rowID = i/width;
        int colID = i%width;
        
        if( labels[i] != imgID )
            continue;
        
        const Eigen::Vector3f &color = clgData.colorIDs[ labels[i] ];
        
        if( type==CV_8UC1 )
        {
            const uchar &src = clgData.patchSet[labels[i]].at<uchar>(rowID,colID);
            cv::Vec3b &trg = coloredImg.at<cv::Vec3b>(rowID,colID);
            
            for(int j=0;j<3;j++)
                trg[j] = color(j)*alpha;
            trg[0] += src*beta;
        }
        
        if( type==CV_8UC3 )
        {
            const cv::Vec3b &src = clgData.patchSet[labels[i]].at<cv::Vec3b>(rowID,colID);
            cv::Vec3b &trg = coloredImg.at<cv::Vec3b>(rowID,colID);
            for(int j=0;j<3;j++)
                trg[j] = color(j)*alpha + src[j]*beta;
        }
        
        if( type==CV_8UC4 )
        {
            const cv::Vec4b &src = clgData.patchSet[labels[i]].at<cv::Vec4b>(rowID,colID);
            cv::Vec4b &trg = coloredImg.at<cv::Vec4b>(rowID,colID);
            for(int j=0;j<3;j++)
                trg[j] = color(j)*alpha + src[j]*beta;
            trg[3] = src[3];
        }
        
    }
    
    return coloredImg;
}
