//
//  Plane.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "Plane.h"

Plane::Plane()
{
    
}

Plane::Plane(float a, float b, float c, float d)
{
    this->a = a;
    this->b = b;
    this->c = c;
    this->d = -d; //notice..
}

Plane::~Plane()
{
    
}

const Plane Plane::ground(0.0f,0.0f,1.0f,0.0f);

void Plane::init(const std::vector<float> &params)
{
    if( params.size() != 4 )
        std::cout << "params number is not 4" << std::endl;
    
    a = params[0];
    b = params[1];
    c = params[2];
    d = -params[3]; //notice..
}

//initialize by points using SVD
// 4xn matrix, n = # of pts
//return error
void Plane::init(const Eigen::MatrixXf &pts)
{
    //the SVD of A is the plane parameters a,b,c,d
    Eigen::MatrixXf A;
    A = pts.transpose();
    A.col(3).setConstant(-1);
    
    Eigen::JacobiSVD<Eigen::MatrixXf> svd(A, Eigen::ComputeThinU | Eigen::ComputeThinV);
    Eigen::Vector4f sol = svd.matrixV().col(3);
    
    this->a = sol(0);
    this->b = sol(1);
    this->c = sol(2);
    this->d = sol(3);
}

//calculate the error to the points
void Plane::calcError
(Eigen::VectorXf &error, const Eigen::MatrixXf &pts)const
{
    Eigen::Vector4f sol;
    sol << a, b, c, d;
    
    Eigen::MatrixXf A;
    A = pts.transpose();
    A.col(3).setConstant(-1);
    
    error = A*sol;
}

//invert the plane parameters
void Plane::inverseNormal()
{
    this->a = -this->a;
    this->b = -this->b;
    this->c = -this->c;
    this->d = -this->d;
}


//get the normal of the plane
void Plane::getNormal(Eigen::Vector3f &n)const
{
    n(0) = a;
    n(1) = b;
    n(2) = c;
    n.normalize();
}

//get the intersection of the ray and the plane
void Plane::getIntersection
(Eigen::Vector4f &pt3d,
 const Eigen::Vector2f &pt2d,
 const Eigen::Matrix4f &K,
 const Eigen::Matrix4f &M      )const
{
    //inverse of projection matrix
    Eigen::Matrix4f invP;
    invP = (K*M).inverse();
    
    float t = getDepth(pt2d, invP);
    
    pt3d(0) = pt2d(0) * t;
    pt3d(1) = pt2d(1) * t;
    pt3d(2) = t;
    pt3d(3) = 1.0f;
    
    pt3d = invP * pt3d;
}

void Plane::getIntersection
(Eigen::Vector4f &pt3d,
 const Eigen::Vector4f &p,
 const Eigen::Vector4f &v      )const
{
    float t = ( d - a*p(0) - b*p(1) - c*p(2) )/( a*v(0) + b*v(1) + c*v(2) );
    pt3d = p + t*v;
}

//get the depth t
// X          xt
// Y = invP * yt
// Z           t
// invP.row(3) = [0 0 0 1]
float Plane::getDepth
(const Eigen::Vector2f &pt2d, const  Eigen::Matrix4f &invP )const
{
    float tmpA = ( invP(0,0)*pt2d(0) + invP(0,1)*pt2d(1) + invP(0,2) )*a;
    float tmpB = ( invP(1,0)*pt2d(0) + invP(1,1)*pt2d(1) + invP(1,2) )*b;
    float tmpC = ( invP(2,0)*pt2d(0) + invP(2,1)*pt2d(1) + invP(2,2) )*c;
    
    float t = d - a*invP(0,3) - b*invP(1,3) - c*invP(2,3);
    t /= (tmpA+tmpB+tmpC);
    
    return t;
}

//get the nearest point on the plane
void Plane::getNearestPoint
(Eigen::Vector4f &pt3dn, const Eigen::Vector4f &pt3d)const
{
    float t = a*pt3d(0) + b*pt3d(1) + c*pt3d(2) - d;
    t /= -(a*a+b*b+c*c);
    
    pt3dn(0) = pt3d(0) + a*t;
    pt3dn(1) = pt3d(1) + b*t;
    pt3dn(2) = pt3d(2) + c*t;
    pt3dn(3) = 1.0f;
}

//get the smallest distance to the plane
float Plane::getShortestDist(const Eigen::Vector4f &pt3d)const
{
    float dist = std::abs( a*pt3d(0) + b*pt3d(1) + c*pt3d(2) - d );
    dist /= std::sqrt( a*a + b*b + c*c );
    
    return dist;
}

//pcl plane estimation
bool Plane::estimatePlane
(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud)
{    
    // Segment a plane
    pcl::SACSegmentation<pcl::PointXYZ> seg;
    seg.setOptimizeCoefficients (true);
    seg.setModelType (pcl::SACMODEL_PLANE);
    seg.setMethodType (pcl::SAC_RANSAC);
    seg.setDistanceThreshold (0.01);
    seg.setInputCloud (cloud);
    
    pcl::ModelCoefficients::Ptr coefficients (new pcl::ModelCoefficients);
    pcl::PointIndices::Ptr inliers (new pcl::PointIndices);
    seg.segment (*inliers, *coefficients);
    
    if ( inliers->indices.size() == 0 )
        return false;
    
    // set up the ground
    this->init( coefficients->values );
    
    return true;
}

//get the depth map of the plane
cv::Mat Plane::getDepthMap(const Eigen::Matrix4f &prjM, const cv::Size frameSz)const
{
    Eigen::Matrix4f invP = prjM.inverse();
    cv::Mat depthMap = cv::Mat::zeros(frameSz, CV_32FC1);
    
    for(int i=0;i<frameSz.width;i++)
    for(int j=0;j<frameSz.height;j++)
    {
        Eigen::Vector2f pt2d;
        pt2d << i, j;
        
        float &curDepth = depthMap.at<float>(j,i);
        curDepth = this->getDepth( pt2d, invP );
    }
    
    return depthMap;
}

