//
//  lsqSparseToDenseMap.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "lsqSparseToDenseMap.h"

//get bilinear coefficient
void lsqSparseToDenseMap::getBilinearCoeffs(int &gi, int &gj, float &alpha, float &beta, const cv::Point2f &v)
{
    gi = v.y / gridSz;
    gj = v.x / gridSz;
    
    alpha = ( v.x - gridVs(gi,gj)(0) ) / gridSz;
    beta  = ( v.y - gridVs(gi,gj)(1) ) / gridSz;
}

//check boundary
bool lsqSparseToDenseMap::boundaryCheck(const cv::Point2f &v)
{
    return v.x>=0 && v.x<width && v.y>=0 && v.y<height;
}

//2D to 1D: index mapping
int lsqSparseToDenseMap::V2DtoV1D(int row, int col)
{
    bool testY = row>=0 && row<gridNumY+1;
    bool testX = col>=0 && col<gridNumX+1;
    
    if( testY && testX )
        return row * (gridNumX+1) + col;
    
    return -1;
}

//set matrix for value alignment
void lsqSparseToDenseMap::setValueAlignment
( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int &curRow, float fweight )
{
    int ptNum = (int)vs.size();
    if( ptNum == 0 )
        return;
    
    float w = fweight / ptNum;
    
    // value alignment
    for( int i=0; i<ptNum; i++ )
    {
        int gi, gj;
        float alpha, beta;
        this->getBilinearCoeffs( gi, gj, alpha, beta, vs[i] );
        
        int luGridV = this->V2DtoV1D( gi  , gj   );
        int ldGridV = this->V2DtoV1D( gi+1, gj   );
        int ruGridV = this->V2DtoV1D( gi  , gj+1 );
        int rdGridV = this->V2DtoV1D( gi+1, gj+1 );
        
        float luCoef = ( 1.0f - alpha ) * ( 1.0f - beta );
        float ldCoef = ( 1.0f - alpha ) * beta;
        float ruCoef = alpha * ( 1.0f - beta );
        float rdCoef = alpha * beta;
        
        luCoef *= w;
        ldCoef *= w;
        ruCoef *= w;
        rdCoef *= w;
        
        A_trips.push_back( Eigen::Triplet<float>( curRow, luGridV, luCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, ldGridV, ldCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, ruGridV, ruCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, rdGridV, rdCoef ) );
        b(curRow++) = (float)values[i] * w;
    }
}

//set value similarity
void lsqSparseToDenseMap::setSmlrty
(std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int &curRow, float simWeight )
{
    int gvNum = (gridNumX+1)*(gridNumY+1);
    simWeight /= gvNum;
    
    //for each vertex
    for( int gi=0; gi<gridNumY+1; gi++ ) for( int gj=0; gj<gridNumX+1; gj++ )
    {
        //4-neighbors
        //  0
        //1 4 2
        //  3
        int NS[5];
        NS[0] = this->V2DtoV1D( gi-1,   gj );
        NS[1] = this->V2DtoV1D(   gi, gj-1 );
        NS[2] = this->V2DtoV1D(   gi, gj+1 );
        NS[3] = this->V2DtoV1D( gi+1,   gj );
        NS[4] = this->V2DtoV1D(   gi,   gj );
        
        int nCount = 0;
        for(int i=0;i<4;i++)
            nCount += NS[i]>=0;
        
        float w = simWeight / nCount;
        
        for(int i=0;i<4;i++)
            if( NS[i]>=0 )
                A_trips.push_back( Eigen::Triplet<float>( curRow, NS[i], w ) );
        
        A_trips.push_back( Eigen::Triplet<float>( curRow, NS[4], -simWeight ) );
        b(curRow++) = 0.0f;
    }
}

//solve sparse linear system
float lsqSparseToDenseMap::solve
( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int nrows, int ncols )
{
    // solve sparse linear transform
    Eigen::SparseMatrix<float> A;
    A.resize( nrows, ncols );
    A.setFromTriplets( A_trips.begin(), A_trips.end() );
    
    Eigen::SparseQR< Eigen::SparseMatrix<float>, Eigen::COLAMDOrdering<int> > solver;
    solver.compute(A);
    
    Eigen::VectorXf x(ncols);
    x = solver.solve(b);
    
    float ansDiff = 0.0f;
    
    //copy results
    for( int gi=0; gi<gridNumY+1; gi++ ) for( int gj=0; gj<gridNumX+1; gj++ )
    {
        int curVID = this->V2DtoV1D( gi, gj );
        
        float oriValue = this->gridValues(gi,gj);
        this->gridValues(gi,gj) = x(curVID);
        float aftValue = this->gridValues(gi,gj);
        
        ansDiff += std::abs(oriValue-aftValue);
    }
    
    int vNum = (gridNumY+1) * (gridNumX+1);
    return ansDiff/vNum;
}

//constructor
lsqSparseToDenseMap::lsqSparseToDenseMap(int width, int height, int gridSz)
{
    this->width = width;
    this->height = height;
    this->gridSz = gridSz;
    
    this->gridNumX = (int)std::ceil( (float)width /gridSz );
    this->gridNumY = (int)std::ceil( (float)height/gridSz );
    
    //standard grids
    gridVs.resize( gridNumY+1, gridNumX+1 );
    for(int i=0;i<gridNumY+1;i++) for(int j=0;j<gridNumX+1;j++)
        gridVs(i,j) << gridSz*j, gridSz*i;

    //init value
    this->gridValues.resize( gridNumY+1, gridNumX+1 );
    this->gridValues.setZero();
}

lsqSparseToDenseMap::~lsqSparseToDenseMap()
{
    
}

//add data
void lsqSparseToDenseMap::addValue(const cv::Point2f &v, float value)
{
    if( !boundaryCheck(v) )
        return;
    
    this->vs.push_back(v);
    this->values.push_back(value);
}

//go least square optimization
void lsqSparseToDenseMap::goLSQ(float valueW, float simW)
{
    //num of input values
    int ptNum = (int)this->vs.size();
    
    //num of grid points
    int gridPtNum = (gridNumX+1)*(gridNumY+1);
    
    // the dim of A
    int nrows = ptNum + gridPtNum;
    int ncols = gridPtNum;
    
    //Ax=b
    std::vector< Eigen::Triplet<float> > A_trips;
    Eigen::VectorXf b(nrows);
    int curRow = 0;
    
    if( this->vs.size() == 0 )
    {
        //init value
        this->gridValues.setZero();
        return;
    }
    
    // Value Alignment (row#=ptNum)
    this->setValueAlignment( A_trips, b, curRow, valueW );
    
    // Similarity of Edges (row#=gridEdgeNum)
    this->setSmlrty( A_trips, b, curRow, simW );
    
    // solve sparse linear transform
    this->solve( A_trips, b, nrows, ncols );
}

//get the dense data map
cv::Mat lsqSparseToDenseMap::getDenseMap()
{
    cv::Mat denseMap = cv::Mat( height, width, CV_32FC1, cv::Scalar(0.0f) );
    
    for(int i=0;i<height;i++) for(int j=0;j<width;j++)
    {
        int gi, gj;
        float alpha, beta;
        
        getBilinearCoeffs( gi, gj, alpha, beta, cv::Point2f(j,i) );
        
        const float &luV = gridValues( gi,   gj   );
        const float &ldV = gridValues( gi+1, gj   );
        const float &ruV = gridValues( gi,   gj+1 );
        const float &rdV = gridValues( gi+1, gj+1 );
        
        float luCoef = ( 1.0f - alpha ) * ( 1.0f - beta );
        float ldCoef = ( 1.0f - alpha ) * beta;
        float ruCoef = alpha * ( 1.0f - beta );
        float rdCoef = alpha * beta;
        
        float &value = denseMap.at<float>(i,j);
        value = luV * luCoef + ldV * ldCoef + ruV * ruCoef + rdV * rdCoef;
    }
    
    return denseMap;
}
