//
//  Poly2Tri.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "Poly2Tri.h"

DTIO::DTIO()
{
    
    this->init();
}

DTIO::~DTIO()
{
    this->release();
}

void DTIO::init()
{
    ioData.numberofpoints = 0;
    ioData.numberofpointattributes = 0;
    ioData.numberoftriangles = 0;
    ioData.numberofcorners = 0;
    ioData.numberoftriangleattributes = 0;
    ioData.numberofsegments = 0;
    ioData.numberofholes = 0;
    ioData.numberofregions = 0;
    ioData.numberofedges = 0;
    
    ioData.pointlist = NULL;
    ioData.pointattributelist = NULL;
    ioData.pointmarkerlist = NULL;
    ioData.trianglelist = NULL;
    ioData.triangleattributelist = NULL;
    ioData.trianglearealist = NULL;
    ioData.neighborlist = NULL;
    ioData.segmentlist = NULL;
    ioData.segmentmarkerlist = NULL;
    ioData.holelist = NULL;
    ioData.regionlist = NULL;
    ioData.edgelist = NULL;
    ioData.edgemarkerlist = NULL;
    ioData.normlist = NULL;
}

void DTIO::release()
{
    if( ioData.pointlist )
        free( ioData.pointlist );
    
    if( ioData.pointattributelist )
        free( ioData.pointattributelist );
    
    if( ioData.pointmarkerlist )
        free( ioData.pointmarkerlist );
    
    if( ioData.trianglelist )
        free( ioData.trianglelist );
    
    if( ioData.triangleattributelist )
        free( ioData.triangleattributelist );
    
    if( ioData.trianglearealist )
        free( ioData.trianglearealist );
    
    if( ioData.neighborlist )
        free( ioData.neighborlist );
    
    if( ioData.segmentlist )
        free( ioData.segmentlist );
    
    if( ioData.segmentmarkerlist )
        free( ioData.segmentmarkerlist );
    
    if( ioData.holelist )
        free( ioData.holelist );
    
    if( ioData.regionlist )
        free( ioData.regionlist );
    
    if( ioData.edgelist )
        free( ioData.edgelist );
    
    if( ioData.edgemarkerlist )
        free( ioData.edgemarkerlist );
    
    if( ioData.normlist )
        free( ioData.normlist );

    this->init();
}

std::ostream & operator<< (std::ostream &os, const DTIO& obj)
{
    bool markers = true;
    bool reporttriangles = true;
    bool reportneighbors = true;
    bool reportsegments = true;
    bool reportedges = true;
    bool reportnorms = true;
    
    int i, j;
    
    for (i = 0; i < obj.ioData.numberofpoints; i++) {
        printf("Point %4d:", i);
        for (j = 0; j < 2; j++) {
            printf("  %.6g", obj.ioData.pointlist[i * 2 + j]);
        }
        if (obj.ioData.numberofpointattributes > 0) {
            printf("   attributes");
        }
        for (j = 0; j < obj.ioData.numberofpointattributes; j++) {
            printf("  %.6g",
                   obj.ioData.pointattributelist[i * obj.ioData.numberofpointattributes + j]);
        }
        if (markers) {
            printf("   marker %d\n", obj.ioData.pointmarkerlist[i]);
        } else {
            printf("\n");
        }
    }
    printf("\n");
    
    if (reporttriangles || reportneighbors) {
        for (i = 0; i < obj.ioData.numberoftriangles; i++) {
            if (reporttriangles) {
                printf("Triangle %4d points:", i);
                for (j = 0; j < obj.ioData.numberofcorners; j++) {
                    printf("  %4d", obj.ioData.trianglelist[i * obj.ioData.numberofcorners + j]);
                }
                if (obj.ioData.numberoftriangleattributes > 0) {
                    printf("   attributes");
                }
                for (j = 0; j < obj.ioData.numberoftriangleattributes; j++) {
                    printf("  %.6g", obj.ioData.triangleattributelist[i *
                                                               obj.ioData.numberoftriangleattributes + j]);
                }
                printf("\n");
            }
            if (reportneighbors) {
                printf("Triangle %4d neighbors:", i);
                for (j = 0; j < 3; j++) {
                    printf("  %4d", obj.ioData.neighborlist[i * 3 + j]);
                }
                printf("\n");
            }
        }
        printf("\n");
    }
    
    if (reportsegments) {
        for (i = 0; i < obj.ioData.numberofsegments; i++) {
            printf("Segment %4d points:", i);
            for (j = 0; j < 2; j++) {
                printf("  %4d", obj.ioData.segmentlist[i * 2 + j]);
            }
            if (markers) {
                printf("   marker %d\n", obj.ioData.segmentmarkerlist[i]);
            } else {
                printf("\n");
            }
        }
        printf("\n");
    }
    
    if (reportedges) {
        for (i = 0; i < obj.ioData.numberofedges; i++) {
            printf("Edge %4d points:", i);
            for (j = 0; j < 2; j++) {
                printf("  %4d", obj.ioData.edgelist[i * 2 + j]);
            }
            if (reportnorms && (obj.ioData.edgelist[i * 2 + 1] == -1)) {
                for (j = 0; j < 2; j++) {
                    printf("  %.6g", obj.ioData.normlist[i * 2 + j]);
                }
            }
            if (markers) {
                printf("   marker %d\n", obj.ioData.edgemarkerlist[i]);
            } else {
                printf("\n");
            }
        }
        printf("\n");
    }
    return os;
}

//polygon triangulation
void Poly2Tri::execute
(const std::vector< Eigen::Vector2f > &points, std::vector< Eigen::Vector3i > &triangles)
{
    DTIO in, out;
    
    int ptNum = (int)points.size();
    in.ioData.numberofpoints = ptNum;
    in.ioData.pointlist = (float *) malloc( ptNum * 2 * sizeof(float) );
    for(int i=0;i<ptNum;i++)
    {
        in.ioData.pointlist[2*i] = points[i](0);
        in.ioData.pointlist[2*i+1] = points[i](1);
    }
    
    in.ioData.pointmarkerlist = (int *) malloc( ptNum * sizeof(int) );
    for(int i=0;i<ptNum;i++)
        in.ioData.pointmarkerlist[i] = 1;
    
    in.ioData.numberofsegments = ptNum;
    in.ioData.segmentlist = (int *) malloc( ptNum * 2 * sizeof(int) );
    for(int i=0;i<ptNum;i++)
    {
        in.ioData.segmentlist[2*i] = i;
        in.ioData.segmentlist[2*i+1] = (i+1)%ptNum;
    }
    
    in.ioData.segmentmarkerlist = (int *) malloc( ptNum * sizeof(int) );
    for(int i=0;i<ptNum;i++)
        in.ioData.segmentmarkerlist[i] = 1;
    
    std::string cmd = "Qpzen";
    triangulate((char*)cmd.c_str(), &(in.ioData), &(out.ioData), NULL);
    
    triangles.resize( out.ioData.numberoftriangles );
    for (int i = 0; i < out.ioData.numberoftriangles; i++)
        for (int j = 0; j < out.ioData.numberofcorners; j++)
            triangles[i](j) = out.ioData.trianglelist[ i * out.ioData.numberofcorners + j ];
    
    in.release();
    out.release();
}
