//
//  Utils_Ocv.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "Utils_Ocv.h"

Utils_Ocv Utils_Ocv::inst;

Utils_Ocv::Utils_Ocv()
{
    cv::Mat depmap = cv::Mat::ones(100, 256, CV_8UC1);
    
    for(int j=0;j<256;j++)
        for(int i=0;i<100;i++)
            depmap.at<uchar>(i,j) = j%256;
    
    // Apply the colormap:
    cv::applyColorMap( depmap, colormap, cv::COLORMAP_JET );
}

void Utils_Ocv::getColorBGR ( cv::Vec3b &color, int value, int maxValue, int minValue)
{
    value = std::max( value, minValue );
    value = std::min( value, maxValue );
    
    float colorID = 0.0f;
    if( maxValue > minValue )
        colorID = (float)( value - minValue )*255.0f / ( maxValue - minValue );
    
    color = colormap.at<cv::Vec3b>(0,(int)colorID);
}
void Utils_Ocv::getColorBGR ( Eigen::Vector3f &color, int value, int maxValue, int minValue)
{
    cv::Vec3b color3b;
    getColorBGR( color3b, value, maxValue, minValue );
    color << (float)color3b[0], (float)color3b[1], (float)color3b[2];
}
void Utils_Ocv::getColorRGB ( cv::Vec3b &color, int value, int maxValue, int minValue)
{
    getColorBGR( color, value, maxValue, minValue );
    std::swap(color[0], color[2]);
}
void Utils_Ocv::getColorRGB ( Eigen::Vector3f &color, int value, int maxValue, int minValue)
{
    getColorBGR( color, value, maxValue, minValue );
    std::swap(color[0], color[2]);
}

cv::Mat Utils_Ocv::getGradient(const cv::Mat &img)
{
    /// Generate grad_x and grad_y
    cv::Mat grad_x, grad_y;
    
    /// Gradient
    cv::Sobel( img, grad_x, CV_32F, 1, 0, 3, 1, 0, cv::BORDER_DEFAULT );
    cv::Sobel( img, grad_y, CV_32F, 0, 1, 3, 1, 0, cv::BORDER_DEFAULT );
    cv::Mat grad = 0.5 * ( cv::abs(grad_x) + cv::abs(grad_y) );
    
    return grad;
}

cv::Mat Utils_Ocv::getGradSaliency(const cv::Mat &img)
{
    cv::Mat grad = getGradient(img);
    cv::Mat tmp = grad * 0;
    cv::Mat saliencyMap = sumAbsDiff(tmp, grad);
    return saliencyMap;
}

cv::Mat Utils_Ocv::sumAbsDiff(const cv::Mat &img1, const cv::Mat &img2)
{
    cv::Mat absDiffMat;
    cv::absdiff(img1, img2, absDiffMat);
    
    std::vector<cv::Mat> channels;
    cv::split( absDiffMat, channels );
    
    cv::Mat absDiffSum = cv::Mat::zeros( img1.size(), CV_32FC1 );
    for(int i=0;i<(int)channels.size();i++)
    {
        channels[i].convertTo(channels[i], CV_32F);
        absDiffSum += channels[i];
    }
    
    return absDiffSum;
}

cv::Mat Utils_Ocv::appendAlpha(const cv::Mat &img, const cv::Mat &msk, bool toEraseBackGround)
{
    assert (img.channels()==3);
    assert (msk.channels()==1);

    std::vector<cv::Mat> imgRGBA;
    cv::split( img, imgRGBA );

    cv::Mat maskedImg;
    imgRGBA.push_back(msk);
    
    if( toEraseBackGround )
    for(int i=0;i<3;i++)
        imgRGBA[i].setTo(0, msk==0);

    cv::merge( imgRGBA, maskedImg );
    return maskedImg;
}

//get alpha mask
cv::Mat Utils_Ocv::getAlphaMask(const cv::Mat &img)
{
    assert( img.channels()==4 );
    
    std::vector<cv::Mat> channels;
    cv::split( img, channels );
    return channels[3];
}

void Utils_Ocv::visualizeEnergy(const cv::Mat &img, std::string winName)
{
    cv::Mat tmp = img.clone();
    cv::normalize(tmp, tmp, 0.0, 1.0, cv::NORM_MINMAX);
    cv::imshow(winName, tmp);
    cv::waitKey(1);
}

//convert grayscale image to R/G/B/W
void Utils_Ocv::gray8UC1ToColor8UC4(cv::Mat &colorImg, const cv::Mat &grayImg, const Eigen::Vector4f &bgra)
{
    assert( grayImg.channels()==1 );
    
    std::vector<cv::Mat> chs(4);
    for(int i=0;i<4;i++)
        chs[i] = bgra(i) * grayImg;
    
    cv::merge(chs, colorImg);
}

//get contour map
cv::Mat Utils_Ocv::getContourMap(const cv::Mat &grayImg, float strkSz)
{
    cv::Mat canny_output;
    std::vector<std::vector<cv::Point> > contours;
    std::vector<cv::Vec4i> hierarchy;
    
    /// Detect edges using canny
    int thresh = 100;
    Canny( grayImg, canny_output, thresh, thresh*2, 3 );
    /// Find contours
    findContours( canny_output, contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_SIMPLE, cv::Point(0, 0) );
    
    /// Draw contours
    cv::Mat contourMap = cv::Mat::zeros( canny_output.size(), CV_8UC1 );
    for( int i = 0; i< contours.size(); i++ )
        cv::drawContours( contourMap, contours, i, cv::Scalar(255), strkSz, 8, hierarchy, 0 );
    
    return contourMap;
}

//get bounding box of the fg mask
bool Utils_Ocv::getFGBB(cv::Rect &fgbb, const cv::Mat &mask)
{
    int width = mask.cols;
    int height = mask.rows;
    
    int x_min = width-1;
    int x_max = 0;
    int y_min = height-1;
    int y_max = 0;
    
    for(int i=0;i<height;i++) for(int j=0;j<width;j++)
    {
        if( mask.at<uchar>(i,j)==0 )
            continue;
        
        if( i < y_min ) y_min = i;
        if( i > y_max ) y_max = i;
        
        if( j < x_min ) x_min = j;
        if( j > x_max ) x_max = j;
    }
    
    if( x_max < x_min || y_max < y_min )
        return false;
    
    fgbb = cv::Rect( x_min, y_min, x_max-x_min+1, y_max-y_min+1 );
    return true;
}

//draw arrow
void Utils_Ocv::drawArrow(cv::Mat& img, cv::Point pStart, cv::Point pEnd, int len, int alpha,
               cv::Scalar& color, int thickness, int lineType)
{
    cv::Point arrow;
    double angle = atan2((double)(pStart.y - pEnd.y), (double)(pStart.x - pEnd.x));
    cv::line(img, pStart, pEnd, color, thickness, lineType);
    arrow.x = pEnd.x + len * std::cos(angle + M_PI * alpha / 180.0f);
    arrow.y = pEnd.y + len * std::sin(angle + M_PI * alpha / 180.0f);
    cv::line(img, pEnd, arrow, color, thickness, lineType);
    arrow.x = pEnd.x + len * std::cos(angle - M_PI * alpha / 180.0f);
    arrow.y = pEnd.y + len * std::sin(angle - M_PI * alpha / 180.0f);
    cv::line(img, pEnd, arrow, color, thickness, lineType);
}

//weighted Blurring
cv::Mat Utils_Ocv::weightedGaussianBlur
( const cv::Mat &img, const cv::Mat &bSize, float sigma )
{
    assert( img.type()==CV_8UC1 || img.type()==CV_8UC3 || img.type()==CV_8UC4 );
    assert( bSize.type()==CV_32FC1 );
    assert( bSize.rows == img.rows && bSize.cols == img.cols );
    
    int width = img.cols;
    int height = img.rows;
    cv::Rect imgRegion(0,0,width,height);
    
    cv::Mat blurredImg = img.clone();
    for(int i=0;i<height;i++)for(int j=0;j<width;j++)
    {
        int ksz = std::max( bSize.at<float>(i,j), 0.0f );
        ksz += ksz%2==0;
        int hksz = ksz/2;
        
        cv::Mat kernel1D = cv::getGaussianKernel( ksz, sigma, CV_32F );
        cv::Mat kernel2D = kernel1D * kernel1D.t();
        
        if( img.type()==CV_8UC1 )
        {
            float weight = 0.0f;
            float v = 0;
            for(int gi=-hksz;gi<=hksz;gi++)for(int gj=-hksz;gj<=hksz;gj++)
            {
                if( !imgRegion.contains( cv::Point(j+gj, i+gi)) )
                    continue;
                
                const uchar &src = img.at<uchar>( i+gi, j+gj );
                float gWeight = kernel2D.at<float>(gi+hksz,gj+hksz);
                weight += gWeight;
                v += gWeight * src;
            }
            v = v/weight;
            blurredImg.at<uchar>(i,j) = v;
        }
        else if( img.type()==CV_8UC3 )
        {
            float weight = 0.0f;
            cv::Vec3f v = 0;
            for(int gi=-hksz;gi<=hksz;gi++)for(int gj=-hksz;gj<=hksz;gj++)
            {
                if( !imgRegion.contains( cv::Point(j+gj, i+gi)) )
                    continue;
                
                const cv::Vec3b &src = img.at<cv::Vec3b>( i+gi, j+gj );
                float gWeight = kernel2D.at<float>(gi+hksz,gj+hksz);
                weight += gWeight;
                v += gWeight * src;
            }
            v = v/weight;
            blurredImg.at<cv::Vec3b>(i,j) = v;
        }
        else if( img.type()==CV_8UC4 )
        {
            float weight = 0.0f;
            cv::Vec4f v = 0;
            for(int gi=-hksz;gi<=hksz;gi++)for(int gj=-hksz;gj<=hksz;gj++)
            {
                if( !imgRegion.contains( cv::Point(j+gj, i+gi)) )
                    continue;
                
                const cv::Vec4b &src = img.at<cv::Vec4b>( i+gi, j+gj );
                float gWeight = kernel2D.at<float>(gi+hksz,gj+hksz);
                weight += gWeight;
                v += gWeight * src;
            }
            v = v/weight;
            blurredImg.at<cv::Vec4b>(i,j) = v;
        }
    }
    
    return blurredImg;
}

//horizontally align
cv::Mat Utils_Ocv::hAlignImages(cv::Mat &img1, cv::Mat &img2)
{
    int height = (int)std::max(img1.rows, img2.rows);
    int width  = (int)(img1.cols + img2.cols);
    
    cv::Rect imgRegion1(0,0,img1.cols, img1.rows);
    cv::Rect imgRegion2(img1.cols,0,img2.cols, img2.rows);
    
    cv::Mat out = cv::Mat::zeros(height, width, CV_8UC4);
    
    for(int i=0;i<height;i++)for(int j=0;j<width;j++)
    {
        if( imgRegion1.contains(cv::Point(j,i)) )
            out.at<cv::Vec4b>(i,j) = img1.at<cv::Vec4b>(i,j);
        if( imgRegion2.contains(cv::Point(j,i)) )
            out.at<cv::Vec4b>(i,j) = img2.at<cv::Vec4b>(i,j-img1.cols);
    }
    
    return out;
}

//blend color with the image
void Utils_Ocv::blendColor(cv::Mat &img, const cv::Vec3b &color, float alpha)
{
    assert( img.type()==CV_8UC3 || img.type()==CV_8UC4 );
    assert( alpha >= 0.0f && alpha <= 1.0f );
    
    float beta = 1.0f - alpha;
    
    for(int i=0;i<img.rows;i++) for(int j=0;j<img.cols;j++)
    {
        if( img.channels()==3 )
        {
            cv::Vec3b &src = img.at<cv::Vec3b>(i,j);
            for(int j=0;j<3;j++)
                src[j] = color[j]*alpha + src[j]*beta;
        }
        else if( img.channels()==4 )
        {
            cv::Vec4b &src = img.at<cv::Vec4b>(i,j);
            for(int j=0;j<3;j++)
                src[j] = color[j]*alpha + src[j]*beta;
        }
    }
}

//shed color on the image
void Utils_Ocv::shedColor(cv::Mat &img, const cv::Vec3b &color)
{
    assert( img.type()==CV_8UC3 || img.type()==CV_8UC4 );
    
    Eigen::Vector3f colorf;
    colorf << (float)color[0],(float)color[1],(float)color[2];
    colorf = colorf / 255.0f;
    
    for(int i=0;i<img.rows;i++) for(int j=0;j<img.cols;j++)
    {
        if( img.channels()==3 )
        {
            cv::Vec3b &src = img.at<cv::Vec3b>(i,j);
            for(int j=0;j<3;j++)
                src[j] = colorf(j) * src[j];
        }
        else if( img.channels()==4 )
        {
            cv::Vec4b &src = img.at<cv::Vec4b>(i,j);
            for(int j=0;j<3;j++)
                src[j] = colorf(j) * src[j];
        }
    }
}

//255-rgb
void Utils_Ocv::invertRGB(cv::Mat &img)
{
    assert( img.type()==CV_8UC4 );
    
    std::vector<cv::Mat> cs;
    cv::split(img, cs);
    
    for(int i=0;i<3;i++)
        cs[i] = 255 - cs[i];
    
    cv::merge(cs, img);
}

//draw importance map
void Utils_Ocv::drawImpWeight(cv::Mat &importanceMap, const cv::Point2f &pt, int hksz)
{
    assert( importanceMap.type() == CV_32FC1 );
    
    int width = importanceMap.cols;
    int height = importanceMap.rows;
    
    int ksz = hksz*2+1;
    float sigma = 0.3*((ksz-1)*0.5 - 1) + 0.8; //default sigma as getGaussianKernel
    cv::Mat kernel1D = cv::Mat( ksz, 1, CV_32FC1, cv::Scalar(0.0f) );
    for(int i=0;i<ksz;i++)
        kernel1D.at<float>(i,0) = -( (i-hksz)*(i-hksz) ) / ( 2 * sigma * sigma );
    cv::exp( kernel1D, kernel1D );
    cv::Mat kernel2D = kernel1D * kernel1D.t();
    
    for(int i=0;i<ksz;i++)for(int j=0;j<ksz;j++)
    {
        int ptx = pt.x + j - hksz;
        int pty = pt.y + i - hksz;
        
        if( ptx >= 0 && ptx < width && pty >= 0 && pty < height )
        {
            float w = kernel2D.at<float>(i,j);
            float &imp = importanceMap.at<float>(pty, ptx);
            imp = std::max( w, imp );
        }
    }
}

inline void makecolorwheel(std::vector<cv::Scalar> &colorwheel)
{
    int RY = 15;
    int YG = 6;
    int GC = 4;
    int CB = 11;
    int BM = 13;
    int MR = 6;
    
    int i;
    
    for (i = 0; i < RY; i++) colorwheel.push_back(cv::Scalar(255,	   255*i/RY,	 0));
    for (i = 0; i < YG; i++) colorwheel.push_back(cv::Scalar(255-255*i/YG, 255,		 0));
    for (i = 0; i < GC; i++) colorwheel.push_back(cv::Scalar(0,		   255,		 255*i/GC));
    for (i = 0; i < CB; i++) colorwheel.push_back(cv::Scalar(0,		   255-255*i/CB, 255));
    for (i = 0; i < BM; i++) colorwheel.push_back(cv::Scalar(255*i/BM,	   0,		 255));
    for (i = 0; i < MR; i++) colorwheel.push_back(cv::Scalar(255,	   0,		 255-255*i/MR));
}

//used to visualize motion field
void Utils_Ocv::motionToColor(cv::Mat flow, cv::Mat &color)
{
    #define UNKNOWN_FLOW_THRESH 1e9
    
    if (color.empty())
        color.create(flow.rows, flow.cols, CV_8UC3);
    
    static std::vector<cv::Scalar> colorwheel; //Scalar r,g,b
    if (colorwheel.empty())
        makecolorwheel(colorwheel);
    
    // determine motion range:
    float maxrad = -1;
    
    // Find max flow to normalize fx and fy
    for (int i= 0; i < flow.rows; ++i)
    {
        for (int j = 0; j < flow.cols; ++j)
        {
            cv::Vec2f flow_at_point = flow.at<cv::Vec2f>(i, j);
            float fx = flow_at_point[0];
            float fy = flow_at_point[1];
            if ((fabs(fx) >  UNKNOWN_FLOW_THRESH) || (fabs(fy) >  UNKNOWN_FLOW_THRESH))
                continue;
            float rad = sqrt(fx * fx + fy * fy);
            maxrad = maxrad > rad ? maxrad : rad;
        }
    }
    
    for (int i= 0; i < flow.rows; ++i)
    {
        for (int j = 0; j < flow.cols; ++j)
        {
            uchar *data = color.data + color.step[0] * i + color.step[1] * j;
            cv::Vec2f flow_at_point = flow.at<cv::Vec2f>(i, j);
            
            float fx = flow_at_point[0] / maxrad;
            float fy = flow_at_point[1] / maxrad;
            if ((fabs(fx) >  UNKNOWN_FLOW_THRESH) || (fabs(fy) >  UNKNOWN_FLOW_THRESH))
            {
                data[0] = data[1] = data[2] = 0;
                continue;
            }
            float rad = sqrt(fx * fx + fy * fy);
            
            float angle = atan2(-fy, -fx) / CV_PI;
            float fk = (angle + 1.0) / 2.0 * (colorwheel.size()-1);
            int k0 = (int)fk;
            int k1 = (k0 + 1) % colorwheel.size();
            float f = fk - k0;
            //f = 0; // uncomment to see original color wheel
            
            for (int b = 0; b < 3; b++) 
            {
                float col0 = colorwheel[k0][b] / 255.0;
                float col1 = colorwheel[k1][b] / 255.0;
                float col = (1 - f) * col0 + f * col1;
                if (rad <= 1)
                    col = 1 - rad * (1 - col); // increase saturation with radius
                else
                    col *= .75; // out of range
                data[2 - b] = (int)(255.0 * col);
            }
        }
    }
}

cv::Scalar Utils_Color::WHITE = cv::Scalar(255,255,255,255);
cv::Scalar Utils_Color::BLACK = cv::Scalar(000,000,000,255);
cv::Scalar Utils_Color::GRAY  = cv::Scalar(128,128,128,255);
cv::Scalar Utils_Color::RED   = cv::Scalar(000,000,255,255);
cv::Scalar Utils_Color::GREEN = cv::Scalar(000,255,000,255);
cv::Scalar Utils_Color::BLUE  = cv::Scalar(255,000,000,255);

cv::Scalar Utils_Color::WHITE_A = cv::Scalar(255,255,255,128);
cv::Scalar Utils_Color::BLACK_A = cv::Scalar(000,000,000,128);
cv::Scalar Utils_Color::GRAY_A  = cv::Scalar(128,128,128,128);
cv::Scalar Utils_Color::RED_A   = cv::Scalar(000,000,255,128);
cv::Scalar Utils_Color::GREEN_A = cv::Scalar(000,255,000,128);
cv::Scalar Utils_Color::BLUE_A  = cv::Scalar(255,000,000,128);


VideoOutput::VideoOutput( std::string filename, cv::Size frameSz, double fps){
    this->fps     = fps;
    this->frameSz = frameSz;
    int fourcc = CV_FOURCC('m', 'p', '4', 'v');
    this->outputVideo.open( filename.c_str(), fourcc, this->fps, this->frameSz);
}

VideoOutput::~VideoOutput(){}

void VideoOutput::writeFrame(cv::Mat &current_frame){
    this->outputVideo << current_frame;
}

void VideoOutput::finishWriting(){
    this->outputVideo.release();
}