//
//  Utils_Ogl.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "Utils_Ogl.h"

//=================================================
// private auxiliary functions
//=================================================
//draw a quad between -1 and 1 using texture imageID
void drawQuad(GLuint imageID)
{
    if(imageID==0){
        std::cout << "texture id is zero." << std::endl;
        return;
    }
    glBindTexture(GL_TEXTURE_2D,imageID);
    
    glBegin(GL_QUADS);
    glTexCoord2f(0.f,1.f);
    glVertex3f(-1.0f,-1.0f,0.0f);
    glTexCoord2f(1.f,1.f);
    glVertex3f(1.0f,-1.0f,0.0f);
    glTexCoord2f(1.f,0.0f);
    glVertex3f(1.0f,1.0f,0.0f);
    glTexCoord2f(0.f,0.0f);
    glVertex3f(-1.0f,1.0f,0.0f);
    glEnd();
    
    glBindTexture(GL_TEXTURE_2D,0);
}

//=================================================
// interface functions
//=================================================
//convert Mat to openGL texture ID
GLuint Utils_Ogl::matToGL(const cv::Mat &img)
{
    GLuint textID;
    
    glGenTextures( 1 , &textID );
    glBindTexture( GL_TEXTURE_2D, textID );
    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_MIN_FILTER , GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_MAG_FILTER , GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_WRAP_S     , GL_CLAMP_TO_BORDER  );
    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_WRAP_T     , GL_CLAMP_TO_BORDER  );
    
    int nchannels = img.channels();
    
    switch (nchannels) {
        case 1:
            glTexImage2D( GL_TEXTURE_2D, 0 , nchannels , img.cols, img.rows, 0 ,
                         GL_LUMINANCE , GL_UNSIGNED_BYTE , img.ptr() );
            break;
            
        case 3:
            glTexImage2D( GL_TEXTURE_2D, 0 , nchannels , img.cols, img.rows, 0 ,
                         GL_BGR , GL_UNSIGNED_BYTE , img.ptr() );
            break;
            
        case 4:
            glTexImage2D( GL_TEXTURE_2D, 0 , nchannels , img.cols, img.rows, 0 ,
                         GL_BGRA , GL_UNSIGNED_BYTE , img.ptr() );
            break;
            
        default:
            break;
    }
    
    return textID;
}

//init function
void Utils_Ogl::InitGL()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 2.0);
    
    glPointSize(5);
    glLineWidth(1.0f);
    
    glDisable(GL_DITHER);
}

//render a Mat in openGL
void Utils_Ogl::renderMat(const cv::Mat &img)
{
    if(img.empty()){
        std::cout << "Mat is empty, and it cannot be rendered."<< std::endl;
        return;
    }
    
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    
    GLuint imageID = matToGL( img );
    drawQuad(imageID);
    glDeleteTextures(1,&imageID);
}

//read color buffer
cv::Size Utils_Ogl::getViewPortSize()
{
    GLint vp [4];
    glGetIntegerv (GL_VIEWPORT, vp);
    cv::Size sz;
    sz.width  = vp[2];
    sz.height = vp[3];
    return sz;
}
cv::Mat Utils_Ogl::readColorBuffer(int ltx,int lty,int w,int h)
{
    cv::Mat fbFrame = cv::Mat::zeros( h, w, CV_8UC4 );
    
    glReadBuffer( GL_BACK );
    glReadPixels(ltx, lty, w, h, GL_BGRA, GL_UNSIGNED_BYTE, fbFrame.data );
    
    cv::Mat fliped;
    cv::flip( fbFrame, fliped, 0 );
    
    return fliped;
}
cv::Mat Utils_Ogl::readWholeBuffer()
{
    cv::Size sz = getViewPortSize();
    return readColorBuffer(0,0,sz.width,sz.height);
}

//set/reset projection & modelview by K & M & W & H
void Utils_Ogl::setMatrix(int W, int H,
               const Eigen::Matrix4f &K,
               const Eigen::Matrix4f &M )
{
    //intrinsics
    const float alpha = K(0,0);
    const float beta  = K(1,1);
    const float skew  = K(0,1);
    const float u0    = K(0,2);
    const float v0    = K(1,2);
    
    //params of normalized device coordinates (NDC)
    const float L = 0.0f;
    const float R = (float)W;
    const float B = (float)H;
    const float T = 0.0f;
    const float N = 0.0001f;
    const float F = 1000.0f;
    
    // init intrinsics matrix for openGL
    Eigen::Matrix4f tproj = Eigen::Matrix4f::Zero();
    tproj <<
    alpha, skew,    u0,    0.0f,
    0.0f,  beta,    v0,    0.0f,
    0.0f,  0.0f,  -N-F,     N*F,
    0.0f,  0.0f,  1.0f,    0.0f;
    
    //rendering
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    //NDC -> OGL can render the stuff on the viewport
    glOrtho(L,R,B,T,N,F);
    //projection matrix (perserve the z-value)
    glMultMatrixf( tproj.data() );
    
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    //camera motion
    glMultMatrixf( M.data() );
}

//set/reset projection & modelview to Identity
void Utils_Ogl::setMatrixIdentity()
{
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
}

void Utils_Ogl::resetMatrix()
{
    glPopMatrix();
    
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    
    glMatrixMode(GL_MODELVIEW);
}

//draw x,y,z axis
void Utils_Ogl::drawAxis()
{
    glPushAttrib( GL_CURRENT_BIT | GL_POLYGON_BIT | GL_CURRENT_BIT );
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    
    glBegin(GL_LINES);
    glColor4f(1.0f,0.0f,0.0f,1.0f);
    glVertex3d(0.0,0.0,0.0);
    glVertex3d(1.0,0.0,0.0);
    glColor4f(0.0f,1.0f,0.0f,1.0f);
    glVertex3d(0.0,0.0,0.0);
    glVertex3d(0.0,1.0,0.0);
    glColor4f(0.0f,0.0f,1.0f,1.0f);
    glVertex3d(0.0,0.0,0.0);
    glVertex3d(0.0,0.0,1.0);
    glEnd();
    
    glBegin(GL_POINTS);
    glColor4f(0.0f,0.0f,0.0f,1.0f);
    glVertex3d(0.0,0.0,0.0);
    
    glColor4f(1.0f,0.0f,0.0f,1.0f);
    glVertex3d(1.0,0.0,0.0);
    
    glColor4f(0.0f,1.0f,0.0f,1.0f);
    glVertex3d(0.0,1.0,0.0);
    
    glColor4f(0.0f,0.0f,1.0f,1.0f);
    glVertex3d(0.0,0.0,1.0);
    
    glEnd();
    
    
    glPopAttrib();
}

//draw standard cube
void Utils_Ogl::drawCube()
{
    static const float ver[8][3] =
    {
        { -1.0f, -1.0f, 1.0f },
        { -1.0f,  1.0f, 1.0f },
        {  1.0f,  1.0f, 1.0f },
        {  1.0f, -1.0f, 1.0f },
        { -1.0f, -1.0f,-1.0f },
        { -1.0f,  1.0f,-1.0f },
        {  1.0f,  1.0f,-1.0f },
        {  1.0f, -1.0f,-1.0f }
    };
    
    static const GLint order[6][4] =
    {
        {0,3,2,1},
        {2,3,7,6},
        {0,4,7,3},
        {1,2,6,5},
        {4,5,6,7},
        {0,1,5,4}
    };

    glBegin(GL_QUADS);
    for(int i=0;i<6;i++)
        for(int j=0;j<4;j++)
            glVertex3fv( ver[order[i][j]] );
    glEnd();
}

//draw pyramid
void Utils_Ogl::drawPyramid()
{
    static const float ver[5][3] =
    {
        {  0.0f,  0.0f, 0.0f },
        { -1.0f, -1.0f, 1.0f },
        { -1.0f,  1.0f, 1.0f },
        {  1.0f,  1.0f, 1.0f },
        {  1.0f, -1.0f, 1.0f }
    };
    
    static const GLint order[4][3] =
    {
        {2,0,1},
        {1,0,4},
        {4,0,3},
        {3,0,2}
    };
    
    glBegin(GL_TRIANGLES);
    for(int i=0;i<4;i++)
        for(int j=0;j<3;j++)
            glVertex3fv( ver[order[i][j]] );
    glEnd();
}

//draw standard sphere
void Utils_Ogl::drawSphere( int slices, int stacks)
{   
    for( int i=0; i<stacks; i++ )
    {
        float stack0 = M_PI * (float)i / stacks;
        float r0 = std::sin(stack0);
        float z0 = std::cos(stack0);
        
        float stack1 = M_PI * (float)(i+1) / stacks;
        float r1 = std::sin(stack1);
        float z1 = std::cos(stack1);
        
        glBegin(GL_QUADS);
        for( int j=0; j<slices; j++ )
        {
            float slice0 = 2.0f * M_PI * (float)j / slices;
            float sx0 = std::cos(slice0);
            float sy0 = std::sin(slice0);
            
            float slice1 = 2.0f * M_PI * (float)(j+1) / slices;
            float sx1 = std::cos(slice1);
            float sy1 = std::sin(slice1);
            
            glVertex3f( r0 * sx0, r0 * sy0, z0 );
            glVertex3f( r1 * sx0, r1 * sy0, z1 );
            glVertex3f( r1 * sx1, r1 * sy1, z1 );
            glVertex3f( r0 * sx1, r0 * sy1, z0 );
        }
        glEnd();
    }
}

//draw XY plane
void Utils_Ogl::drawGridXY(int xw, int yw)
{
    glPushAttrib( GL_POLYGON_BIT | GL_CURRENT_BIT );
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBegin(GL_QUADS);
    
    for(int x=-xw;x<xw;x++)
    {
        float xc1 = (float)x/xw;
        float xc2 = (float)(x+1)/xw;
        
        for(int y=-yw;y<yw;y++)
        {
            float yc1 = (float)y/yw;
            float yc2 = (float)(y+1)/yw;
    
            glVertex3f(xc1,yc1,0.0f);
            glVertex3f(xc2,yc1,0.0f);
            glVertex3f(xc2,yc2,0.0f);
            glVertex3f(xc1,yc2,0.0f);
        }
    }
    
    glEnd();
    glPopAttrib();
}

inline void renderColoredGridXY(int xw, int yw, float alpha, float r, float g, float b, float diff=1.0f)
{
    glPushAttrib( GL_CURRENT_BIT );
    glBegin(GL_QUADS);
    
    for(int x=-xw;x<xw;x++)
    {
        float xc1 = (float)x/xw;
        float xc2 = (float)(x+1)/xw;
        
        for(int y=-yw;y<yw;y++)
        {
            //cornflower color (0.392, 0.584, 0.929)
            float c = 1.0f + diff * (std::abs(x+y)%2);
            glColor4f( r/c, g/c, b/c, alpha );
            
            float yc1 = (float)y/yw;
            float yc2 = (float)(y+1)/yw;
            
            glVertex3f(xc1,yc1,0.0f);
            glVertex3f(xc2,yc1,0.0f);
            glVertex3f(xc2,yc2,0.0f);
            glVertex3f(xc1,yc2,0.0f);
        }
    }
    
    glEnd();
    
    glColor4f( r/1.5f, g/1.5f, b/1.5f, alpha );
    Utils_Ogl::drawGridXY(xw,yw);
    
    glPopAttrib();
}

void Utils_Ogl::blueGridXY(int xw, int yw, float alpha)
{
    renderColoredGridXY(xw, yw, alpha, 0.392f, 0.584f, 0.929f);
}
void Utils_Ogl::whiteGridXY(int xw, int yw, float alpha)
{
    renderColoredGridXY(xw, yw, alpha, 0.7882f, .8156f, .8431f, 0.1f);
}
void Utils_Ogl::redGridXY(int xw, int yw, float alpha)
{
    renderColoredGridXY(xw, yw, alpha, 0.929f, 0.284f, 0.192f);
}



//draw viewport border
void Utils_Ogl::drawViewportBorder(float width, const Eigen::Vector3f &color)
{
    Utils_Ogl::setMatrixIdentity();
    
    glPushAttrib( GL_CURRENT_BIT | GL_DEPTH_BUFFER_BIT | GL_POLYGON_BIT | GL_LINE_BIT );
    
    glDisable(GL_DEPTH_TEST);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glColor3f( 1.0f, 0.0f, 0.0f );
    glLineWidth(10);
    
    glBegin(GL_QUADS);
    glVertex3f(-1.0f,-1.0f,0.0f);
    glVertex3f(1.0f,-1.0f,0.0f);
    glVertex3f(1.0f,1.0f,0.0f);
    glVertex3f(-1.0f,1.0f,0.0f);
    glEnd();
    
    glPopAttrib();
    
    Utils_Ogl::resetMatrix();
}

//draw background color
void Utils_Ogl::drawBackgroundColor(float r, float g, float b)
{
    Utils_Ogl::setMatrixIdentity();
    
    glPushAttrib( GL_CURRENT_BIT | GL_DEPTH_BUFFER_BIT );
    glDisable(GL_DEPTH_TEST);
    
    glColor3f(r,g,b);
    glBegin(GL_QUADS);
    glVertex3f(-1.0f,-1.0f,0.0f);
    glVertex3f(1.0f,-1.0f,0.0f);
    glVertex3f(1.0f,1.0f,0.0f);
    glVertex3f(-1.0f,1.0f,0.0f);
    glEnd();

    glPopAttrib();
    
    Utils_Ogl::resetMatrix();
}


//is FBO supported
bool Utils_Ogl::FBO::supportFBO()
{
    const GLubyte * strExt = glGetString (GL_EXTENSIONS);
    GLboolean support = gluCheckExtension ((const GLubyte*)"GL_EXT_framebuffer_object",strExt); // 5
    return support;
}

//constructor
Utils_Ogl::FBO::FBO()
{
    this->isBuffersReady = false;
}

bool Utils_Ogl::FBO::start(int width, int height)
{
    if( !Utils_Ogl::FBO::supportFBO() )
    {
        std::cout << "FBO is not supported" << std::endl;
        return false;
    }
    
    if( this->isBuffersReady )
        return true;
    
    this->width = width;
    this->height = height;
    
    //frame buffer
    glGenFramebuffersEXT(1, &this->frameBuffer);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, this->frameBuffer);
    
    //render buffer
    glGenRenderbuffersEXT(1, &this->colorRenderBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, this->colorRenderBuffer);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA8, this->width, this->height);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_RENDERBUFFER_EXT, this->colorRenderBuffer);
    
    //depth buffer
    glGenRenderbuffersEXT(1, &this->depthRenderBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, this->depthRenderBuffer);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, this->width, this->height);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, this->depthRenderBuffer);
    
    GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    this->isBuffersReady = status == GL_FRAMEBUFFER_COMPLETE_EXT;
    
    return this->isBuffersReady;
}

//dump frame buffer in fbo
cv::Mat Utils_Ogl::FBO::readFrameBuffer(int channelNum)
{
    assert(this->isBuffersReady);
    
    cv::Mat frame = Utils_Ogl::readWholeBuffer();
    
    if( channelNum == 1 )
        cv::cvtColor( frame, frame, CV_RGBA2GRAY, channelNum );
    if( channelNum == 3 )
        cv::cvtColor( frame, frame, CV_RGBA2RGB, channelNum );
    
    return frame;
}

//write the frame to an image
void Utils_Ogl::FBO::writeFrameBuffer(std::string imgPath, int channelNum)
{
    assert(this->isBuffersReady);
    cv::Mat frame = this->readFrameBuffer(channelNum);
    cv::imwrite( imgPath, frame );
}

void Utils_Ogl::FBO::end()
{
    if( !this->isBuffersReady )
        return;
    
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
    glDeleteRenderbuffersEXT(1, &this->colorRenderBuffer);
    glDeleteRenderbuffersEXT(1, &this->depthRenderBuffer);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glDeleteFramebuffersEXT(1, &this->frameBuffer);
    this->isBuffersReady = false;
}









