//
//  CoGEPWarper.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "CoGEPWarper.h"

//clear all data
void CoGEPWarper::clear()
{
    this->cics.clear();
    this->gepws.clear();
    this->gepwWeights.clear();
}

//add gep warps
void CoGEPWarper::addGepWarp( GEPWarper &gepw, float gepwWeight )
{
    assert( gepwWeight > 0.0f );
    
    this->gepws.push_back(gepw);
    this->gepwWeights.push_back(gepwWeight);
}

//cross-image positional constraints
void CoGEPWarper::addCoPointConstraint
(int i, const cv::Point2f &vi, int j, const cv::Point2f &vj, float pWeight)
{
    CrossImgCrr cic;
    cic.i = i;
    cic.j = j;
    cic.vi = vi;
    cic.vj = vj;
    cic.pWeight = pWeight;
    
    this->cics.push_back(cic);
}


Eigen::VectorXf CoGEPWarper::optimizeForWarps
(float ftrWeight, float simWeight, float edgeWeight, float coWeight, float btmWeight)
{
    const int warpNum = (int)this->gepws.size();
    
    //normalize warp weights
    Eigen::VectorXf nmlzdGepWs;
    nmlzdGepWs.resize(warpNum);
    for(int i=0;i<warpNum;i++)
        nmlzdGepWs(i) = this->gepwWeights[i];
    nmlzdGepWs /= nmlzdGepWs.sum();
    
    //normalize shared point weights
    Eigen::VectorXf nmlzdSPWs;
    const int cicNum = (int)this->cics.size();
    nmlzdSPWs.resize(cicNum);
    for(int i=0;i<cicNum;i++)
        nmlzdSPWs(i) = this->cics[i].pWeight;
    nmlzdSPWs /= nmlzdSPWs.sum();
    
    //error values for each warp
    Eigen::VectorXf warpErrors(warpNum);


    //calculate the parameters required by the warping
    //====================================================================================
    std::vector<int> nrows(warpNum), ncols(warpNum), epNums(warpNum);
    for(int i=0;i<warpNum;i++)
    {
        GEPWarper &curWarper = this->at(i);
        
        //point crr #
        int ptNum = (int)curWarper.vFroms.size() + (int)curWarper.vFromsHarder.size();
        
        //num of grid points
        int gridPtNum = (curWarper.gridNumX+1)*(curWarper.gridNumY+1);
        
        //num of grid triangles(each quad contains 4 triangles)
        int gridTriNum = 4 * curWarper.gridNumX * curWarper.gridNumY;
        
        //number of edge points
        int sumEdgePtNum = 0;
        int edgeNum = (int)curWarper.lines.size();
        for(int j=0;j<edgeNum;j++)
            sumEdgePtNum += (int)curWarper.lines[j].evs_from.size();
        epNums[i] = sumEdgePtNum;
        
        // the dim of A
        nrows[i] = 2 * ( ptNum + gridTriNum ) + sumEdgePtNum;
        ncols[i] = 2 * gridPtNum;
    }
    
    std::vector<int> start_nrows(warpNum,0);
    std::vector<int> start_ncols(warpNum,0);
    for(int i=1;i<warpNum;i++)
    {
        start_nrows[i] += nrows[i-1] + start_nrows[i-1];
        start_ncols[i] += ncols[i-1] + start_ncols[i-1];
    }
    
    //the size of CoMatrix
    int totalCols = std::accumulate( ncols.begin(), ncols.end(), 0 );
    int totalRows = std::accumulate( nrows.begin(), nrows.end(), 0 );
    totalRows += 2*(int)cics.size();
    //====================================================================================
    
    //prepare the sparse linear system
    std::vector< Eigen::Triplet<double> > A_trips_total;
    Eigen::VectorXd b_total( totalRows );
    
    //setup warps
    //====================================================================================
    //collect the matrix in each warper
    for(int i=0;i<warpNum;i++)
    {
        std::vector< Eigen::Triplet<float> > A_trips;
        Eigen::VectorXf b(nrows[i]);
        int curRow = 0;
        
        GEPWarper &curWarper = this->at(i);
        
        // Feature Alignment (row#=ptNum)
        curWarper.setPtCrrspndnc( A_trips, b, curRow, ftrWeight * nmlzdGepWs(i) );
        curWarper.setHarderPtCrrspndnc( A_trips, b, curRow, btmWeight * nmlzdGepWs(i) );
        
        // Similarity of Quads (row#=gridTriNum)
        curWarper.setSmlrty( A_trips, b, curRow, simWeight * nmlzdGepWs(i) );
        
        // Edge Straightness (row#=edgePtNum)
        curWarper.setStrghtnss( A_trips, b, curRow, edgeWeight * nmlzdGepWs(i), epNums[i] );
        
        //copy data - Sparse Matrix A
        int tripNum = (int)A_trips.size();
        for(int j=0;j<tripNum;j++)
        {
            //copy the matix to the big matrix
            int row = A_trips[j].row() + start_nrows[i];
            int col = A_trips[j].col() + start_ncols[i];
            float v = A_trips[j].value();
            
            Eigen::Triplet<double> trip( row, col, (double)v );
            A_trips_total.push_back( trip );
        }
        
        //copy data - Vector b
        for(int j=0;j<nrows[i];j++)
            b_total( j + start_nrows[i] ) = (double) b( j );
    }
    //====================================================================================
    
    
    //setup copoints
    //====================================================================================
    //set co-image positional constraint
    for(int i=0;i<cicNum;i++)
    {
        int curRow = totalRows - 2*cicNum + 2*i;
        
        const CrossImgCrr &cic = cics[i];
        float w = coWeight * nmlzdSPWs(i);
        
        int igi, igj;
        float ialpha, ibeta;
        const GEPWarper &warperi = this->at(cic.i);
        warperi.getBilinearCoeffs( igi, igj, ialpha, ibeta, cic.vi );
        
        int vidi_lu = 2 * warperi.V2DtoV1D(igi  , igj  ) + start_ncols[cic.i];
        int vidi_ld = 2 * warperi.V2DtoV1D(igi+1, igj  ) + start_ncols[cic.i];
        int vidi_ru = 2 * warperi.V2DtoV1D(igi  , igj+1) + start_ncols[cic.i];
        int vidi_rd = 2 * warperi.V2DtoV1D(igi+1, igj+1) + start_ncols[cic.i];
        
        float luCoefi = w * ( 1.0f - ialpha ) * ( 1.0f - ibeta );
        float ldCoefi = w * ( 1.0f - ialpha ) * ibeta;
        float ruCoefi = w * ialpha * ( 1.0f - ibeta );
        float rdCoefi = w * ialpha * ibeta;
        
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidi_lu, luCoefi ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidi_ld, ldCoefi ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidi_ru, ruCoefi ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidi_rd, rdCoefi ) );
        
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidi_lu+1, luCoefi ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidi_ld+1, ldCoefi ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidi_ru+1, ruCoefi ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidi_rd+1, rdCoefi ) );
        
        int jgi, jgj;
        float jalpha, jbeta;
        const GEPWarper &warperj = this->at(cic.j);
        warperj.getBilinearCoeffs( jgi, jgj, jalpha, jbeta, cic.vj );
        
        int vidj_lu = 2 * warperj.V2DtoV1D(jgi  , jgj  ) + start_ncols[cic.j];
        int vidj_ld = 2 * warperj.V2DtoV1D(jgi+1, jgj  ) + start_ncols[cic.j];
        int vidj_ru = 2 * warperj.V2DtoV1D(jgi  , jgj+1) + start_ncols[cic.j];
        int vidj_rd = 2 * warperj.V2DtoV1D(jgi+1, jgj+1) + start_ncols[cic.j];
        
        float luCoefj = w * ( 1.0f - jalpha ) * ( 1.0f - jbeta );
        float ldCoefj = w * ( 1.0f - jalpha ) * jbeta;
        float ruCoefj = w * jalpha * ( 1.0f - jbeta );
        float rdCoefj = w * jalpha * jbeta;
        
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidj_lu, -luCoefj ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidj_ld, -ldCoefj ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidj_ru, -ruCoefj ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow, vidj_rd, -rdCoefj ) );
        
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidj_lu+1, -luCoefj ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidj_ld+1, -ldCoefj ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidj_ru+1, -ruCoefj ) );
        A_trips_total.push_back( Eigen::Triplet<double>( curRow+1, vidj_rd+1, -rdCoefj ) );
        
        b_total(curRow) = b_total(curRow+1) = 0.0f;
    }
    //====================================================================================
    
    //solve
    // solve sparse linear transform
    Eigen::SparseMatrix<double> A;
    A.resize( totalRows, totalCols );
    A.setFromTriplets( A_trips_total.begin(), A_trips_total.end() );
    
    Eigen::SparseQR< Eigen::SparseMatrix<double>, Eigen::COLAMDOrdering<int> > solver;
    solver.compute(A);
    
    Eigen::VectorXd x(totalCols);
    x = solver.solve(b_total);
    
    Eigen::VectorXd rowErrors = A * x - b_total;
    rowErrors = rowErrors.cwiseAbs();
    
    //copy optimized results
    float ansDiff = 0.0f;
    for(int i=0;i<warpNum;i++)
    {
        warpErrors(i) = rowErrors.block( start_nrows[i], 0, nrows[i], 1 ).sum();
        
        GEPWarper &curWarper = this->at(i);
        
        for( int gi=0; gi<curWarper.gridNumY+1; gi++ ) for( int gj=0; gj<curWarper.gridNumX+1; gj++ )
        {
            int curVID = 2 * curWarper.V2DtoV1D( gi, gj ) + start_ncols[i];
            
            Eigen::Vector3f oriPt = curWarper.gridVs(gi,gj);
            curWarper.gridVs(gi,gj) << x(curVID+0), x(curVID+1), 1.0f;
            Eigen::Vector3f aftPt = curWarper.gridVs(gi,gj);
            
            ansDiff += (oriPt-aftPt).norm();
        }
        
        //update edge parameters
        curWarper.updateEdgeData();
    }

    return warpErrors;
}
