//
//  GEPWarper.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "GEPWarper.h"

inline void fillTriangle( cv::Mat &canvas, const cv::Point &p1, const cv::Point &p2, const cv::Point &p3, cv::Scalar color )
{
    int npts = 3;
    cv::Point tri[3] = { p1, p2, p3 };
    const cv::Point* ptri[1] = { tri };
    cv::fillPoly( canvas, ptri, &npts, 1, color );
}

//calc quad similarity weight using saliency
void GEPWarper::setSaliencyWeight(const cv::Mat &saliency)
{
    cv::Mat triIndexMap = cv::Mat::zeros( height, width, CV_32FC1 );
    
    for ( int i = 0; i < gridNumY; i++ ) for ( int j = 0; j < gridNumX; j++ )
    {
        cv::Point lu( gridOriVs(  i,  j)(0), gridOriVs(  i,  j)(1) );
        cv::Point ru( gridOriVs(  i,j+1)(0), gridOriVs(  i,j+1)(1) );
        cv::Point ld( gridOriVs(i+1,  j)(0), gridOriVs(i+1,  j)(1) );
        cv::Point rd( gridOriVs(i+1,j+1)(0), gridOriVs(i+1,j+1)(1) );
        cv::Point ctr = ( lu + ru + ld + rd ) * 0.25;
        
        float ind0 = 4 * ( i * gridNumX + j );
        
        fillTriangle( triIndexMap, lu, ctr, ru, cv::Scalar(ind0+0) );
        fillTriangle( triIndexMap, ru, ctr, rd, cv::Scalar(ind0+1) );
        fillTriangle( triIndexMap, ctr, ld, rd, cv::Scalar(ind0+2) );
        fillTriangle( triIndexMap, lu, ld, ctr, cv::Scalar(ind0+3) );
    }
    
    Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic> tmpGridWeights;
    tmpGridWeights.resize( 4*gridNumY, gridNumX );
    tmpGridWeights.setZero();
    
    for(int i=0;i<height;i++) for(int j=0;j<width;j++)
    {
        int index = (int)triIndexMap.at<float>(i,j);
        int triID = index % 4;
        int gridID = index / 4;
        int gridIDY = ( gridID / gridNumX )*4 + triID;
        int gridIDX = gridID % gridNumX;
        
        tmpGridWeights(gridIDY, gridIDX) += saliency.at<float>(i,j);
    }
    tmpGridWeights /= tmpGridWeights.maxCoeff();
    
    for ( int i = 0; i < 4*gridNumY; i++ ) for ( int j = 0; j < gridNumX; j++ )
        tmpGridWeights(i,j) = 0.3f + 0.7f * tmpGridWeights(i,j);
    tmpGridWeights /= tmpGridWeights.sum();
    
    this->gridWeights.resize( gridNumY, gridNumX );
    for ( int i = 0; i < gridNumY; i++ ) for ( int j = 0; j < gridNumX; j++ )
        this->gridWeights(i,j).setZero();
    for ( int i = 0; i < 4*gridNumY; i++ ) for ( int j = 0; j < gridNumX; j++ )
        this->gridWeights(i/4,j)(i%4) = tmpGridWeights(i,j);
}

//set matrix for similarity ( sum of diff of triangles in quad  )
void GEPWarper::setSmlrty( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int &curRow, float simWeight )
{
    for( int gi=0; gi<gridNumY; gi++ ) for( int gj=0; gj<gridNumX; gj++ )
    {
        //A B
        //C D
        int A = this->V2DtoV1D(   gi, gj   );//A
        int B = this->V2DtoV1D(   gi, gj+1 );//B
        int C = this->V2DtoV1D( gi+1, gj   );//C
        int D = this->V2DtoV1D( gi+1, gj+1 );//D
        
        int triOrders[4][3] = { {A,B,D}, {B,D,C}, {D,C,A}, {C,A,B} };
        
        Eigen::Vector4f weight;
        Eigen::Vector4f quadWeight = gridWeights(gi,gj);
        weight(0) = simWeight * 0.5f * ( quadWeight(0) + quadWeight(1) );
        weight(1) = simWeight * 0.5f * ( quadWeight(1) + quadWeight(2) );
        weight(2) = simWeight * 0.5f * ( quadWeight(2) + quadWeight(3) );
        weight(3) = simWeight * 0.5f * ( quadWeight(3) + quadWeight(0) );
        
        for(int i=0;i<4;i++)
        {
            //the weight of the neighboring vertices
            Eigen::VectorXf weightX(6);
            Eigen::VectorXf weightY(6);
            weightX << -1.0f,  0.0f,  1.0f, 1.0f, 0.0f, -1.0f;
            weightY <<  0.0f, -1.0f, -1.0f, 1.0f, 1.0f,  0.0f;
            weightX *= weight(i);
            weightY *= weight(i);
            
            for(int j=0;j<3;j++)
            {
                A_trips.push_back( Eigen::Triplet<float>( curRow,   2*triOrders[i][j], weightX(2*j+0) ) );
                A_trips.push_back( Eigen::Triplet<float>( curRow, 2*triOrders[i][j]+1, weightX(2*j+1) ) );
                b( curRow ) = 0.0f;
                
                A_trips.push_back( Eigen::Triplet<float>( curRow+1,   2*triOrders[i][j], weightY(2*j+0) ) );
                A_trips.push_back( Eigen::Triplet<float>( curRow+1, 2*triOrders[i][j]+1, weightY(2*j+1) ) );
                b( curRow+1 ) = 0.0f;
            }
            curRow+=2;
        }
    }
}

//update edge points/parameters after warping
void GEPWarper::updateEdgeData()
{
    int edgeNum = (int)this->lines.size();
    for(int i=0;i<edgeNum;i++)
    {
        //0. recover the edge points
        Line &curLine = this->lines[i];
        int edgePtNum = (int)curLine.evs_from.size();
        std::vector< cv::Point2f > evs_to(edgePtNum);
        for(int j=0;j<edgePtNum;j++)
        {
            const cv::Point2f &curEPtFrom = curLine.evs_from[j];
            
            // calc bi coeff
            int gi, gj;
            float alpha, beta;
            this->getBilinearCoeffs(gi, gj, alpha, beta, curEPtFrom);
            
            float luCoef = ( 1.0f - alpha ) * ( 1.0f - beta );
            float ldCoef = ( 1.0f - alpha ) * beta;
            float ruCoef = alpha * ( 1.0f - beta );
            float rdCoef = alpha * beta;
            
            Eigen::Vector3f newEPtTo;
            newEPtTo  = luCoef * this->gridVs(gi,gj);
            newEPtTo += ldCoef * this->gridVs(gi+1,gj);
            newEPtTo += ruCoef * this->gridVs(gi,gj+1);
            newEPtTo += rdCoef * this->gridVs(gi+1,gj+1);
            
            cv::Point2f &curEPtTo = evs_to[j];
            curEPtTo.x = newEPtTo(0);
            curEPtTo.y = newEPtTo(1);
        }
        
        //1. fit an edge, calc the params
        cv::Vec4f line;
        cv::fitLine( evs_to, line, CV_DIST_L2, 0, 0.01, 0.01 );

        //edge parameters
        float alpha = std::atan2( line[1], line[0] );
        curLine.sinAlpha = std::sin(alpha);
        curLine.cosAlpha = std::cos(alpha);
        curLine.b = curLine.sinAlpha * line[2] - curLine.cosAlpha * line[3];
    }
}

// set edge straightness penalty
void GEPWarper::setStrghtnss
( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int &curRow, float edgeWeight, int sumEdgePtNum )
{
    float w = edgeWeight / sumEdgePtNum;
    
    int edgeNum = (int)this->lines.size();
    for(int i=0;i<edgeNum;i++)
    {
        Line &curLine = this->lines[i];
        int edgePtNum = (int)curLine.evs_from.size();
        for(int j=0;j<edgePtNum;j++)
        {
            const cv::Point2f &curEPt = curLine.evs_from[j];
            
            // calc bi coeff
            int gi = curEPt.y / gridSz;
            int gj = curEPt.x / gridSz;
            
            float alpha = ( curEPt.x - gridOriVs(gi,gj)(0) ) / gridSz;
            float beta  = ( curEPt.y - gridOriVs(gi,gj)(1) ) / gridSz;
            
            int luGridV = this->V2DtoV1D( gi  , gj   );
            int ldGridV = this->V2DtoV1D( gi+1, gj   );
            int ruGridV = this->V2DtoV1D( gi  , gj+1 );
            int rdGridV = this->V2DtoV1D( gi+1, gj+1 );
            
            float luCoef = ( 1.0f - alpha ) * ( 1.0f - beta );
            float ldCoef = ( 1.0f - alpha ) * beta;
            float ruCoef = alpha * ( 1.0f - beta );
            float rdCoef = alpha * beta;
            
            luCoef *= w;
            ldCoef *= w;
            ruCoef *= w;
            rdCoef *= w;
            
            float sinAlpha = curLine.sinAlpha;
            float cosAlpha = curLine.cosAlpha;
            
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*luGridV,    luCoef*sinAlpha ) );
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ldGridV,    ldCoef*sinAlpha ) );
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ruGridV,    ruCoef*sinAlpha ) );
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*rdGridV,    rdCoef*sinAlpha ) );
            
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*luGridV+1, -luCoef*cosAlpha ) );
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ldGridV+1, -ldCoef*cosAlpha ) );
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ruGridV+1, -ruCoef*cosAlpha ) );
            A_trips.push_back( Eigen::Triplet<float>( curRow, 2*rdGridV+1, -rdCoef*cosAlpha ) );
            
            b(curRow++) = w * curLine.b;
        }
    }
}


GEPWarper::GEPWarper(const cv::Mat &img, int gridSz):Warper(img, gridSz)
{
    this->updateEdgeData();
    
    Utils_Ocv &uocv = Utils_Ocv::instance();
    cv::Mat saliencyMap = uocv.getGradient(img);
    cv::Mat tmp = saliencyMap * 0;
    saliencyMap = uocv.sumAbsDiff(tmp, saliencyMap);
    this->setSaliencyWeight(saliencyMap);
}

//add edge constraint
void GEPWarper::addEdgeConstraint( const cv::Point2f &ev1, const cv::Point2f &ev2, int interval, float initAlpha, float initB )
{
    cv::Rect imgRegion(0,0,width,height);
    
    cv::Point v1 = ev1;
    cv::Point v2 = ev2;
    
    bool isInImg = cv::clipLine( imgRegion, v1, v2 );
    if( !isInImg )
        return;
    
    cv::Point2f ev1in = v1;
    cv::Point2f ev2in = v2;
    
    Line l;
    l.sinAlpha = std::sin(initAlpha);
    l.cosAlpha = std::cos(initAlpha);
    l.b = initB;
    
    //line points
    cv::Point2f vpt = ev2in - ev1in;
    float vLen = std::sqrt( vpt.dot(vpt) );
    int samplePtNum = std::ceil( vLen / interval );
    for(int i=0; i<=samplePtNum; i++)
    {
        float w = (float)i/samplePtNum;
        cv::Point2f p = ev1in * w + ev2in * (1.0f-w);
        l.evs_from.push_back(p);
    }
    
    this->lines.push_back(l);
}

//get warping results
//=================================================================

//get warped feature map (white point, black background)
cv::Mat GEPWarper::getPointMap( cv::Size osz )const
{
    cv::Mat pointMap = cv::Mat( osz, CV_8UC1, cv::Scalar(0) );
    
    // points
    std::vector< cv::Point2f >::const_iterator vit = this->vFroms.begin();
    for( ; vit!=this->vFroms.end(); vit++ )
    {
        cv::Point2f vTo;
        this->remapPoint( *vit, vTo );
        cv::circle( pointMap, vTo, 0, cv::Scalar(255), -1 );
    }
    
    // harder points
    std::vector< cv::Point2f >::const_iterator vit_h = this->vFromsHarder.begin();
    for( ; vit_h!=this->vFromsHarder.end(); vit_h++ )
    {
        cv::Point2f vTo;
        this->remapPoint( *vit_h, vTo );
        cv::circle( pointMap, vTo, 0, cv::Scalar(255), -1 );
    }
    
    return pointMap;
}

//get warped edge map (white edge, black background)
cv::Mat GEPWarper::getEdgeMap( cv::Size osz, float edgeRange )const
{
    cv::Mat edgeMap = cv::Mat( osz, CV_8UC1, cv::Scalar(0) );
    
    std::vector< Line >::const_iterator lit = this->lines.begin();
    for( ; lit!=this->lines.end(); lit++ )
    {
        int evNum = (int)lit->evs_from.size();
        std::vector<cv::Point2f> evs_to(evNum);
        
        for(int i=0;i<evNum;i++)
            this->remapPoint( lit->evs_from[i], evs_to[i] );
        
        for(int i=0;i<evNum-1;i++)
            cv::line( edgeMap, evs_to[i], evs_to[i+1], cv::Scalar(255), edgeRange, CV_AA );
    }
    
    return edgeMap;
}

//get warped edge corner map (white edge, black background)
cv::Mat GEPWarper::getEdgeCornerMap( cv::Size osz )const
{
    cv::Mat edgeCornerMap = cv::Mat( osz, CV_8UC1, cv::Scalar(0) );
    
    std::vector< Line >::const_iterator lit = this->lines.begin();
    for( ; lit!=this->lines.end(); lit++ )
    {
        int evNum = (int)lit->evs_from.size();

        std::vector<cv::Point2f> evs_to(2);
        this->remapPoint( lit->evs_from[0], evs_to[0] );
        this->remapPoint( lit->evs_from[evNum-1], evs_to[1] );
            
        cv::circle( edgeCornerMap, evs_to.front(), 1, cv::Scalar(255), -1 );
        cv::circle( edgeCornerMap,  evs_to.back(), 1, cv::Scalar(255), -1 );
    }
    
    return edgeCornerMap;
}

//get importance map
cv::Mat GEPWarper::getImportanceMap( cv::Size osz )const
{
    //initial importance map
    cv::Mat importanceMap = cv::Mat( osz, CV_32FC1, cv::Scalar(0.0f) );
    
    Utils_Ocv &uocv = Utils_Ocv::instance();
    
    // points
    int vNum = (int)this->vFroms.size();
    for( int i=0; i<vNum; i++ )
    {
        cv::Point2f vTo;
        this->remapPoint( this->vFroms[i], vTo );
        uocv.drawImpWeight( importanceMap, vTo, this->gridSz * this->weights[i] );
    }
    
    // harder points
    int hvNum = (int)this->vFromsHarder.size();
    for( int i=0; i<hvNum; i++ )
    {
        cv::Point2f vTo;
        this->remapPoint( this->vFromsHarder[i], vTo );
        uocv.drawImpWeight( importanceMap, vTo, this->gridSz * this->weightsHarder[i] );
    }
    
    // edges
    cv::Mat edgeMap = this->getEdgeMap( osz, 1.0f );
    for(int i=0;i<osz.height;i++)for(int j=0;j<osz.width;j++)
    {
        uchar ec = edgeMap.at<uchar>(i,j);
        if( ec == 0 )
            continue;
        
        float w = ec / 255.0f;
        uocv.drawImpWeight( importanceMap, cv::Point(j,i), gridSz * w );
    }
    
    return importanceMap;
}

//get warp distortion map
cv::Mat GEPWarper::getDistortionMap(cv::Size osz)const
{
    Eigen::Matrix<Eigen::Vector4f, Eigen::Dynamic, Eigen::Dynamic> distortionWeights;
    distortionWeights.resize( gridNumY, gridNumX );
    
    for( int gi=0; gi<gridNumY; gi++ ) for( int gj=0; gj<gridNumX; gj++ )
    {
        //A B
        //C D
        // {A,B,D}, {B,D,C}, {D,C,A}, {C,A,B}
        int vOrders[4][6] = {   {gi  ,gj  ,gi  ,gj+1,gi+1,gj+1},
                                {gi  ,gj+1,gi+1,gj+1,gi+1,gj  },
                                {gi+1,gj+1,gi+1,gj  ,gi  ,gj  },
                                {gi+1,gj  ,gi  ,gj  ,gi  ,gj+1} };
        
        Eigen::Vector4f weight;
        Eigen::Vector4f quadWeight = gridWeights(gi,gj);
        weight(0) = 0.5f * ( quadWeight(0) + quadWeight(1) );
        weight(1) = 0.5f * ( quadWeight(1) + quadWeight(2) );
        weight(2) = 0.5f * ( quadWeight(2) + quadWeight(3) );
        weight(3) = 0.5f * ( quadWeight(3) + quadWeight(0) );
        
        //calc grid errors
        Eigen::Vector4f triErrors;
        triErrors.setZero();
        
        for(int i=0;i<4;i++)
        {
            //the weight of the neighboring vertices
            Eigen::VectorXf weightX(6);
            Eigen::VectorXf weightY(6);
            weightX << -1.0f,  0.0f,  1.0f, 1.0f, 0.0f, -1.0f;
            weightY <<  0.0f, -1.0f, -1.0f, 1.0f, 1.0f,  0.0f;
            
            float dx=0.0f, dy=0.0f;
            for(int j=0;j<3;j++)
            {
                const Eigen::Vector3f &ptij = this->gridVs( vOrders[i][j*2], vOrders[i][j*2+1] );
                
                dx += ptij(0) * weightX(2*j+0);
                dx += ptij(1) * weightX(2*j+1);
                
                dy += ptij(0) * weightY(2*j+0);
                dy += ptij(1) * weightY(2*j+1);
            }
            
            triErrors(i) += std::abs(dx) + std::abs(dy);
        }
        
        //A B
        //C D
        Eigen::Vector4f &triErrors2 = distortionWeights(gi,gj);
        triErrors2(0) = triErrors(3) + triErrors(0);
        triErrors2(1) = triErrors(0) + triErrors(1);
        triErrors2(2) = triErrors(1) + triErrors(2);
        triErrors2(3) = triErrors(2) + triErrors(3);
        triErrors2 /= 2.0f;
    }
    
    //initial distortion map
    cv::Mat distortionMap = cv::Mat( osz, CV_32FC1, cv::Scalar(0.0f) );
    
    for( int gi=0; gi<gridNumY; gi++ ) for( int gj=0; gj<gridNumX; gj++ )
    {
        //draw the error
        cv::Point PA( this->gridVs(gi  , gj  )(0), this->gridVs(gi  , gj  )(1) );
        cv::Point PB( this->gridVs(gi  , gj+1)(0), this->gridVs(gi  , gj+1)(1) );
        cv::Point PC( this->gridVs(gi+1, gj  )(0), this->gridVs(gi+1, gj  )(1) );
        cv::Point PD( this->gridVs(gi+1, gj+1)(0), this->gridVs(gi+1, gj+1)(1) );
        cv::Point PCTR = (PA+PB+PC+PD)*0.25f;
        
        Eigen::Vector4f &triErrors2 = distortionWeights(gi,gj);
        
        cv::Point tris[4][3]{ {PA,PB,PCTR}, {PB,PD,PCTR}, {PD,PC,PCTR}, {PC,PA,PCTR} };
        for(int i=0;i<4;i++)
        {
            int npts = 3;
            const cv::Point* ptri[1] = { tris[i] };
            cv::fillPoly( distortionMap, ptri, &npts, 1, cv::Scalar( triErrors2(i) ) );
        }
    }
    
    return distortionMap;
}

//=================================================================


