//
//  Warper.cpp
//  ivWarping
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "Warper.h"

//get bilinear coefficient
void Warper::getBilinearCoeffs(int &gi, int &gj, float &alpha, float &beta, const cv::Point2f &v)const
{
    gi = v.y / gridSz;
    gj = v.x / gridSz;

    alpha = ( v.x - gridOriVs(gi,gj)(0) ) / gridSz;
    beta  = ( v.y - gridOriVs(gi,gj)(1) ) / gridSz;
}

//check boundary
bool Warper::boundaryCheck(const cv::Point2f &v)const
{
    return v.x>=0 && v.x<width && v.y>=0 && v.y<height;
}

//2D to 1D: index mapping
int Warper::V2DtoV1D(int row, int col)const 
{
    bool testY = row>=0 && row<gridNumY+1;
    bool testX = col>=0 && col<gridNumX+1;
    
    if( testY && testX )
        return row * (gridNumX+1) + col;
    
    return -1;
}

//set matrix for point correspondence
void Warper::setPtCrrspndnc
( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int &curRow, float fweight )
{
    if( vFroms.size() == 0 )
        return;
    
    // Feature Alignment
    int ptNum = (int)vFroms.size();
    for( int i=0; i<ptNum; i++ )
    {
        float w = fweight * weights[i] / this->weightSum;
        
        int gi, gj;
        float alpha, beta;
        this->getBilinearCoeffs(gi, gj, alpha, beta, vFroms[i]);
        
        int luGridV = this->V2DtoV1D( gi  , gj   );
        int ldGridV = this->V2DtoV1D( gi+1, gj   );
        int ruGridV = this->V2DtoV1D( gi  , gj+1 );
        int rdGridV = this->V2DtoV1D( gi+1, gj+1 );
        
        float luCoef = ( 1.0f - alpha ) * ( 1.0f - beta );
        float ldCoef = ( 1.0f - alpha ) * beta;
        float ruCoef = alpha * ( 1.0f - beta );
        float rdCoef = alpha * beta;
        
        luCoef *= w;
        ldCoef *= w;
        ruCoef *= w;
        rdCoef *= w;
        
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*luGridV, luCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ldGridV, ldCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ruGridV, ruCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*rdGridV, rdCoef ) );
        b(curRow++) = (float)vTos[i].x * w;
        
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*luGridV+1, luCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ldGridV+1, ldCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ruGridV+1, ruCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*rdGridV+1, rdCoef ) );
        b(curRow++) = (float)vTos[i].y * w;
    }
}
void Warper::setHarderPtCrrspndnc
( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int &curRow, float fweight )
{
    if( vFromsHarder.size() == 0 )
        return;
    
    // Feature Alignment
    int ptNum = (int)vFromsHarder.size();
    for( int i=0; i<ptNum; i++ )
    {
        float w = fweight * weightsHarder[i] / this->weightHarderSum;
    
        int gi, gj;
        float alpha, beta;
        this->getBilinearCoeffs(gi, gj, alpha, beta, vFromsHarder[i]);
        
        int luGridV = this->V2DtoV1D( gi  , gj   );
        int ldGridV = this->V2DtoV1D( gi+1, gj   );
        int ruGridV = this->V2DtoV1D( gi  , gj+1 );
        int rdGridV = this->V2DtoV1D( gi+1, gj+1 );
        
        float luCoef = ( 1.0f - alpha ) * ( 1.0f - beta );
        float ldCoef = ( 1.0f - alpha ) * beta;
        float ruCoef = alpha * ( 1.0f - beta );
        float rdCoef = alpha * beta;
        
        luCoef *= w;
        ldCoef *= w;
        ruCoef *= w;
        rdCoef *= w;
        
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*luGridV, luCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ldGridV, ldCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ruGridV, ruCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*rdGridV, rdCoef ) );
        b(curRow++) = (float)vTosHarder[i].x * w;
        
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*luGridV+1, luCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ldGridV+1, ldCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*ruGridV+1, ruCoef ) );
        A_trips.push_back( Eigen::Triplet<float>( curRow, 2*rdGridV+1, rdCoef ) );
        b(curRow++) = (float)vTosHarder[i].y * w;
    }
}

//solve sparse linear system
float Warper::solveGrids( std::vector< Eigen::Triplet<float> > &A_trips, Eigen::VectorXf &b, int nrows, int ncols )
{
    // solve sparse linear transform
    Eigen::SparseMatrix<float> A;
    A.resize( nrows, ncols );
    A.setFromTriplets( A_trips.begin(), A_trips.end() );
    
    Eigen::SparseQR< Eigen::SparseMatrix<float>, Eigen::COLAMDOrdering<int> > solver;
    solver.compute(A);
    
    Eigen::VectorXf x(ncols);
    x = solver.solve(b);
    
    float ansDiff = 0.0f;
    
    //copy results
    for( int gi=0; gi<gridNumY+1; gi++ ) for( int gj=0; gj<gridNumX+1; gj++ )
    {
        int curVID = this->V2DtoV1D( gi, gj );
        
        Eigen::Vector3f oriPt = this->gridVs(gi,gj);
        this->gridVs(gi,gj) << x(2*curVID+0), x(2*curVID+1), 1.0f;
        Eigen::Vector3f aftPt = this->gridVs(gi,gj);
        
        ansDiff += (oriPt-aftPt).norm();
    }
    
    int vNum = (gridNumY+1) * (gridNumX+1);
    return ansDiff/vNum;
}


Warper::Warper(const cv::Mat &img, int gridSz)
{
    this->img = img.clone();
    this->width  = img.cols;
    this->height = img.rows;
    
    this->gridSz = gridSz;
    this->gridNumX = (int)std::ceil( (float)width /gridSz );
    this->gridNumY = (int)std::ceil( (float)height/gridSz );
    
    this->gridOriVs.resize( gridNumY+1, gridNumX+1 );
    this->gridVs.resize( gridNumY+1, gridNumX+1 );
    
    //initial quads
    for(int i=0;i<gridNumY+1;i++) for(int j=0;j<gridNumX+1;j++)
    {
        this->gridOriVs(i,j) << gridSz*j, gridSz*i;
        this->gridVs(i,j)    << gridSz*j, gridSz*i, 1.0f;
    }
    
    this->weightSum = this->weightHarderSum = 0.0f;
}

//add point correspondence constraint
void Warper::addPointConstraint( const cv::Point2f &vFrom, const cv::Point2f &vTo, float weight )
{
    if( !boundaryCheck(vFrom) )
        return;
    
    this->vFroms.push_back(vFrom);
    this->vTos.push_back(vTo);
    this->weights.push_back(weight);
    
    this->weightSum += weight;
}
void Warper::addHarderPointConstraint( const cv::Point2f &vFrom, const cv::Point2f &vTo, float weight )
{
    if( !boundaryCheck(vFrom) )
        return;
    
    this->vFromsHarder.push_back(vFrom);
    this->vTosHarder.push_back(vTo);
    this->weightsHarder.push_back(weight);
    
    this->weightHarderSum += weight;
}

//remap the point to the warped space
void Warper::remapPoint( const cv::Point2f &vFrom, cv::Point2f &vTo )const
{
    int gi, gj;
    float alpha, beta;
    this->getBilinearCoeffs( gi, gj, alpha, beta, vFrom );
    
    Eigen::Vector3f x1 = (1.0f-alpha) * gridVs(gi,gj)   + alpha * gridVs(gi,gj+1);
    Eigen::Vector3f x2 = (1.0f-alpha) * gridVs(gi+1,gj) + alpha * gridVs(gi+1,gj+1);
    Eigen::Vector3f vTo3f = (1.0f-beta) * x1 + beta * x2;

    vTo.x = vTo3f(0);
    vTo.y = vTo3f(1);
}

//get the deformed image
cv::Mat Warper::getDeformedImg(cv::Size osz, bool rgrid)const 
{
    cv::Mat deformedImg;
    
    //render on the frame buffer object
    Utils_Ogl::FBO fboObject;
    bool isFboReady = fboObject.start( osz.width, osz.height );
    if( isFboReady )
    {
        Eigen::Matrix4f I4f = Eigen::Matrix4f::Identity();
        Utils_Ogl::setMatrix( osz.width, osz.height, I4f, I4f );
        
        glPushAttrib( GL_CURRENT_BIT | GL_TRANSFORM_BIT | GL_ENABLE_BIT );
        Utils_Ogl::InitGL();
        glViewport(0,0,osz.width,osz.height);
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        glDisable(GL_CULL_FACE);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_BLEND);
        glMatrixMode(GL_TEXTURE);
        glPushMatrix();
        {
            glLoadIdentity();
            glScalef( 1.0f/width, 1.0/height, 1.0f );
            
            GLuint texID = Utils_Ogl::matToGL(this->img);
            glBindTexture(GL_TEXTURE_2D,texID);
            
            glBegin(GL_QUADS);
            {
                for ( int i = 0; i < gridNumY; i++ ) for ( int j = 0; j < gridNumX; j++ )
                {
                    glTexCoord2fv( gridOriVs(i,j).data() );
                    glVertex3fv(      gridVs(i,j).data() );
                    
                    glTexCoord2fv( gridOriVs(i,j+1).data() );
                    glVertex3fv(      gridVs(i,j+1).data() );
                    
                    glTexCoord2fv( gridOriVs(i+1,j+1).data() );
                    glVertex3fv(      gridVs(i+1,j+1).data() );
                    
                    glTexCoord2fv( gridOriVs(i+1,j).data() );
                    glVertex3fv(      gridVs(i+1,j).data() );
                }
            }
            glEnd();
            
            glBindTexture(GL_TEXTURE_2D, 0);
            glDeleteTextures(1,&texID);
            
            if(rgrid)
            {
                for ( int i = 0; i < gridNumY; i++ ) for ( int j = 0; j < gridNumX; j++ )
                {
                    glBegin(GL_LINE_LOOP);
                    glVertex3fv( gridVs(i,j).data() );
                    glVertex3fv( gridVs(i,j+1).data() );
                    glVertex3fv( gridVs(i+1,j+1).data() );
                    glVertex3fv( gridVs(i+1,j).data() );
                    glEnd();
                }
            }
        }
        glPopMatrix();
        glPopAttrib();
        
        Utils_Ogl::resetMatrix();
        
        deformedImg = fboObject.readFrameBuffer( this->img.channels() );
        
        fboObject.end();
    }
    
    return deformedImg;
}

cv::Mat Warper::getDeformedImg(const cv::Mat &img, cv::Size osz, bool rgrid)
{
    cv::Mat tmp = this->img;
    this->img = img;
    cv::Mat deformedImg = this->getDeformedImg(osz, rgrid);
    this->img = tmp;
    return deformedImg;
}


