//
//  ivAppObjPose.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppObjPose.h"

ivAppObjPose::ivAppObjPose()
{
    this->reset();
}

void ivAppObjPose::reset()
{
    this->_scl.setConstant(1.0f);
    this->_trs.setConstant(0.0f);
    this->_rot = Eigen::AngleAxis<float>( 0, Eigen::Vector3f::UnitX() );
}

void ivAppObjPose::setFromMatrix( Eigen::Matrix4f m )
{
    this->_trs = m.block(0,3,3,1);
    
    //decompose R*S by SVD
    Eigen::Matrix3f subM = m.block(0,0,3,3);
    Eigen::JacobiSVD<Eigen::MatrixXf> svd( subM, Eigen::ComputeThinU | Eigen::ComputeThinV );
    this->_scl = svd.singularValues();
    Eigen::Matrix3f rot = svd.matrixU();
    this->_rot = rot;
}

//get the transformation
Eigen::Matrix4f ivAppObjPose::gTransform()
{
    //scaling matrix
    Eigen::Matrix4f sm = Eigen::Matrix4f::Identity(4, 4);
    for(int i=0;i<3;i++)
        sm(i,i) = this->_scl(i);
    
    //rotation matrix
    Eigen::Matrix4f rm = Eigen::Matrix4f::Identity(4, 4);
    rm.block(0,0,3,3) = this->_rot.toRotationMatrix();
    
    //translation matrix
    Eigen::Matrix4f tm = Eigen::Matrix4f::Identity(4, 4);
    tm.block(0,3,3,1) = this->_trs;
    
    //return the matrix
    Eigen::Matrix4f gTransform;
    gTransform = tm * rm * sm;
    return gTransform;
}

Eigen::Matrix4f ivAppObjPose::rtTransform()
{
    //rotation matrix
    Eigen::Matrix4f rm = Eigen::Matrix4f::Identity(4, 4);
    rm.block(0,0,3,3) = this->_rot.toRotationMatrix();
    
    //translation matrix
    Eigen::Matrix4f tm = Eigen::Matrix4f::Identity(4, 4);
    tm.block(0,3,3,1) = this->_trs;
    
    //return the matrix
    Eigen::Matrix4f rtTransform;
    rtTransform = tm * rm;
    return rtTransform;
}

//interpolation
ivAppObjPose ivAppObjPose::interpolate(float alpha,ivAppObjPose &pos)
{
    ivAppObjPose result;
    result._scl = alpha * this->_scl + (1.0f-alpha) * pos._scl;
    result._trs = alpha * this->_trs + (1.0f-alpha) * pos._trs;
    result._rot = this->_rot.slerp( (1.0f-alpha), pos._rot );
    return result;
}

//operators
ivAppObjPose ivAppObjPose::operator+(const ivAppObjPose& pose)
{
    ivAppObjPose result;
    result._scl = _scl + pose._scl;
    result._trs = _trs + pose._trs;
    result._rot = _rot * pose._rot;
    return result;
}

ivAppObjPose ivAppObjPose::operator-(const ivAppObjPose& pose)
{
    ivAppObjPose result;
    result._scl = _scl - pose._scl;
    result._trs = _trs - pose._trs;
    result._rot = _rot * pose._rot.inverse();
    return result;
}

ivAppObjPose ivAppObjPose::operator*(float& scalar)
{
    ivAppObjPose result;
    result._scl = _scl * scalar;
    result._trs = _trs * scalar;
    
    Eigen::AngleAxis< float > aa;
    aa.fromRotationMatrix( _rot.toRotationMatrix() );
    float &ang = aa.angle();
    if (ang > M_PI) ang -= 2.0*M_PI;
    ang = std::fmod( ang*scalar ,2.0*M_PI);
    result._rot = aa;
    
    return result;
}

float ivAppObjPose::magnitude()
{
    float _trs_mag = _trs.norm();
    float _scl_mag = _scl.norm();
    
    Eigen::AngleAxis< float > aa;
    aa.fromRotationMatrix( _rot.toRotationMatrix() );
    float &ang = aa.angle();
    if (ang > M_PI) ang -= 2.0*M_PI;
    ang = std::fmod( ang ,2.0*M_PI);
    float _rot_mag = std::abs(ang);
    
    return _trs_mag + _scl_mag + _rot_mag;
}

ivAppObjPose ivAppObjPose::normalized()
{
    ivAppObjPose result;
    float mag = this->magnitude();
    result._scl = _scl / mag;
    result._trs = _trs / mag;
    
    Eigen::AngleAxis< float > aa;
    aa.fromRotationMatrix( _rot.toRotationMatrix() );
    float &ang = aa.angle();
    if (ang > M_PI) ang -= 2.0*M_PI;
    ang = std::fmod( ang / mag ,2.0*M_PI);
    result._rot = aa;
    
    return result;
}





//io
std::istream& operator>>(std::istream& is, ivAppObjPose& pose)
{
    is >> pose._scl(0) >> pose._scl(1) >> pose._scl(2);
    is >> pose._trs(0) >> pose._trs(1) >> pose._trs(2);
    is >> pose._rot.w() >> pose._rot.x() >> pose._rot.y() >> pose._rot.z();
    return is;
}

std::ostream& operator<<(std::ostream& os, const ivAppObjPose& pose)
{
    os << pose._scl(0) << " " << pose._scl(1) << " " << pose._scl(2) << " " ;
    os << pose._trs(0) << " " << pose._trs(1) << " " << pose._trs(2) << " " ;
    os << pose._rot.w() << " " << pose._rot.x() << " " << pose._rot.y() << " " << pose._rot.z() << std::endl;
    return os;
}




