//
//  ivAppSfM.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppSfM.h"

//calculate camPosition
void frameData::setup(const Eigen::Matrix4f &K)
{
    Eigen::Vector3f mRt = M.block(0,3,3,1);
    mRt = -1 * M.block(0,0,3,3).inverse() * mRt;
    camPosition.block(0,0,3,1) = mRt;
    camPosition(3) = 1.0f;
    
    unitCam = mRt;
    unitCam.normalize();
    
    KM = K * M;
    inverseM = M.inverse();
}

//calculate vpidIDs
void frameData::prepareIndexLookup(int objPtNum)
{
    this->vpidIDs = std::vector<int>( objPtNum, -1 );
    int vptNum = (int)this->visiblePIDs.size();
    for(int j=0;j<vptNum;j++)
        this->vpidIDs[ this->visiblePIDs[j] ] = j;
}

//=================================================
// adjust SfM data
//=================================================
void ivAppSfM::adjustByM(Eigen::Matrix4f &m)
{
    //local to global matrix
    Eigen::Matrix4f m_inv = m.inverse();
    
    //transform the 3d points
    for( int i=0; i<vPointData.size(); i++)
        vPointData[i].point = m * vPointData[i].point;
    
    //transform the camera motion
    int sfmFrameNum = (int)vFrameData.size();
    for( int i=0; i<sfmFrameNum; i++ )
        vFrameData[i].M = vFrameData[i].M * m_inv;

    //accumulate the normalization transform
    this->normalizationTransform = m * this->normalizationTransform;
    
    this->setupCamPositions();
}

//void ivAppSfM::adjustByScale(float scale)
//{
//    //the scaling matrix
//    Eigen::Matrix4f S;
//    S.setIdentity();
//    S.block(0,0,3,3) *= scale;
//    
//    //normalize the 3d points
//    for( int i=0; i<vPointData.size(); i++)
//        vPointData[i].point = S * vPointData[i].point;
//    
//    //normalize the camera
//    int sfmFrameNum = (int)vFrameData.size();
//    for( int i=0; i<sfmFrameNum; i++ )
//    {
//        frameData &curFrameData = vFrameData[i];
//        curFrameData.M.block(0,3,4,1) = S * curFrameData.M.block(0,3,4,1);
//    }
//    
//    this->setupCamPositions();
//}

void ivAppSfM::setupCamPositions()
{
    int fn = (int)vFrameData.size();
    for(int i=0;i<fn;i++)
        this->vFrameData[i].setup(this->K);
}

//=================================================
// Transfrom To Local Space
// project the selected points to the ground
// estimate 3 axis by PCA
// transform & normalize the point cloud
//=================================================
void ivAppSfM::transformToLocalSpace( Plane &ground, std::vector<int> &ptsid )
{
    int targetPtNum = (int)ptsid.size();
    
    //a ground & a point are not enough for determing a local space
    if( targetPtNum<=1 )
        return;
    
    //project the points to the ground
    cv::Mat projectedPts( targetPtNum, 3, CV_32FC1 );
    for(int i=0;i<targetPtNum;i++)
    {
        Eigen::Vector4f pt3dn;
        ground.getNearestPoint( pt3dn, vPointData[ ptsid[i] ].point );
        
        for(int j=0;j<3;j++)
            projectedPts.at<float>(i,j) = pt3dn(j);
    }

    //estimate the main axis
    cv::PCA pcaObj(projectedPts,cv::Mat(),CV_PCA_DATA_AS_ROW, 1);
    
    //center
    Eigen::Vector3f center;
    cv2eigen( pcaObj.mean, center );
    
    // vec_x
    Eigen::Vector3f vec_x;
    cv2eigen( pcaObj.eigenvectors, vec_x );
    vec_x.normalize();
    
    // vec_z
    Eigen::Vector3f vec_z;
    ground.getNormal(vec_z);
    
    // vec_y
    Eigen::Vector3f vec_y;
    vec_y = vec_z.cross( vec_x );

    //global to local matrix
    Eigen::Matrix4f g2l;
    g2l.setIdentity();
    g2l.block(0,0,1,3) = vec_x.transpose();
    g2l.block(1,0,1,3) = vec_y.transpose();
    g2l.block(2,0,1,3) = vec_z.transpose();
    g2l.block(0,3,3,1) = g2l.block(0,0,3,3) * -center;
    adjustByM(g2l);
    
    //find normalization matrix
    std::vector<float> scales;
    for(int i=0;i<targetPtNum;i++)
        for(int j=0;j<3;j++)
            scales.push_back( (float)std::abs( vPointData[ ptsid[i] ].point(j) ) );
    float maxScale = *std::max_element( scales.begin(), scales.end() );
    Eigen::Matrix4f g2l_scl = Eigen::Matrix4f::Identity();
    g2l_scl(0,0) = g2l_scl(1,1) = g2l_scl(2,2) = 1.0f / maxScale;
    adjustByM(g2l_scl);
    
    //z=0
    std::vector<float> params{0.0f,0.0f,1.0f,0.0f};
    ground.init(params);
}

//=================================================
// Rendering
//=================================================
void ivAppSfM::render( int framePos )const
{
    glColor4ub(0,255,0,255);
    //glColor4ub(255,255,255,255);
    
    //the tracking data
    const frameData &fd = vFrameData[ framePos ];
    
    //point cloud
    glBegin(GL_POINTS);
    std::vector<int>::const_iterator it3d = fd.visiblePIDs.begin();
    for(;it3d!=fd.visiblePIDs.end();it3d++)
        glVertex4fv( vPointData[*it3d].point.data() );
    glEnd();
    
    Utils_Ogl::drawAxis();
}

void ivAppSfM::renderCameraAxis(float ptSz, float lineWidth, float axisSize, int nframeID)const
{
    glPushAttrib(GL_CURRENT_BIT | GL_POINT_BIT | GL_LINE_BIT );
    
    glPointSize( ptSz );
    glLineWidth( lineWidth );
    glColor4ub(255, 255, 255, 255);
    
    //cameras orientation
    glPushMatrix();
    glMultMatrixf( vFrameData[nframeID].inverseM.data() );
    glScalef(axisSize, axisSize, axisSize);
    Utils_Ogl::drawAxis();
    glPopMatrix();

    glPopAttrib();
}

void ivAppSfM::renderCameraCube(float boxSize, int nframeID, bool isUnit)const
{
    Eigen::Matrix4f inverseM = vFrameData[nframeID].inverseM;

    if( isUnit )
        inverseM.block(0,3,3,1).normalize();
    
    glPushMatrix();
    glMultMatrixf( inverseM.data() );
    glScalef(boxSize, boxSize, boxSize);
    
    glDisable(GL_CULL_FACE);
    glPushAttrib(GL_POLYGON_BIT);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    
    Utils_Ogl::drawCube();
    
    glPopAttrib();
    glEnable(GL_CULL_FACE);

    glPopMatrix();
}

void ivAppSfM::renderCameraPyramid(float pyramidSize, int nframeID, bool isUnit)const
{
    Eigen::Matrix4f inverseM = vFrameData[nframeID].inverseM;
    
    if( isUnit )
        inverseM.block(0,3,3,1).normalize();
    
    glPushMatrix();
    glMultMatrixf( inverseM.data() );
    glScalef(pyramidSize, pyramidSize, pyramidSize);

    glDisable(GL_CULL_FACE);

    Utils_Ogl::drawPyramid();
    
    glPushAttrib( GL_POLYGON_BIT | GL_CURRENT_BIT );
    glColor4ub(0,0,0,255);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    Utils_Ogl::drawPyramid();
    glPopAttrib();
    
    glEnable(GL_CULL_FACE);

    glPopMatrix();
}

void ivAppSfM::renderCameraSphere(float sphereSize, int nframeID, bool isUnit)const
{
    Eigen::Matrix4f inverseM = vFrameData[nframeID].inverseM;
    
    if( isUnit )
        inverseM.block(0,3,3,1).normalize();
    
    glPushMatrix();
    glMultMatrixf( inverseM.data() );
    glTranslatef(0.0f, 0.0f, -sphereSize);
    glScalef(sphereSize, sphereSize, sphereSize);
    Utils_Ogl::drawSphere(5,5);
    glPopMatrix();
}


void ivAppSfM::renderPoints(float ptSz)const
{
    glPushAttrib(GL_CURRENT_BIT | GL_POINT_BIT );
    
    glPointSize( ptSz );
    glColor4ub(255, 255, 255, 255);
    
    //structure points
    glBegin(GL_POINTS);
    std::vector<pointData>::const_iterator itpd = vPointData.begin();
    for(;itpd!=vPointData.end();itpd++)
        glVertex4fv( itpd->point.data() );
    glEnd();
    
    //normals
    glBegin(GL_LINES);
    for(itpd = vPointData.begin();itpd!=vPointData.end();itpd++)
    {
        Eigen::Vector4f tmpPoint = itpd->point + 0.05 * itpd->normal;
        glVertex4fv( itpd->point.data() );
        glVertex4fv( tmpPoint.data() );
    }
    glEnd();
    
    glPopAttrib();
}

void ivAppSfM::renderCameraPath(float lineWidth, float axisSize, bool renderOrientation)const
{
    glPushAttrib(GL_CURRENT_BIT | GL_LINE_BIT );
    
    glLineWidth( lineWidth );
    glColor4ub(255, 255, 255, 255);
    
    //frame #
    int fNum = (int)vFrameData.size();
    
    if( renderOrientation )
    {
        //cameras orientation
        for(int i=0;i<fNum;i++)
        {
            glPushMatrix();
            glMultMatrixf( vFrameData[i].inverseM.data() );
            glScalef(axisSize, axisSize, axisSize);
            Utils_Ogl::drawAxis();
            glPopMatrix();
        }
    }
    
    //camera path
    glColor4ub(0, 0, 0, 255);
    glBegin(GL_LINE_STRIP);
    for(int i=0;i<fNum;i++)
        glVertex4fv(vFrameData[i].camPosition.data());
    glEnd();
    
    glPopAttrib();
}

//=================================================
// Normal Estimation
//=================================================
void ivAppSfM::estimateNormal(int NeighborNum)
{
    // setup the pointcloud for plane fitting
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    cloud->width = (int)vPointData.size();
    cloud->height = 1;
    cloud->points.resize (cloud->width * cloud->height);
    for (size_t i = 0; i < cloud->points.size(); ++i)
    {
        cloud->points[i].x = vPointData[i].point(0);
        cloud->points[i].y = vPointData[i].point(1);
        cloud->points[i].z = vPointData[i].point(2);
    }
    
    // Create the normal estimation class, and pass the input dataset to it
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> normal_estimation;
    normal_estimation.setInputCloud (cloud);
    
    // Create an empty kdtree representation, and pass it to the normal estimation object.
    // Its content will be filled inside the object, based on the given input dataset (as no other search surface is given).
    pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
    normal_estimation.setSearchMethod (tree);
    
    // Output datasets
    pcl::PointCloud<pcl::Normal>::Ptr cloud_normals (new pcl::PointCloud<pcl::Normal>);
    
    // Use all neighbors in a sphere of radius 3cm
    normal_estimation.setKSearch(NeighborNum);
    
    // Compute the features
    normal_estimation.compute (*cloud_normals);
    
    for( size_t i=0; i<cloud->points.size(); ++i )
    {
        vPointData[i].normal(0) = -cloud_normals->points[i].normal_x;
        vPointData[i].normal(1) = -cloud_normals->points[i].normal_y;
        vPointData[i].normal(2) = -cloud_normals->points[i].normal_z;
        vPointData[i].normal(3) = 0.0f;
    }
}

//=================================================
// Plane Estimation
//=================================================
bool ivAppSfM::estimatePlaneIn
(Plane &p, int framePos, cv::Mat &polygonRegion)
{
    std::vector<int> ptids_in;
    
    std::vector<int> &ptids = this->vFrameData[framePos].visiblePIDs;
    std::vector<Eigen::Vector2f> &msr2ds = this->vFrameData[framePos].visibleMsr2ds;
    int vpNum = (int)msr2ds.size();
    for(int i=0;i<vpNum;i++)
    {
        Eigen::Vector2f &msr2d = msr2ds[i];
        uchar c = polygonRegion.at<uchar>( msr2d(1), msr2d(0) );
        if( c > 127 )
            ptids_in.push_back( ptids[i] );
    }
    
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    cloud->width = (int)ptids_in.size();
    cloud->height = 1;
    cloud->points.resize (cloud->width * cloud->height);
    for (size_t j = 0; j < cloud->points.size(); ++j)
    {
        int ptid = ptids_in[j];
        Eigen::Vector4f &curPoint = this->vPointData[ptid].point;
        cloud->points[j].x = curPoint(0);
        cloud->points[j].y = curPoint(1);
        cloud->points[j].z = curPoint(2);
    }

    if( ptids_in.size()<5 )
        return false;
    
    return p.estimatePlane(cloud);
}

//=================================================
// Read/Write Object(Local) SfM
//=================================================
inline void readMatrix4( std::ifstream &ifs, Eigen::Matrix4f &m )
{
    Eigen::Matrix4f tmp;
    for(int i=0;i<16;i++)
        ifs >> tmp(i);
    
    m = tmp.transpose();
}
inline void readVector3( std::ifstream &ifs, Eigen::Vector4f &v )
{
    for(int i=0;i<3;i++)
        ifs >> v(i);
}
void ivAppSfM::readObjSfM(std::string fname)
{
    std::ifstream ifs(fname);
    
    //read frame size
    ifs >> this->width >> this->height;
    
    //read camera intrinsics K
    readMatrix4(ifs, this->K);
    
    //read point cloud
    int ptNum;
    ifs >> ptNum;
    this->vPointData.clear();
    this->vPointData.resize(ptNum);
    for(int i=0;i<ptNum;i++)
    {
        pointData &curPD = this->vPointData[i];
        
        curPD.point(3) = 1.0f;
        readVector3(ifs, curPD.point);
        
        this->vPointData[i].visibleFIDs.clear();
    }
    
    //read camera pose and visible points list
    this->msrNum = 0;
    int fNum;
    ifs >> fNum;
    this->vFrameData.clear();
    this->vFrameData.resize(fNum);
    for(int i=0;i<fNum;i++)
    {
        frameData &curFD = this->vFrameData[i];
        
        readMatrix4(ifs, curFD.M);
        
        int visiblePtNum;
        ifs >> visiblePtNum;
        this->msrNum += visiblePtNum;
        curFD.visibleMsr2ds.resize(visiblePtNum);
        curFD.visiblePIDs.resize(visiblePtNum);
        for(int j=0;j<visiblePtNum;j++)
        {
            //2d measurement
            ifs >> curFD.visibleMsr2ds[j](0);
            ifs >> curFD.visibleMsr2ds[j](1);
            //point id
            ifs >> curFD.visiblePIDs[j];
            
            //visible frame id for each point
            this->vPointData[ curFD.visiblePIDs[j] ].visibleFIDs.push_back(i);
        }
        
        //calculate vpidIDs
        curFD.prepareIndexLookup(ptNum);
        
        //calculate camera position
        curFD.setup(this->K);
    }
    
    //peek first if there exists the normalization transform
    this->normalizationTransform.setIdentity();
    std::string Normalization_Tranform;
    if( ifs>>Normalization_Tranform )
        readMatrix4(ifs, this->normalizationTransform);
    
    ifs.close();
    
    //post process
    //reason - the original measurement 2d is not the same as the reprojected one
    //we aim to align the projected point, so it would be better to be same.
    //erase the measurements points by reprojection
    for(int i=0;i<fNum;i++)
    {
        frameData &curFD = this->vFrameData[i];
        
        Eigen::Matrix4f prjM = this->K * curFD.M;
        
        int visiblePtNum = (int)curFD.visiblePIDs.size();
        for(int j=0;j<visiblePtNum;j++)
        {
            int ptid = curFD.visiblePIDs[j];
            Eigen::Vector4f newMsr = prjM * this->vPointData[ ptid ].point;
            curFD.visibleMsr2ds[j](0) = newMsr(0)/newMsr(2);
            curFD.visibleMsr2ds[j](1) = newMsr(1)/newMsr(2);
        }
    }
}

//SfM Writing
void ivAppSfM::writeObjSfM(std::string fname)
{
    std::ofstream ofs(fname);
    
    //output frame size
    ofs << this->width << " " << this->height << std::endl;
    
    //output camera intrinsics K
    ofs << this->K << std::endl;
    
    //output point cloud
    int ptNum = (int)this->vPointData.size();
    ofs << ptNum << std::endl;
    for(int i=0;i<ptNum;i++)
    {
        pointData &curPD = this->vPointData[i];
        ofs << curPD.point.block(0,0,3,1).transpose() << std::endl;
    }
    
    //output camera pose and visible points list
    int fNum = (int)this->vFrameData.size();
    ofs << fNum << std::endl;
    for(int i=0;i<fNum;i++)
    {
        frameData &curFD = this->vFrameData[i];
        ofs << curFD.M << std::endl;
        int visiblePtNum = (int)curFD.visiblePIDs.size();
        ofs << visiblePtNum << std::endl;
        for(int j=0;j<visiblePtNum;j++)
        {
            ofs << curFD.visibleMsr2ds[j].transpose() << " ";
            ofs << curFD.visiblePIDs[j] << std::endl;
        }
        ofs << std::endl;
    }
    
    ofs << "Normalization_Tranform" << std::endl;
    ofs << this->normalizationTransform << std::endl;
    
    ofs.close();
}
