//
//  ivAppSfM.h
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#ifndef __ivModeling__ivAppSfM__
#define __ivModeling__ivAppSfM__

#include "Plane.h"
#include "Utils_Ogl.h"

//the sfm data in one frame
class frameData
{
public:
    //camera motion (rotation/translation)
    Eigen::Matrix4f KM;
    Eigen::Matrix4f M;
    Eigen::Matrix4f inverseM;
    
    //visible point cloud ids
    std::vector<int> visiblePIDs;
    
    //the index in the visiblePIDs (inverse of visiblePIDs)
    //-1 means the points is not visible in the frame
    std::vector<int> vpidIDs;
    
    //calculate vpidIDs
    void prepareIndexLookup(int objPtNum);
    
    //visible point projected to image domain
    std::vector<Eigen::Vector2f> visibleMsr2ds;
    
    //the camera position
    Eigen::Vector4f camPosition;
    
    //the project of the camera in unit shpere
    Eigen::Vector3f unitCam;
    
    //calculate camPosition
    void setup(const Eigen::Matrix4f &K);
};

class pointData
{
public:
    //the coordinate of the points
    Eigen::Vector4f point;
    
    //point normal
    Eigen::Vector4f normal;
    
    //the id of frames in which the point is visible
    std::vector< int > visibleFIDs;
};

class ivAppSfM
{
public:
    //=================================================
    // video information
    //=================================================
    //size of frames
    int width, height;
    //2D measurement number
    int msrNum;
    
    //=================================================
    // SfM data
    //=================================================
    //camera intrinsics (fixed across frames)
    Eigen::Matrix4f K;
    //the camera motion and point cloud mask in each frame
    std::vector<frameData> vFrameData;
    //point cloud
    std::vector<pointData> vPointData;

    //=================================================
    // Normalization Transformation
    //=================================================
    Eigen::Matrix4f normalizationTransform;
    
    //=================================================
    // adjust SfM data
    //=================================================
public:
    // S=m*S, Mi=Mi*m_inv
    void adjustByM(Eigen::Matrix4f &m);
    // scale camera pos and structures
    //void adjustByScale(float scale);

    //setup the camera position
    void setupCamPositions();

    //=================================================
    // Transfrom To Local Space
    // project the selected points to the ground
    // estimate 3 axis by PCA
    // transform & normalize the point cloud
    //=================================================
    void transformToLocalSpace( Plane &ground, std::vector<int> &ptsid );
    
public:
    //=================================================
    // Rendering
    //=================================================
    virtual void render( int framePos )const ;
    void renderCameraAxis(float ptSz, float lineWidth, float axisSize, int nframeID)const;
    void renderCameraCube(float boxSize, int nframeID, bool isUnit)const;
    void renderCameraPyramid(float pyramidSize, int nframeID, bool isUnit)const;
    void renderCameraSphere(float sphereSize, int nframeID, bool isUnit)const;
    void renderPoints(float lineWidth)const;
    void renderCameraPath(float lineWidth, float axisSize, bool renderOrientation)const;
    
    //=================================================
    // Normal Estimation
    //=================================================
    void estimateNormal(int NeighborNum);
    
    //=================================================
    // Plane Estimation
    //=================================================
    bool estimatePlaneIn(Plane &p, int framePos,
                         cv::Mat &polygonRegion);
    
    //=================================================
    // Read/Write Object(Local) SfM
    //=================================================
    void readObjSfM (std::string fname);
    void writeObjSfM(std::string fname);
};

#endif /* defined(__ivModeling__ivAppSfM__) */
