//
//  ivAppSfM_OBJ.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by sulaispig on 9/1/14.
//  Copyright (c) 2014 nthu. All rights reserved.
//

#include "ivAppSfM_OBJ.h"

//=================================================
// ivPreprocessing - Normalization
//=================================================

//collect the object sfm data
void ivAppSfM_OBJ::init( ivAppSfM_SCN *ivaSfM_SCN, ivAppVideo *ivaMask_OBJ )
{
    this->vFrameData.clear();
    this->vPointData.clear();
    
    //copy frame size
    this->width  = ivaSfM_SCN->width;
    this->height = ivaSfM_SCN->height;
    
    //copy intrinsics
    this->K = ivaSfM_SCN->K;
    
    //copy normalization transform
    this->normalizationTransform = ivaSfM_SCN->normalizationTransform;
    
    //if there is no object mask video
    if( !ivaMask_OBJ->isReady )
    {
        //copy sfm data
        this->vFrameData = ivaSfM_SCN->vFrameData;
        this->vPointData = ivaSfM_SCN->vPointData;
        return;
    }
    
    int fNum     = (int)ivaSfM_SCN->vFrameData.size();
    int sfmPtNum = (int)ivaSfM_SCN->vPointData.size();
    
    //copy camera motions
    this->vFrameData.resize( fNum );
    for(int i=0;i<fNum;i++)
    {
        this->vFrameData[i].M = ivaSfM_SCN->vFrameData[i].M;
        this->vFrameData[i].camPosition = ivaSfM_SCN->vFrameData[i].camPosition;
    }
    
    //filer out those not on object mask
    std::vector<bool> isPointsVisible( sfmPtNum, false );
    std::vector<bool> isPointsMasked( sfmPtNum, true );
    for( int i=0; i<fNum; i++ )
    {
        frameData &fdi = ivaSfM_SCN->vFrameData[i];
        std::vector<Eigen::Vector2f> &vMsr2ds = fdi.visibleMsr2ds;
        std::vector<int> &vPIDs = fdi.visiblePIDs;
        
        cv::Mat &curMask = ivaMask_OBJ->frames[i];
        
        int ptNum = (int)vMsr2ds.size();
        for( int j=0; j<ptNum; j++ )
        {
            cv::Vec3b &c = curMask.at<cv::Vec3b>( vMsr2ds[j](1), vMsr2ds[j](0) );
            isPointsVisible[ vPIDs[j] ] = true;
            if( c[0]<127 )
                isPointsMasked[ vPIDs[j] ] = false;
        }
    }
    
    std::vector<bool> isObjectPoints( sfmPtNum );
    for( int i=0; i<sfmPtNum; i++ )
        isObjectPoints[i] = isPointsMasked[i] && isPointsVisible[i];
    
    //calc the index mapping
    int counter = 0;
    std::vector<int> oldPointIDs;//from new to old
    std::vector<int> newPointIDs( sfmPtNum, -1 );//from old to new
    for( int i=0; i<sfmPtNum; i++ )
        if( isObjectPoints[i] )
        {
            oldPointIDs.push_back(i);
            newPointIDs[i] = counter++;
        }
    
    //store the object feature points
    this->vPointData.resize( counter );
    for(int i=0;i<counter;i++)
        this->vPointData[i].point = ivaSfM_SCN->vPointData[ oldPointIDs[i] ].point;
    
    //store the visible pointID & measurements
    for( int i=0; i<fNum; i++ )
    {
        frameData &fdi = ivaSfM_SCN->vFrameData[i];
        std::vector<Eigen::Vector2f> &vMsr2ds = fdi.visibleMsr2ds;
        std::vector<int> &vPIDs = fdi.visiblePIDs;
        
        int ptNum = (int)vPIDs.size();
        for(int j=0;j<ptNum;j++)
        {
            int oldPID = vPIDs[j];
            int newPID = newPointIDs[oldPID];
            
            if( !isObjectPoints[oldPID] )
                continue;
            
            this->vFrameData[i].visiblePIDs.push_back( newPID );
            this->vFrameData[i].visibleMsr2ds.push_back( vMsr2ds[j] );
            
            this->vPointData[newPID].visibleFIDs.push_back(i);
        }
    }
    
    //transform to object space
    std::vector<float> planeNormal{0.0f,0.0f,1.0f,0.0f};
    Plane ground;
    ground.init(planeNormal);
    std::vector<int> ptsid;
    for(int i=0;i<counter;i++)
        ptsid.push_back(i);
    this->transformToLocalSpace( ground, ptsid );
}

//render diff cluster with diff color
void ivAppSfM_OBJ::render( int framePos )
{
    //the tracking data
    frameData &fd = vFrameData[ framePos ];
    
    //point cloud
    std::vector<int>::iterator it3d = fd.visiblePIDs.begin();
    for(;it3d!=fd.visiblePIDs.end();it3d++)
    {
        pointData &curPD = vPointData[*it3d];
        Eigen::Vector4f npt = curPD.point;
        
        glBegin(GL_POINTS);
        glVertex4fv( curPD.point.data() );
        glEnd();
        
        glBegin(GL_LINES);
        glVertex4fv( curPD.point.data() );
        glVertex4fv( npt.data() );
        glEnd();
    }
    
    Utils_Ogl::drawAxis();
}

