//
//  ivAppSfM_SCN.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by sulaispig on 8/29/14.
//  Copyright (c) 2014 nthu. All rights reserved.
//

#include "ivAppSfM_SCN.h"
//=================================================
// pre-processing
//=================================================
//transform to ground space
void ivAppSfM_SCN::transform2GroundSpace()
{
    //the id of labeled ground points
    std::vector<int> ptsid;
    for( int i=0; i<isPointOnGround.size(); i++ )
        if( isPointOnGround[i] )
            ptsid.push_back(i);
    
    //the set of ground points
    std::vector<Eigen::Vector4f> gndPoints;
    
    //collect the ground points
    int fNum = (int)this->vFrameData.size();
    Eigen::Matrix4f refMotionInv = this->vFrameData[0].M.inverse();
    std::vector<int>::iterator it = ptsid.begin();
    for(;it!=ptsid.end();it++)
    {
        Eigen::Vector4f &curPoint = this->vPointData[*it].point;
        gndPoints.push_back( curPoint );
        
        if( !this->isPlanarMotion )
            continue;
        
        for(int j=1;j<fNum;j++)
        {
            //its motion
            Eigen::Matrix4f relM = refMotionInv * this->vFrameData[j].M;
            Eigen::Vector4f relMPoint = relM * curPoint;
            gndPoints.push_back( relMPoint );
        }
    }
    
    // setup the pointcloud for plane fitting
    int gndPtNum = (int)gndPoints.size();
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    cloud->width = gndPtNum;
    cloud->height = 1;
    cloud->points.resize(gndPtNum);
    for(int i=0;i<gndPtNum;i++)
    {
        cloud->points[i].x = gndPoints[i](0);
        cloud->points[i].y = gndPoints[i](1);
        cloud->points[i].z = gndPoints[i](2);
    }
    
    //plane estimation
    Plane ground;
    if ( !ground.estimatePlane(cloud) )
    {
        std::cout << "[X] unable to fit a plane for the ground." << std::endl;
        return;
    }
    
    //correct the orientation of the ground
    Eigen::Vector2f pt2d;
    pt2d << width*0.5, height*0.5;
    Eigen::Vector4f pt3d;
    ground.getIntersection( pt3d, pt2d, K, vFrameData[0].M );
    Eigen::Vector4f pt2cam = vFrameData[0].camPosition - pt3d;
    Eigen::Vector3f n;
    ground.getNormal(n);
    if( n.dot( pt2cam.block(0,0,3,1) ) < 0.0f )
        ground.inverseNormal();
    
    //transform to ground space
    this->transformToLocalSpace( ground, ptsid );
}

//=================================================
// Parsing Voodoo
//=================================================
//convert the format of camera
inline void convertCamera
(const float params[28],
 Eigen::Matrix4f &K,
 Eigen::Matrix4f &M)
{
    //camera intrinsics
    K.setIdentity();
    K(0,0) = params[20] / params[14];
    K(1,1) = params[20] / params[15];
    K(0,2) = params[18] + params[16]*0.5;
    K(1,2) = params[19] + params[17]*0.5;
    
    //camera motion
    Eigen::Matrix4f r,t;
    r.setIdentity();
    r(0,0) = params[22];
    r(0,1) = params[23];
    r(0,2) = params[24];
    r(1,0) = params[25];
    r(1,1) = params[26];
    r(1,2) = params[27];
    r(2,0) = params[03];
    r(2,1) = params[04];
    r(2,2) = params[05];
    t.setIdentity();
    t(0,3) = -params[00];
    t(1,3) = -params[01];
    t(2,3) = -params[02];
    M = r*t;
}

//parse camera parameters & Point Cloud
bool ivAppSfM_SCN::parseCameraParamsPointCloud(std::string tcName)
{
    std::string camPtsCloudFile = ivDirParams::getCamPnt(tcName);
    
    std::ifstream ifs( camPtsCloudFile.c_str() );
    
    if( !ifs.is_open() )
        return false;
    
    //read the head comments
    std::string cmt;
    for( int i=0; i<39; i++ )
        std::getline( ifs, cmt );
    
    //read camera params
    while(true)
    {
        std::string tmp;
        ifs >> tmp;
        
        if(tmp=="#timeindex")
        {
            int frameID;
            ifs >> frameID;
            
            float camParams[28];
            for(int i=0;i<28;i++)
                ifs >> camParams[i];
            
            frameData fd;
            convertCamera(camParams, this->K, fd.M);
            fd.setup(this->K);
            
            //set the size of frames
            this->width  = camParams[16];
            this->height = camParams[17];
            
            this->vFrameData.push_back(fd);
        }
        else
            break;
    }
    
    //read point cloud
    std::getline( ifs, cmt );
    std::getline( ifs, cmt );
    std::getline( ifs, cmt );
    pointData pnt;
    while(true)
    {
        bool test = (bool)(ifs >> pnt.point(0));
        ifs >> pnt.point(1);
        ifs >> pnt.point(2);
        pnt.point(3) = 1.0f;
        
        if( !test )
            break;
        
        this->vPointData.push_back(pnt);
    }
    
    ifs.close();
    return true;
}

//parse Point Cloud
bool ivAppSfM_SCN::parsePointCloud(std::string tcName)
{
    std::string ptsCloudFile = ivDirParams::getPntDat(tcName);
    
    std::ifstream ifs( ptsCloudFile.c_str() );
    
    if( !ifs.is_open() )
        return false;
    
    this->msrNum = 0;
    
    std::string frmStr;
    while( ifs >> frmStr )
    {
        int frameID;
        ifs >> frameID;
        
        int ptNum;
        ifs >> ptNum;
        this->msrNum += ptNum;
        
        for(int i=0;i<ptNum;i++)
        {
            Eigen::Vector2f pt2d;
            int ptID;
            
            ifs >> pt2d(0) >> pt2d(1);
            ifs >> ptID;
            
            this->vFrameData[frameID].visibleMsr2ds.push_back( pt2d );
            this->vFrameData[frameID].visiblePIDs.push_back(ptID);
            
            this->vPointData[ptID].visibleFIDs.push_back(frameID);
        }
    }
    
    ifs.close();
    return true;
}

//=================================================
// read SfM in voodoo format
//=================================================
bool ivAppSfM_SCN::readVoodooSfM(std::string tcName)
{
    this->vFrameData.clear();
    this->vPointData.clear();
    
    //if it is Moving Object Case
    this->isPlanarMotion = tcName[0]=='M';
    
    //do parsing
    bool isOK1 = this->parseCameraParamsPointCloud(tcName);
    bool isOK2 = this->parsePointCloud(tcName);
    
    //normalization transform is identity initially
    this->normalizationTransform.setIdentity();

    //make up that the first frame cannot be read
    this->vFrameData[0] = this->vFrameData[1];
    
    this->isPointOnGround = std::vector<bool>( vPointData.size(), false );
    
    return isOK1 && isOK2;
}

//=================================================
// Ground Estimation
//=================================================
//add/remove ground points
void ivAppSfM_SCN::collectGroundPoints(int framePos, cv::Mat curStroke, bool isAdd)
{
    //collect the points
    frameData &curFrameData = this->vFrameData[framePos];
    std::vector<Eigen::Vector2f> &msr2ds = curFrameData.visibleMsr2ds;
    std::vector<int> &PIDs = curFrameData.visiblePIDs;
    
    //for each visible point at framePos
    int ptNum = (int)msr2ds.size();
    for(int i=0;i<ptNum;i++)
    {
        Eigen::Vector2f &msr2d = msr2ds[i];
        int pid = PIDs[i];
        
        //if the point is on the stroke
        uchar c = curStroke.at<uchar>( msr2d(1), msr2d(0) );
        if( isAdd )
            isPointOnGround[pid] = isPointOnGround[pid] || (c>127);
        else
            isPointOnGround[pid] = isPointOnGround[pid] && !(c>127);
    }
    this->transform2GroundSpace();
}

//=================================================
// Rendering
//=================================================
void ivAppSfM_SCN::render( int framePos )
{
    //the tracking data
    frameData &fd = vFrameData[ framePos ];
    
    glPushAttrib( GL_CURRENT_BIT );
    
    //point cloud
    glBegin(GL_POINTS);
    std::vector<int>::iterator it3d = fd.visiblePIDs.begin();
    for(;it3d!=fd.visiblePIDs.end();it3d++)
    {
        if( !isPointOnGround[*it3d] )
            glColor4ub(255,255,255,255);
        else
            glColor4ub(255,0,0,255);
        glVertex4fv( vPointData[*it3d].point.data() );
    }
    glEnd();

    //ground point
    glColor4ub(255,0,0,255);
    glBegin(GL_LINES);
    it3d = fd.visiblePIDs.begin();
    for(;it3d!=fd.visiblePIDs.end();it3d++)
        if( isPointOnGround[*it3d] )
        {
            Eigen::Vector4f &curP = vPointData[*it3d].point;
            glVertex4fv( curP.data() );
            glVertex4f( curP(0), curP(1), curP(2)+0.1f, 1.0f );
        }
    glEnd();
    
    glPopAttrib();
    
    //draw axis
    Utils_Ogl::drawAxis();
    
}







