//
//  ivAppTvModel.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppTvModel.h"

//find the front primitive that contain the point
//(primitives ids are in ascending order)
std::vector<int> ivAppTvModel::findNeighborPrims(int ptid)
{
    std::vector<int> trgPrims;
    
    int primNum = (int)this->vtvPrim.size();
    for(int i=0;i<primNum;i++)
        if( this->vtvPrim[i].ifContainPt(ptid) )
            trgPrims.push_back(i);
    
    return trgPrims;
}
std::vector<int> ivAppTvModel::findFrontNeighborPrims(int framePos, int ptid)
{
    std::vector<int> trgPrims;
    
    std::vector<int> nPrims = this->findNeighborPrims(ptid);
    std::vector<int>::iterator it = nPrims.begin();
    for(;it!=nPrims.end();it++)
        if( this->vtvPrim[*it].isFront(framePos) )
            trgPrims.push_back(*it);
    
    return trgPrims;
}

//remove isolated & useless(not connected) points
void ivAppTvModel::rmIsolatedPoints(std::vector<bool> &isIsolated)
{
    int vptNum = (int)this->vtvPoint.size();
    
    //check if the isolated points are still in other prims
    for(int i=0;i<(int)this->vtvPrim.size();i++)
        this->vtvPrim[i].markVertex(isIsolated, false);
    
    //reorder the 3d points/tracks
    std::vector<int> newPtInd(vptNum,-1);
    int counter = 0;
    for(int i=0;i<vptNum;i++)
    {
        if( isIsolated[i] )
            continue;
        
        newPtInd[i] = counter;
        this->vtvPoint[counter] = this->vtvPoint[i];
        counter++;
    }
    this->vtvPoint.resize(counter);
    
    //index remapping
    for(int i=0;i<(int)this->vtvPrim.size();i++)
        this->vtvPrim[i].remapIndex(newPtInd);
}

//unify overlapping points in different primitive
void ivAppTvModel::unifyPoint(int framePos, int ptid)
{
    int target = this->selectPoint(framePos, *this->getPoint2D(framePos, ptid), ptid);
    if( target==-1 )
        return;
    
    //the target point cant be in neighboring primitives
    std::vector<int> nPrims = this->findNeighborPrims(ptid);
    for(int i=0;i<(int)nPrims.size();i++)
        if( this->vtvPrim[ nPrims[i] ].ifContainPt(target) )
            return;
    
    int ptNum = (int)this->vtvPoint.size();
    
    std::vector<int> newInd(ptNum);
    for(int i=0;i<ptNum;i++)
        newInd[i] = i;
    newInd[ptid] = target;
    
    int primNum = (int)this->vtvPrim.size();
    for(int i=0;i<primNum;i++)
        this->vtvPrim[i].remapIndex(newInd);
    
    std::vector<bool> isIsolated(ptNum,false);
    isIsolated[ptid] = true;
    this->rmIsolatedPoints(isIsolated);
}

//get selected primID
int ivAppTvModel::getPrimID(int framePos, int selectedPt)
{
    std::vector<int> neighborPrims;
    if( lblParams::isBackFaceCull )
        neighborPrims = this->findFrontNeighborPrims(framePos,selectedPt);
    else
        neighborPrims = this->findNeighborPrims(selectedPt);
    
    int nPrimNum = (int)neighborPrims.size();
    
    if( nPrimNum<=0 )
        return -1;
    else if( nPrimNum==1 )
        this->primPointer = 0;
    else
        this->primPointer = this->primPointer % nPrimNum;
    
    return neighborPrims[this->primPointer];
}

//remove the primitive
void ivAppTvModel::rmPrimitive(int primID)
{
    if( primID<0 )
        return;
    
    int vptNum = (int)this->vtvPoint.size();
    std::vector<bool> isIsolated( vptNum, false );
    this->vtvPrim[primID].markVertex(isIsolated, true);
    this->vtvPrim.erase( this->vtvPrim.begin()+primID );
    this->rmIsolatedPoints(isIsolated);
}

//clear all data
void ivAppTvModel::init()
{
    this->primPointer = 0;
    ivAppTvModelData::init();
}

//select a point, if the point doesn't exist return -1
//the returned point is not notthisone
int ivAppTvModel::selectPoint(int framePos, const Eigen::Vector2f &p, int notthisone)
{
    //newly added point candidates - loop backward
    int tvptNum = (int)this->vtvPoint.size();
    for(int i=tvptNum-1;i>=0;i--)
    {
        if( this->vtvPoint[i].isReady3D[framePos] )
            break;
        
        Eigen::Vector2f &p2d = *this->getPoint2D(framePos, i);
        if( (p-p2d).norm()<lblParams::POINT_SZ && i!=notthisone )
            return i;
    }
    
    //for each primitive
    int primNum = (int)this->vtvPrim.size();
    for(int i=0;i<primNum;i++)
    {
        bool isPrimFront = this->vtvPrim[i].isFront(framePos);
        if( !isPrimFront && lblParams::isBackFaceCull )
            continue;
        
        std::vector<int> &curPrim = this->vtvPrim[i].polygon;
        
        //the primitive is visible
        int pNum = (int)curPrim.size();
        for(int j=0;j<pNum;j++)
        {
            int tvpid = curPrim[j];
            
            Eigen::Vector2f &p2d = *this->getPoint2D(framePos, tvpid);
            if( (p-p2d).norm()<lblParams::POINT_SZ && tvpid!=notthisone )
                return tvpid;
        }
    }
    
    //no point is selected
    return -1;
}

//push_back a 2d point candidate to be a 3d point
int ivAppTvModel::pushPointCandidate(int fNum, int framePos, const Eigen::Vector2f &p)
{
    tvPoint3D tvp;
    tvp.init(fNum);
    tvp.tvv2D[framePos] = p;
    tvp.isKey[framePos] = true;
    this->vtvPoint.push_back(tvp);
    return (int)this->vtvPoint.size()-1;
}

//pop_back a 2d point candidate
void ivAppTvModel::popPointCandidate(int framePos, int ptid)
{
    if( !this->vtvPoint[ptid].isReady3D[framePos] )
        this->vtvPoint.pop_back();
}

//get the pointer to certain 2d point
Eigen::Vector2f* ivAppTvModel::getPoint2D(int framePos, int ptid)
{
    tvPoint3D &selectedPt3D = this->vtvPoint[ptid];
    return &(selectedPt3D.tvv2D[framePos]);
}



//add a polygon
void ivAppTvModel::addPrimitive(int framePos, ivAppSfM *sfmData, lblPolygon2D &lblPoly)
{
    assert(lblPoly.isReady);
    
    //copy user labeled primitive
    tvTriPrim tpoly;
    tpoly.polygon = lblPoly.prim;
    
    //add the primitive first
    this->vtvPrim.push_back( tpoly );
 
    //set the origin frame
    this->vtvPrim.back().originFrame = framePos;
    
    //the id of the added polygon
    int curPrimID = (int)this->vtvPrim.size()-1;
    
    //estimate/update the 3d position of the primitive
    this->vtvPrim[curPrimID].updatePrimitive3D( this->vtvPoint, framePos, sfmData, lblPoly);
    
    //set all the points in the polygon to be key
    int ptNum = (int)lblPoly.prim.size();
    for(int i=0;i<ptNum;i++)
    {
        int pid = lblPoly.prim[i];
        this->vtvPoint[pid].isKey[framePos] = true;
    }
}

//update picked point & neighboring polygon
void ivAppTvModel::adjustOneRing
(int ptid, int framePos, const Eigen::Vector2f &p, lblPolygon2D &lblPolyAdjust, ivAppSfM *sfmData)
{
    //set key point
    this->vtvPoint[ ptid ].tvv2D[framePos] = p;
    
    //neighboring primitives - ascending order
    std::vector<int> vPrims = this->findFrontNeighborPrims( framePos, ptid );
    int primNum = (int)vPrims.size();
    
    //unify the overlapping points
    if( sfmData )
        this->unifyPoint(framePos, ptid);
    
    //neighboring primitives - descending order
    for( int i=primNum-1; i>=0; i-- )
    {
        int primID = vPrims[i];
        
        std::vector<int> &curPrim = this->vtvPrim[primID].polygon;
        int pNum = (int)curPrim.size();
        
        //simulate a lblPolygon2D
        lblPolyAdjust.reset();
        for(int j=0;j<=pNum;j++)
        {
            int pid = curPrim[ j%pNum ];
            int pid_prev = curPrim[ (j-1+pNum)%pNum ];
            
            //neighboring edges of the selected should be snapped
            Eigen::Vector2f *selectedPt2D = this->getPoint2D(framePos, pid);
            bool isToEdgeSnap = lblParams::isEdgeSnapping && (ptid==pid_prev||ptid==pid);
            lblPolyAdjust.addPoint2D( vtvPoint, framePos, *selectedPt2D, pid, isToEdgeSnap );
        }
        
        //no need to estimate 3D data
        if( sfmData == NULL )
            continue;
        
        //the adjusted polygon is not good
        if( !lblPolyAdjust.isReady )
            continue;
        
        //set keys
        for(int j=0;j<=pNum;j++)
        {
            int pid = curPrim[ j%pNum ];
            int pid_prev = curPrim[ (j-1+pNum)%pNum ];
            int pid_post = curPrim[ (j+1)%pNum ];
            
            //set keys
            bool isKey = ptid==pid;
            bool isNeighborKey = ptid==pid_prev || ptid==pid_post;
            if( (isNeighborKey && lblParams::isEdgeSnapping) || isKey )
                this->vtvPoint[pid].isKey[framePos] = true;
        }
        
        //estimate/update the 3d position of the primitive
        this->vtvPrim[primID].updatePrimitive3D( this->vtvPoint, framePos, sfmData, lblPolyAdjust);
    }
}

//cancel the adjusting
void ivAppTvModel::cancelAdjusting(int framePos, ivAppSfM *sfmData, int selectedPt)
{
    int pNum = (int)this->vtvPoint.size();
    std::vector<bool> isNeighbor = std::vector<bool>(pNum, false);
    
    std::vector<int> nPrims = this->findFrontNeighborPrims(framePos, selectedPt);
    std::vector<int>::iterator it = nPrims.begin();
    for(;it!=nPrims.end();it++)
        this->vtvPrim[*it].markVertex(isNeighbor, true);
    
    Eigen::Matrix4f prjM = sfmData->K * sfmData->vFrameData[framePos].M;
    for(int i=0;i<pNum;i++)
        if(isNeighbor[i])
            this->vtvPoint[i].recoverPoint2D(framePos, prjM);
}

//remove neighboring primitives
void ivAppTvModel::rmOneRingPrims(int framePos, int selectedPt)
{
    std::vector<int> nPrims = this->findNeighborPrims(selectedPt);
    std::vector<int>::reverse_iterator it = nPrims.rbegin();
    for(; it!=nPrims.rend(); it++ )
        this->rmPrimitive(*it);
}

//remove the primitive
void ivAppTvModel::rmPrimitive(int framePos, int selectedPt)
{
    int primID = this->getPrimID(framePos, selectedPt);
    this->rmPrimitive(primID);
}

//set the critical of the selected prim
void ivAppTvModel::setPrimCritical(int framePos, int fNum, int selectedPt)
{
    int primID = this->getPrimID(framePos, selectedPt);
    if( primID<0 )
        return;
    
    this->vtvPrim[primID].switchCritical( this->vtvPoint, framePos, fNum );
}



//render visible 3d data
void ivAppTvModel::renderOCV(cv::Mat &canvas, int framePos, int selectedPt)
{
    int primNum = (int)vtvPrim.size();
    
    //all prims
    for(int i=primNum-1;i>=0;i--)
        this->vtvPrim[i].renderOCV( this->vtvPoint, canvas, framePos, selectedPt, false);
    
    //selected prim
    int primID = this->getPrimID( framePos, selectedPt);
    if( primID<0 )
        return;
    
    this->vtvPrim[ primID ].renderOCV(this->vtvPoint, canvas, framePos, selectedPt, true);
}

