//
//  lblPolygon2D.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "lblPolygon2D.h"

//check if there are repeated points
inline bool checkDist(Eigen::Vector2f &p1, Eigen::Vector2f &p2)
{
    float dist = ( p1 - p2 ).norm();
    return dist<=lblParams::POINT_SZ;
}

//return the nearest point on the line(inf)
inline void project2Line( const cv::Vec4f &line, Eigen::Vector2f &point )
{
    //      vab
    // a -------------- b
    //  \
    //   \  vap
    //    \
    //     * p
    
    cv::Point2f p( point(0), point(1) );
    cv::Point2f a( line[2], line[3] );
    cv::Point2f b( line[2]+line[0], line[3]+line[1] );
    
    cv::Point2f vab = b - a;
    cv::Point2f vap = p - a;
    
    float dotProduct = vab.dot(vap);
    float sqr_vba = vab.x * vab.x + vab.y * vab.y;
    float param_t = dotProduct / sqr_vba;
    p = a + param_t * vab;
    
    point(0) = p.x;
    point(1) = p.y;
}

//snap p1 p2 to edge
inline void snap2Edge(Eigen::Vector2f &p1, Eigen::Vector2f &p2, cv::Mat &edgeMap)
{
    int width  = edgeMap.cols;
    int height = edgeMap.rows;
    
    Eigen::Vector2f vp1p2 = p1-p2;
    int edgeLen = (int)vp1p2.norm();
    if( edgeLen < lblParams::SNAP_WID )
        return;
    
    //get the edge map in target region
    cv::Mat targetEdgeMap = cv::Mat::zeros( height, width, CV_8UC1 );
    cv::Point2f cvp1( p1(0), p1(1) );
    cv::Point2f cvp2( p2(0), p2(1) );
    cv::line( targetEdgeMap, cvp1, cvp2, cv::Scalar(255), lblParams::SNAP_WID );
    targetEdgeMap = targetEdgeMap & edgeMap;

    //snap to neighboring edge
    std::vector<cv::Point2f> epts;
    for(int i=0;i<height;i++)for(int j=0;j<width;j++)
    {
        if( targetEdgeMap.at<uchar>(i,j)>0 )
            epts.push_back( cv::Point2f(j,i) );
    }
    epts.push_back(cvp1);
    epts.push_back(cvp2);
    
    cv::Vec4f line;
    //CV_DIST_L2 -> standard unweighted least squares fit
    //OTHERS -> RANSAC
    cv::fitLine( epts, line, CV_DIST_L1, 0, 0.01, 0.01 );
    
    //project the points to the line
    project2Line( line, p1 );
    project2Line( line, p2 );
}

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Polygon_2.h>
typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef K::Point_2 Point;
typedef CGAL::Polygon_2<K> Polygon_2;
//check if the polygon is self intersected
inline bool lblPolygon2D::isSimplePolygon()
{
    int pNum = (int)this->plgn2d.size();
    std::vector<Point> ps(pNum);
    for(int i=0;i<pNum;i++)
    {
        Eigen::Vector2f &curPt = this->plgn2d[i];
        ps[i] = Point( curPt(0), curPt(1) );
    }
    Polygon_2 pgn( &(ps[0]) , &(ps[0])+pNum );
    
    return pgn.is_simple();
}

//reset
void lblPolygon2D::reset()
{
    this->plgn2d.clear();
    this->prim.clear();
    this->isReady = false;
}

//init
void lblPolygon2D::init(ivAppVideoCanny *canny)
{
    this->reset();
    this->canny = canny;
}

//add a 2D point to form a polygon
bool lblPolygon2D::addPoint2D
(tvPoint3DSet &vtvPoint, int framePos, Eigen::Vector2f lblp, int pid, bool toSnapEdge )
{
    if( this->isReady )
        return false;
    
    //set current frame
    this->curFrame = framePos;
    
    int pNum = (int)this->plgn2d.size();
    if( pNum > 0 )
    {
        bool isReturning = checkDist( lblp, this->plgn2d[0] );
        bool isRepeated  = checkDist( lblp, this->plgn2d.back() );
        
        //non-simple polygon is not allowed, point&line are ok
        if( isReturning && pNum>2 && !this->isSimplePolygon() )
            return false;
        
        //repeated same points are not allowed except the first point
        if( !isReturning && isRepeated )
            return false;
        
        if( isReturning )
        {
            lblp = this->plgn2d.front();
            this->isReady = true;
        }
        
        if( toSnapEdge )
        {
            snap2Edge( lblp, plgn2d[pNum-1], canny->edges[framePos] );
            
            if( pid>=0 )
            {
                //assign to the original fields
                vtvPoint[pid].tvv2D[framePos] = lblp;
                vtvPoint[ this->prim[pNum-1] ].tvv2D[framePos] = plgn2d[pNum-1];
            }
        }
    }
    
    if( !this->isReady )
    {
        this->plgn2d.push_back(lblp);
        this->prim.push_back(pid);
    }
    
    return true;
}

//remove the last point
void lblPolygon2D::rmPoint2D()
{
    if( !this->plgn2d.empty() )
    {
        this->plgn2d.pop_back();
        this->prim.pop_back();
    }
}

//rendering
void lblPolygon2D::render2D(cv::Mat &canvas)
{
    //draw current polygon
    int plgnVNum = (int)this->plgn2d.size();
    
    for(int i=0;i<plgnVNum;i++)
    {
        int j = (i+1) % plgnVNum;
        
        Eigen::Vector2f &lblpt2di = this->plgn2d[i];
        Eigen::Vector2f &lblpt2dj = this->plgn2d[j];
        
        cv::Point pt2di( lblpt2di(0), lblpt2di(1) );
        cv::Point pt2dj( lblpt2dj(0), lblpt2dj(1) );
        
        if( lblParams::isEdgeSnapping )
        {
            //edge snapping cases
            bool ESC1 = j==plgnVNum-1 && !isReady; //for labeling points
            bool ESC2 = j==0 && isReady; //the returning point
            if( ESC1 || ESC2 )
                cv::line(canvas, pt2di, pt2dj, Utils_Color::RED_A, lblParams::SNAP_WID );
        }
        cv::line(canvas, pt2di, pt2dj, Utils_Color::RED );
    }
    
    for(int i=0;i<plgnVNum;i++)
    {
        Eigen::Vector2f &lblpt2di = this->plgn2d[i];
        cv::Point pt2di( lblpt2di(0), lblpt2di(1) );
        cv::circle( canvas, pt2di, lblParams::POINT_SZ, Utils_Color::RED_A, -1 );
        cv::circle( canvas, pt2di, lblParams::POINT_SZ, Utils_Color::BLACK );
    }
}

//rendering preview
void lblPolygon2D::render2D_prev
(cv::Mat &canvas, tvPoint3DSet &vtvPoint, int framePos, Eigen::Vector2f lblp)
{
    int pNum = (int)this->plgn2d.size();
    if( pNum == 0 )return;
    
    //a preview polygon
    lblPolygon2D tmpPolygon2D(canny);
    for(int i=0;i<pNum;i++)
        tmpPolygon2D.addPoint2D( vtvPoint, framePos, this->plgn2d[i], this->prim[i], false );
    tmpPolygon2D.addPoint2D( vtvPoint, framePos, lblp, -1, lblParams::isEdgeSnapping );
    
    //render preview
    tmpPolygon2D.render2D(canvas);
}

//get polygon region (CV_8UC1)
cv::Mat lblPolygon2D::getRegion(int width, int height)
{
    cv::Mat canvas = cv::Mat::zeros(height, width, CV_8UC1);
    int npts = (int)this->plgn2d.size();
    std::vector<cv::Point> pts;
    for(int i=0;i<npts;i++)
    {
        Eigen::Vector2f &lblpt2d = this->plgn2d[i];
        cv::Point p( lblpt2d(0), lblpt2d(1) );
        pts.push_back( p );
    }
    const cv::Point* elementPoints[1] = { &pts[0] };
    cv::fillPoly(canvas, elementPoints, &npts, 1, cv::Scalar(255));

    return canvas;
}


