//
//  tvPoint3D.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "tvPoint3D.h"

//get the index of key frames
void tvPoint3D::getKeyFrameIDs(std::vector<int> &keyframeList)
{
    keyframeList.clear();
    
    int timeLen = (int)this->tvv3D.size();
    for(int i=0;i<timeLen;i++)
        if( this->isKey[i] )
            keyframeList.push_back(i);
}

//initialize the point
void tvPoint3D::init( int timeLen )
{
    this->tvv3D = std::vector<Eigen::Vector4f>(timeLen);
    this->tvv2D = std::vector<Eigen::Vector2f>(timeLen);
    this->isKey        = std::vector<bool>(timeLen, false);
    this->isReady3D    = std::vector<bool>(timeLen, false);
    this->isReliable3D = std::vector<bool>(timeLen, false);
}

//recover the 2d point by projection
void tvPoint3D::recoverPoint2D( int framePos, const Eigen::Matrix4f &prjM )
{
    if( !this->isReady3D[framePos] )
        return;
        
    Eigen::Vector4f &pt3d = this->tvv3D[framePos];
    Eigen::Vector2f &pt2d = this->tvv2D[framePos];

    Eigen::Vector4f prjedpt3d = prjM * pt3d;
    pt2d(0) = prjedpt3d(0)/prjedpt3d(2);
    pt2d(1) = prjedpt3d(1)/prjedpt3d(2);
}

//propagate keyframe points to non-keyframe ones
void tvPoint3D::propagateTVPT3D(const ivAppSfM *sfmData)
{
    int fNum = (int)this->tvv3D.size();
    
    std::vector<int> keyframeList;
    this->getKeyFrameIDs(keyframeList);
    
    int nviews = (int)keyframeList.size();
    if( nviews==1 )
    {
        int onlyKey = keyframeList.front();
        bool isReliable = this->isReliable3D[onlyKey];
        int fNum = (int)this->tvv2D.size();
        
        if( !isReliable )
        {
            this->tvv2D.assign( fNum, this->tvv2D[onlyKey] );
            return;
        }
    }
    
    //before the first key
    std::vector<int> keyList;
    if( keyframeList.front() > 0 )
    {
        this->tvv3D[0] = this->tvv3D[keyframeList.front()];
        keyList.push_back(0);
    }
    
    //middle keys
    std::vector<int>::iterator it = keyframeList.begin();
    for(;it!=keyframeList.end();it++)
        keyList.push_back( *it );
    
    
    //after the last key
    if( keyframeList.back() < fNum-1 )
    {
        this->tvv3D[fNum-1] = this->tvv3D[keyframeList.back()];
        keyList.push_back(fNum-1);
    }
    
    //interpolation 3D points
    int startKeyID = 0;
    for(int i=0;i<fNum;i++)
    {
        int prev = keyList[startKeyID];
        int post = keyList[startKeyID+1];
        if( i==post )
            startKeyID++;
        
        float alpha = (float)(i - prev)/(post-prev);
        this->tvv3D[i] = this->tvv3D[prev] * (1.0f-alpha) + this->tvv3D[post] * alpha ;
        this->tvv3D[i](3) = 1.0f;
        
        this->isReady3D[i] = true;
    }
    
    //calculate projection
    for(int i=0;i<fNum;i++)
    {
        if( this->isKey[i] )
            continue;
        
        Eigen::Matrix4f prjM = sfmData->K * sfmData->vFrameData[i].M;
        Eigen::Vector4f prjedPt = prjM * this->tvv3D[i];
        Eigen::Vector2f &curPt2D = this->tvv2D[i];
        curPt2D << prjedPt(0)/prjedPt(2), prjedPt(1)/prjedPt(2);
    }
}

//3d points triangulation
void tvPoint3D::triangulateTVPT3D(const ivAppSfM *sfmData)
{
    std::vector<int> keyframeList;
    this->getKeyFrameIDs(keyframeList);
    
    int nviews = (int)keyframeList.size();
    if(nviews<2)
        return;

    // Collect Projection Matrix and 2D Points
    std::vector<openMVG::Mat34> Ps(nviews);
    openMVG::Mat2X xs(2, nviews);
    for ( int i=0; i<nviews; i++ )
    {
        int keyFrameID = keyframeList[i];

        //Projection Matrix
        Eigen::Matrix4f P = sfmData->K * sfmData->vFrameData[ keyFrameID ].M;
        Ps[i] = P.block(0,0,3,4).cast<double>();
        
        //2D measurements
        Eigen::Vector2f &p = this->tvv2D[ keyFrameID ];
        xs.col(i) = p.cast<double>();
    }
    
    //triangulation by a "Iterated linear method" (openMVG)
    openMVG::Triangulation triangulationObj;
    for ( int i=0; i<nviews; i++ )
        triangulationObj.add( Ps[i], xs.col(i) );
    openMVG::Vec3 X = triangulationObj.compute();

    Eigen::Vector4f pt3d = Eigen::Vector4f::Ones();
    pt3d.block(0,0,3,1) = X.cast<float>();
    
    //estimate the nearest 3D points(in each frame) to X
    for (int i = 0; i < nviews; ++i)
    {
        int keyFrameID = keyframeList[i];
        
        //if it is reliable, no need to estimate a 3d point again
        if( this->isReliable3D[keyFrameID] )
            continue;
            
        Eigen::Matrix4f P = sfmData->K * sfmData->vFrameData[ keyFrameID ].M;
        Eigen::Matrix4f invP = P.inverse();
        Eigen::Vector2f &pt2d = this->tvv2D[ keyFrameID ];
        
        //estimate a plane whose normal is the vector from camera to the pixel
        //the plane intersect the estimated 3d point pt3d
        std::vector<float> params(4);
        params[0] = invP(0,0)*pt2d(0) + invP(0,1)*pt2d(1) + invP(0,2);
        params[1] = invP(1,0)*pt2d(0) + invP(1,1)*pt2d(1) + invP(1,2);
        params[2] = invP(2,0)*pt2d(0) + invP(2,1)*pt2d(1) + invP(2,2);
        params[3] = -( params[0]*pt3d(0) + params[1]*pt3d(1) + params[2]*pt3d(2) );
        
        Plane tmpPlane;
        tmpPlane.init(params);
        
        //project the pt3d to the 3d line
        Eigen::Vector4f newpt3d;
        tmpPlane.getIntersection(newpt3d, pt2d, sfmData->K,
                                 sfmData->vFrameData[ keyFrameID ].M);
        
        //set the points
        this->isReady3D[ keyFrameID ] = true;
        this->tvv3D[ keyFrameID ] = newpt3d;
    }
}

//io
std::istream& operator>>(std::istream& is, tvPoint3D& tvpt)
{
    is >> tvpt.v3D(0) >> tvpt.v3D(1) >> tvpt.v3D(2) >> tvpt.v3D(3);
    
    int key3DReadyFrameNum;
    is >> key3DReadyFrameNum;
    
    for(int i=0;i<key3DReadyFrameNum;i++)
    {
        int fid;
        is >> fid;
        
        int isReliable;
        is >> isReliable;
        tvpt.isReliable3D[fid] = (bool)isReliable;
        
        is >> tvpt.tvv3D[fid](0) >> tvpt.tvv3D[fid](1) >> tvpt.tvv3D[fid](2) >> tvpt.tvv3D[fid](3);
        is >> tvpt.tvv2D[fid](0) >> tvpt.tvv2D[fid](1);
        
        tvpt.isKey[fid] = true;
        tvpt.isReady3D[fid] = true;
    }
    
    return is;
}
std::ostream& operator<<(std::ostream& os, const tvPoint3D& tvpt)
{
    std::vector<int> key3DReadyFrameList;
    
    int fNum = (int)tvpt.tvv2D.size();
    for(int i=0;i<fNum;i++)
    {
        bool isKeyReady3D = tvpt.isKey[i] && tvpt.isReady3D[i];
        if( isKeyReady3D )
            key3DReadyFrameList.push_back(i);
    }
    
    os << tvpt.v3D.transpose() << std::endl;
    
    //# of key & 3d ready points
    os << key3DReadyFrameList.size() << std::endl;
    
    std::vector<int>::iterator it = key3DReadyFrameList.begin();
    for(;it!=key3DReadyFrameList.end();it++)
    {
        os << *it << " ";
        os << tvpt.isReliable3D[*it] << " ";
        os << tvpt.tvv3D[*it].transpose() << " ";
        os << tvpt.tvv2D[*it].transpose() << std::endl;
    }
    
    return os;
}







