//
//  tvTriPrim.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "tvTriPrim.h"

//calc visibility of the primitive in framePos by
//cross product
void tvTriPrim::calcVisibilityByCross(tvPoint3DSet &vtvPoint, int fNum)
{
    this->isVisible.resize(fNum, false);
    
    for(int fid=0;fid<fNum;fid++)
    {
        bool isFront = true;
        int triNum = (int)this->triangles.size();
        for(int i=0;i<triNum;i++)
        {
            //(OCV Coordinate)
            //---------------->x
            // |(x)
            // |
            // v y
            
            Eigen::Vector3i &curTri = this->triangles[i];
            Eigen::Vector2f &p0 = vtvPoint[curTri[0]].tvv2D[fid];
            Eigen::Vector2f &p1 = vtvPoint[curTri[1]].tvv2D[fid];
            Eigen::Vector2f &p2 = vtvPoint[curTri[2]].tvv2D[fid];
            Eigen::Vector2f vp1p0 = p1-p0;
            Eigen::Vector2f vp2p0 = p2-p0;
            isFront = isFront && ( vp1p0(0) * vp2p0(1) - vp1p0(1) * vp2p0(0) ) < 0;
        }
        this->isVisible[fid] = isFront;
    }
}
//critical labeling
void tvTriPrim::calcVisibilityByCritical(tvPoint3DSet &vtvPoint, int fNum)
{
    this->isCritical = std::vector<bool>(fNum, false);
    for(int i=0;i<(int)this->criticalFrames.size();i++)
        this->isCritical[ this->criticalFrames[i] ] = true;

    //find key frames
    std::vector<int> keyFrames;
    for(int i=0;i<fNum;i++)
    {
        bool isKey = true;
        for(int j=0;j<(int)this->polygon.size();j++)
            isKey = isKey && vtvPoint[this->polygon[j]].isKey[i];
        
        if(isKey)
            keyFrames.push_back(i);
    }
    
    //calc visibility
    this->isVisible = std::vector<bool>(fNum, false);
    std::vector<int>::iterator it = keyFrames.begin();
    for(;it!=keyFrames.end();it++)
    {
        //right
        for( int keyfid=*it; keyfid<fNum; keyfid++ )
        {
            this->isVisible[keyfid] = true;
            if( this->isCritical[keyfid] )
                break;
        }

        //left
        for( int keyfid=*it; keyfid>=0; keyfid-- )
        {
            this->isVisible[keyfid] = true;
            if( this->isCritical[keyfid] )
                break;
        }
    }
}

//calculate if the primitive is visible
void tvTriPrim::calcVisibility(tvPoint3DSet &vtvPoint, int fNum)
{
    //if 3d points are not ready
    if( !this->isReady3D(vtvPoint, fNum) )
        return;
    
    int pNum = (int)polygon.size();
    if( pNum>=3 )
        calcVisibilityByCross(vtvPoint, fNum);
    else
        calcVisibilityByCritical(vtvPoint, fNum);
}

//is prim 3d ready
bool tvTriPrim::isReady3D(tvPoint3DSet &vtvPoint, int fNum)
{
    //if 3d points are ready
    bool ready = true;
    for(int j=0;j<(int)this->polygon.size();j++)
        ready = ready && vtvPoint[this->polygon[j]].isReady3D[0];
    return ready;
}

//check if the point is in the polygon
bool tvTriPrim::ifContainPt(int ptid)
{
    std::vector<int>::iterator it = std::find( polygon.begin(), polygon.end(), ptid );
    return it!=polygon.end();
}

//mark points
void tvTriPrim::markVertex(std::vector<bool> &vertexMarks, bool mark)
{
    int vNum = (int)polygon.size();
    for(int i=0;i<vNum;i++)
        vertexMarks[ polygon[i] ] = mark;
}

//constrainted delaunay triangulation on the polygon
void tvTriPrim::cdTriangulate(tvPoint3DSet &vtvPoint)
{
    this->triangles.clear();
    
    int ptNum = (int)this->polygon.size();
    if( ptNum<=2 )
        return;
    
    //collect 2d points
    std::vector< Eigen::Vector2f > pts(ptNum);
    for(int i=0;i<ptNum;i++)
    {
        tvPoint3D &curP = vtvPoint[ this->polygon[i] ];
        pts[i] = curP.tvv2D[ this->originFrame ];
        pts[i](1) = -pts[i](1);
    }
    
    //polygon triangulation
    Poly2Tri::execute( pts, this->triangles );
    
    //remapping
    int triNum = (int)this->triangles.size();
    for(int i=0;i<triNum;i++)
        for(int j=0;j<3;j++)
            this->triangles[i](j) = this->polygon[ this->triangles[i](j) ];
}

//point index remapping
void tvTriPrim::remapIndex(const std::vector<int> &newPtInd)
{
    int vNum = (int)this->polygon.size();
    for(int j=0;j<vNum;j++)
        this->polygon[j] = newPtInd[ this->polygon[j] ];
    
    //update the vertex id in triangles
    int triNum = (int)this->triangles.size();
    for(int i=0;i<triNum;i++)for(int j=0;j<3;j++)
        this->triangles[i](j) = newPtInd[ this->triangles[i](j) ];
}

//tvpoint propagation & prismitive triangulation
void tvTriPrim::digest(tvPoint3DSet &vtvPoint, ivAppSfM *sfmData)
{
    //interpolate 3d points
    std::vector<int>::iterator it = this->polygon.begin();
    for(; it!=this->polygon.end(); it++ )
        vtvPoint[*it].propagateTVPT3D(sfmData);
    
    //polygon triangulation
    this->cdTriangulate(vtvPoint);
    
    int fNum = (int)sfmData->vFrameData.size();
    this->calcVisibility(vtvPoint, fNum);
}

//if the primitive is front in the frame
//(used when in original video)
bool tvTriPrim::isFront(int framePos) const
{
    if( this->isVisible.empty() )
        return true;
    
    return this->isVisible[framePos];
}

//(used when in novel view)
bool tvTriPrim::isFront(  const tvPoint3DSet &vtvPoint, const Eigen::Matrix4f &prjM, int framePos) const
{
    //for each triangle
    int triNum = (int)this->triangles.size();
    if(triNum==0)return true;
    
    Eigen::Vector3i tri = this->triangles[0];
    Eigen::Vector4f pt0 = prjM * vtvPoint[tri(0)].tvv3D[framePos];
    Eigen::Vector4f pt1 = prjM * vtvPoint[tri(1)].tvv3D[framePos];
    Eigen::Vector4f pt2 = prjM * vtvPoint[tri(2)].tvv3D[framePos];
    
    Eigen::Vector2f p0, p1, p2;
    p0 << pt0(0)/pt0(2), pt0(1)/pt0(2);
    p1 << pt1(0)/pt1(2), pt1(1)/pt1(2);
    p2 << pt2(0)/pt2(2), pt2(1)/pt2(2);
    
    Eigen::Vector2f vp1p0 = p1-p0;
    Eigen::Vector2f vp2p0 = p2-p0;
    
    return ( vp1p0(0) * vp2p0(1) - vp1p0(1) * vp2p0(0) ) < 0;
}

//switch critical
void tvTriPrim::switchCritical(tvPoint3DSet &vtvPoint, int framePos, int fNum)
{
    if(this->polygon.size()>=3)
        return;
    
    std::vector<int>::iterator it;
    it = std::find(this->criticalFrames.begin(), this->criticalFrames.end(), framePos);
    
    if( it != this->criticalFrames.end() )
        this->criticalFrames.erase(it);
    else
        this->criticalFrames.push_back(framePos);
    
    this->calcVisibilityByCritical(vtvPoint, fNum);
}

//update 3D primitive using sfmData & lblPoly(contains 2D points of models)
void tvTriPrim::updatePrimitive3D(tvPoint3DSet &vtvPoint, int framePos, ivAppSfM *sfmData, lblPolygon2D &lblPoly)
{
    assert( lblPoly.isReady );
    
    //fit a plane
    cv::Mat polygonRegion = lblPoly.getRegion(sfmData->width, sfmData->height);
    Plane polygonPlane;
    bool isPlaneReady = sfmData->estimatePlaneIn( polygonPlane, framePos, polygonRegion );
    
    int ptNum = (int)lblPoly.prim.size();
    if( isPlaneReady )
    {
        //add 3D polygon
        Eigen::Matrix4f K = sfmData->K;
        Eigen::Matrix4f M = sfmData->vFrameData[framePos].M;
        
        for(int i=0;i<ptNum;i++)
        {
            Eigen::Vector4f pt3d;
            Eigen::Vector2f pt2d = lblPoly.plgn2d[i].cast<float>();
            polygonPlane.getIntersection( pt3d, pt2d, K, M );
            
            int pid = lblPoly.prim[i];
            vtvPoint[pid].tvv3D[framePos] = pt3d;
            vtvPoint[pid].isReady3D[framePos] = true;
            vtvPoint[pid].isReliable3D[framePos] = true;
        }
    }
    else
    {
        for(int i=0;i<ptNum;i++)
            vtvPoint[ lblPoly.prim[i] ].triangulateTVPT3D(sfmData);
    }
    
    //tvpoint propagation & primitive triangulation
    this->digest(vtvPoint, sfmData);
}


//rendering for current view 2D
void tvTriPrim::renderOCV(tvPoint3DSet &vtvPoint, cv::Mat &canvas, int framePos, int selectedPt, bool isSelected)
{
    bool isPrimVisible = this->isFront(framePos);
    if( !isPrimVisible && lblParams::isBackFaceCull )
        return;
    
    int pNum = (int)this->polygon.size();
    
    //render the lines
    for(int j=0;j<pNum;j++)
    {
        int pid1 = this->polygon[j];
        int pid2 = this->polygon[(j+1)%pNum];
        Eigen::Vector2f &p1 = vtvPoint[pid1].tvv2D[framePos];
        Eigen::Vector2f &p2 = vtvPoint[pid2].tvv2D[framePos];
        
        bool isToDrawRed = lblParams::isEdgeSnapping && ( selectedPt==pid1 || selectedPt==pid2 ) ;
        if( isToDrawRed )
        {
            cv::line(canvas, cv::Point(p1(0),p1(1)), cv::Point(p2(0),p2(1)), Utils_Color::RED_A, lblParams::SNAP_WID );
            cv::line(canvas, cv::Point(p1(0),p1(1)), cv::Point(p2(0),p2(1)), Utils_Color::RED );
        }
        else
            cv::line(canvas, cv::Point(p1(0),p1(1)), cv::Point(p2(0),p2(1)), Utils_Color::BLACK );
        
        if( !this->isCritical.empty() && this->isCritical[framePos] )
            cv::line(canvas, cv::Point(p1(0),p1(1)), cv::Point(p2(0),p2(1)), Utils_Color::GREEN );
        
        if( isSelected )
            cv::line(canvas, cv::Point(p1(0),p1(1)), cv::Point(p2(0),p2(1)), Utils_Color::WHITE );
    }
    
    //render the points
    for(int j=0;j<pNum;j++)
    {
        int pid1 = this->polygon[j];
        tvPoint3D &tvpt3d = vtvPoint[pid1];
        Eigen::Vector2f &p1 = tvpt3d.tvv2D[framePos];
        
        cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::BLUE_A, -1 );
        cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::BLACK );
        
        bool isKey = tvpt3d.isKey[framePos];
        bool isReady3D = tvpt3d.isReady3D[framePos];
        bool isReliable3D = tvpt3d.isReliable3D[framePos];
        if( isKey )
        {
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::RED_A, -1 );
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::BLACK );
        }
        if( selectedPt==pid1 )
        {
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::RED, -1 );
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::BLACK );
        }
        if( isReady3D )
        {
            if( isReliable3D )
                cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ/2, Utils_Color::WHITE, -1 );
            else
                cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ/2, Utils_Color::WHITE_A, -1 );
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ/2, Utils_Color::BLACK );
        }

        if( !this->isCritical.empty() && this->isCritical[framePos] )
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::GREEN );
        
        if( isSelected )
            cv::circle( canvas, cv::Point(p1(0),p1(1)), lblParams::POINT_SZ, Utils_Color::WHITE );
    }
}

//rendering for novel view 3D
void tvTriPrim::renderOGL(tvPoint3DSet &vtvPoint, int framePos)
{
    //3D is not ready
    if( this->isVisible.empty() )
        return;
    
    //for each triangle
    int triNum = (int)this->triangles.size();
    glColor4ub(0,0,0,0);
    glBegin(GL_TRIANGLES);
    for(int j=0;j<triNum;j++)
    {
        Eigen::Vector3i tri = this->triangles[j];
        for(int k=0;k<3;k++)
        {
            tvPoint3D &tvpt3d = vtvPoint[tri(k)];
            glVertex4fv( tvpt3d.tvv3D[framePos].data() );
        }
    }
    glEnd();
    
    int ptNum = (int)this->polygon.size();
    glColor4ub(255,255,255,255);
    glBegin(GL_LINE_LOOP);
    for(int j=0;j<ptNum;j++)
    {
        tvPoint3D &tvpt3d = vtvPoint[polygon[j]];
        glVertex4fv( tvpt3d.tvv3D[framePos].data() );
    }
    glEnd();
}

//rendering for optimized points
void tvTriPrim::renderOGL(tvPoint3DSet &vtvPoint)
{
    //3D is not ready
    if( this->isVisible.empty() )
        return;
    
    //for each triangle
    int triNum = (int)this->triangles.size();
    glColor4ub(0,0,0,0);
    glBegin(GL_TRIANGLES);
    for(int j=0;j<triNum;j++)
    {
        Eigen::Vector3i tri = this->triangles[j];
        for(int k=0;k<3;k++)
        {
            tvPoint3D &tvpt3d = vtvPoint[tri(k)];
            glVertex4fv( tvpt3d.v3D.data() );
        }
    }
    glEnd();
    
    int ptNum = (int)this->polygon.size();
    glColor4ub(255,255,255,255);
    glBegin(GL_LINE_LOOP);
    for(int j=0;j<ptNum;j++)
    {
        tvPoint3D &tvpt3d = vtvPoint[polygon[j]];
        glVertex4fv( tvpt3d.v3D.data() );
    }
    glEnd();
}

//io
std::istream& operator>>(std::istream& is, tvTriPrim& prim)
{
    is >> prim.originFrame;
    
    int ptNum;
    is >> ptNum;
    
    prim.polygon.resize(ptNum);
    for(int i=0;i<ptNum;i++)
        is >> prim.polygon[i];
    
    int criticalNum;
    is >> criticalNum;
    prim.criticalFrames.resize(criticalNum);
    for(int i=0;i<criticalNum;i++)
        is >> prim.criticalFrames[i];
    
    return is;
}

std::ostream& operator<<(std::ostream& os, const tvTriPrim& prim)
{
    os << prim.originFrame << " ";
    
    int ptNum = (int)prim.polygon.size();
    os << ptNum << " ";

    for(int i=0;i<ptNum;i++)
        os << prim.polygon[i] << " ";
    
    int criticalNum = (int)prim.criticalFrames.size();
    os << criticalNum << " ";
    for(int i=0;i<criticalNum;i++)
        os << prim.criticalFrames[i] << " ";
    
    os << std::endl;
    
    return os;
}

