//
//  ivAppBA.cpp
//  Interactive Videos: Plausible Video Editing using Sparse Structure Points
//
//  Created by Chia-Sheng Chang on 03/01/16.
//  Copyright (c) 2016 Chia-Sheng Chang. All rights reserved.
//

#include "ivAppBA.h"

#define FULLQUATSZ 4
#define MAXITER 100

inline void _MK_QUAT_FRM_VEC(double *q, double*v)
{
    q[1]=(v)[0];
    q[2]=(v)[1];
    q[3]=(v)[2];
    q[0]=sqrt(1.0 - q[1]*q[1] - q[2]*q[2]- q[3]*q[3]);
}

typedef struct _globs_
{
    double *rot0params; /* initial rotation parameters, combined with a local rotation parameterization */
    double *intrcalib; /* the 5 intrinsic calibration parameters in the order [fu, u0, v0, ar, skew],
                        * where ar is the aspect ratio fv/fu.
                        * Used only when calibration is fixed for all cameras;
                        * otherwise, it is null and the intrinsic parameters are
                        * included in the set of motion parameters for each camera
                        */
    int nccalib; /* number of calibration parameters that must be kept constant.
                  * 0: all parameters are free
                  * 1: skew is fixed to its initial value, all other parameters vary (i.e. fu, u0, v0, ar)
                  * 2: skew and aspect ratio are fixed to their initial values, all other parameters vary (i.e. fu, u0, v0)
                  * 3: meaningless
                  * 4: skew, aspect ratio and principal point are fixed to their initial values, only the focal length varies (i.e. fu)
                  * 5: all intrinsics are kept fixed to their initial values
                  * >5: meaningless
                  * Used only when calibration varies among cameras
                  */
    
    int ncdist; /* number of distortion parameters in Bouguet's model that must be kept constant.
                 * 0: all parameters are free
                 * 1: 6th order radial distortion term (kc[4]) is fixed
                 * 2: 6th order radial distortion and one of the tangential distortion terms (kc[3]) are fixed
                 * 3: 6th order radial distortion and both tangential distortion terms (kc[3], kc[2]) are fixed [i.e., only 2nd & 4th order radial dist.]
                 * 4: 4th & 6th order radial distortion terms and both tangential distortion ones are fixed [i.e., only 2nd order radial dist.]
                 * 5: all distortion parameters are kept fixed to their initial values
                 * >5: meaningless
                 * Used only when calibration varies among cameras and distortion is to be estimated
                 */
    int cnp, pnp, mnp; /* dimensions */
    
    double *ptparams; /* needed only when bundle adjusting for camera parameters only */
    double *camparams; /* needed only when bundle adjusting for structure parameters only */
} Globs;

static void calcImgProjFullR
(double a[5], double qr0[4], double t[3], double M[3], double n[2])
{
    double t1;
    double t11;
    double t13;
    double t17;
    double t2;
    double t22;
    double t27;
    double t3;
    double t38;
    double t46;
    double t49;
    double t5;
    double t6;
    double t8;
    double t9;
    {
        t1 = a[0];
        t2 = qr0[1];
        t3 = M[0];
        t5 = qr0[2];
        t6 = M[1];
        t8 = qr0[3];
        t9 = M[2];
        t11 = -t3*t2-t5*t6-t8*t9;
        t13 = qr0[0];
        t17 = t13*t3+t5*t9-t8*t6;
        t22 = t6*t13+t8*t3-t9*t2;
        t27 = t13*t9+t6*t2-t5*t3;
        t38 = -t5*t11+t13*t22-t27*t2+t8*t17+t[1];
        t46 = -t11*t8+t13*t27-t5*t17+t2*t22+t[2];
        t49 = 1/t46;
        n[0] = (t1*(-t2*t11+t13*t17-t22*t8+t5*t27+t[0])+a[4]*t38+a[1]*t46)*t49;
        n[1] = (t1*a[3]*t38+a[2]*t46)*t49;
        return;
    }
}

static void img_projsS_x
(double *p, struct sba_crsm *idxij, int *rcidxs, int *rcsubs, double *hx, void *adata)
{
    int i, j;
    int cnp, pnp, mnp;
    double *pqr, *pt, *ppt, *pmeas, *Kparms, *camparams, trot[FULLQUATSZ];
    int m, nnz;
    Globs *gl;
    
    gl=(Globs *)adata;
    cnp=gl->cnp; pnp=gl->pnp; mnp=gl->mnp;
    Kparms=gl->intrcalib;
    camparams=gl->camparams;
    
    //n=idxij->nr;
    m=idxij->nc;
    
    for(j=0; j<m; ++j)
    {
        /* j-th camera parameters */
        pqr=camparams+j*cnp;
        pt=pqr+3; // quaternion vector part has 3 elements
        _MK_QUAT_FRM_VEC(trot, pqr);
        
        nnz=sba_crsm_col_elmidxs(idxij, j, rcidxs, rcsubs); /* find nonzero hx_ij, i=0...n-1 */
        
        for(i=0; i<nnz; ++i)
        {
            ppt=p + rcsubs[i]*pnp;
            pmeas=hx + idxij->val[rcidxs[i]]*mnp; // set pmeas to point to hx_ij
            
            calcImgProjFullR(Kparms, trot, pt, ppt, pmeas); // evaluate Q in pmeas
        }
    }
}

void ivAppBA::optimizeStructure
(tvPoint3DSet &vtvPoint, const tvTriPrimSet &vtvPrim, const ivAppSfM *sfmData)
{
    // const params
    const int CNP = 6;
    const int PNP = 3;
    const int MNP = 2;
    
    //camera parameter, 3D structure, projection
    int nframes = (int)sfmData->vFrameData.size();
    int numpts3D = (int)vtvPoint.size();

    cv::vector<double> motstruct( nframes * CNP + numpts3D * PNP );
    for(int i=0;i<nframes;i++)
    {
        const frameData &curFD = sfmData->vFrameData[i];

        //rotation vec
        Eigen::Matrix3f rotMatrix = curFD.M.block(0,0,3,3);
        Eigen::Quaternionf rotq(rotMatrix);
        Eigen::Vector3f rotvec = rotq.vec() / ( rotq.w()>=0.0 ? 1.0 : -1.0 );
        for(int j=0;j<3;j++)
            motstruct[ i*CNP + j ] = rotvec(j);

        //translation vec
        for(int j=0;j<3;j++)
            motstruct[ i*CNP + j + 3 ] = curFD.M(j,3);
    }

    //structure points
    for(int i=0;i<numpts3D;i++) for(int j=0;j<nframes;j++)
    {
        const tvPoint3D &curTVP = vtvPoint[i];
        
        if( !curTVP.isKey[j] || !curTVP.isReady3D[j] )
            continue;
       
        for(int k=0;k<3;k++)
            motstruct[ nframes*CNP + i*PNP + k ] = curTVP.tvv3D[j](k);
    }
    
    //visibility mask
    cv::vector<char> vmask( numpts3D * nframes, 0 );
    for(int i=0;i<nframes;i++)
    {
        std::vector<tvTriPrim>::const_iterator pit = vtvPrim.cbegin();
        for(;pit!=vtvPrim.cend();pit++)
        {
            if( !pit->isFront(i) )
                continue;
            
            const std::vector<int> &roiPts = pit->polygon;
            std::vector<int>::const_iterator ptit = roiPts.cbegin();
            for(;ptit!=roiPts.cend();ptit++)
                vmask[ nframes * *ptit + i ] = 1;
        }
    }
    
    //projection points
    int numprojs = 0;
    for(int i=0;i<nframes;i++) for(int j=0;j<numpts3D;j++)
        numprojs += vmask[ nframes * j + i ];
    cv::vector<double> imgpts( numprojs * MNP );
    
    int k=0;
    for(int j=0;j<numpts3D;j++) for(int i=0;i<nframes;i++)
    {
        if( !vmask[ nframes * j + i ] )
            continue;

        imgpts[ k * MNP + 0 ] = vtvPoint[j].tvv2D[i](0);
        imgpts[ k * MNP + 1 ] = vtvPoint[j].tvv2D[i](1);
        k++;
    }
    
    //intrinsics
    double K[5];
    K[0] = sfmData->K(0,0); // fu
    K[1] = sfmData->K(0,2); // u0
    K[2] = sfmData->K(1,2); // v0
    K[3] = sfmData->K(1,1) / sfmData->K(0,0); // ar
    K[4] = sfmData->K(0,1); // s

    Globs globs;
    globs.cnp = CNP; /* 3 rot params + 3 trans params */
    globs.pnp = PNP; /* euclidean 3D points */
    globs.mnp = MNP; /* image points are 2D */
    globs.ptparams = NULL;
    globs.camparams = motstruct.data();
    globs.intrcalib = K;
    
    // call sparse LM routine (BA for structure only)
    double opts[SBA_OPTSSZ], info[SBA_INFOSZ];
    opts[0] = SBA_INIT_MU;
    opts[1] = SBA_STOP_THRESH;
    opts[2] = SBA_STOP_THRESH;
    opts[3] = SBA_STOP_THRESH;
    opts[4] = 0.0;
    globs.camparams = motstruct.data();
    sba_str_levmar_x(numpts3D, 0, nframes, vmask.data(), motstruct.data()+nframes*CNP,
                     PNP, imgpts.data(), NULL, MNP, img_projsS_x, NULL,
                     (void *)(&globs), MAXITER, 0/*verbose*/, opts, info);
    
    //copy optimized structure points
    for(int i=0;i<numpts3D;i++)
    {
        for(int k=0;k<3;k++)
            vtvPoint[i].v3D(k) = motstruct[ nframes*CNP + i*PNP + k ];
        vtvPoint[i].v3D(3) = 1.0f;
    }
}